@@ -64,8 +64,8 @@ WindowGlobalParent* WindowContext::Canonical() {
   return static_cast<WindowGlobalParent*>(this);
 }
 
-bool WindowContext::IsCached() const {
-  return mBrowsingContext->mCurrentWindowContext != this;
+bool WindowContext::IsCurrent() const {
+  return mBrowsingContext->mCurrentWindowContext == this;
 }
 
 bool WindowContext::IsInBFCache() {
@@ -125,7 +125,7 @@ void WindowContext::AppendChildBrowsingContext(
 
   // If we're the current WindowContext in our BrowsingContext, make sure to
   // clear any cached `children` value.
-  if (!IsCached()) {
+  if (IsCurrent()) {
     BrowsingContext_Binding::ClearCachedChildrenValue(mBrowsingContext);
   }
 }
@@ -139,7 +139,7 @@ void WindowContext::RemoveChildBrowsingContext(
 
   // If we're the current WindowContext in our BrowsingContext, make sure to
   // clear any cached `children` value.
-  if (!IsCached()) {
+  if (IsCurrent()) {
     BrowsingContext_Binding::ClearCachedChildrenValue(mBrowsingContext);
   }
 }
@@ -280,6 +280,11 @@ void WindowContext::DidSet(FieldIndex<IDX_AllowJavascript>, bool aOldValue) {
   RecomputeCanExecuteScripts();
 }
 
+bool WindowContext::CanSet(FieldIndex<IDX_HasActivePeerConnections>, bool,
+                           ContentParent*) {
+  return XRE_IsParentProcess() && IsTop();
+}
+
 void WindowContext::RecomputeCanExecuteScripts(bool aApplyChanges) {
   const bool old = mCanExecuteScripts;
   if (!AllowJavascript()) {
@@ -335,10 +340,11 @@ void WindowContext::DidSet(FieldIndex<IDX_UserActivationState>) {
     USER_ACTIVATION_LOG(
         "Set user gesture start time for %s browsing context 0x%08" PRIx64,
         XRE_IsParentProcess() ? "Parent" : "Child", Id());
-    mUserGestureStart =
-        (GetUserActivationState() == UserActivation::State::FullActivated)
-            ? TimeStamp::Now()
-            : TimeStamp();
+    if (GetUserActivationState() == UserActivation::State::FullActivated) {
+      mUserGestureStart = TimeStamp::Now();
+    } else if (GetUserActivationState() == UserActivation::State::None) {
+      mUserGestureStart = TimeStamp();
+    }
   }
 }
 
@@ -454,13 +460,26 @@ bool WindowContext::HasBeenUserGestureActivated() {
   return GetUserActivationState() != UserActivation::State::None;
 }
 
+DOMHighResTimeStamp WindowContext::LastUserGestureTimeStamp() {
+  MOZ_ASSERT(IsInProcess());
+  if (nsGlobalWindowInner* innerWindow = GetInnerWindow()) {
+    if (Performance* perf = innerWindow->GetPerformance()) {
+      return perf->GetDOMTiming()->TimeStampToDOMHighRes(mUserGestureStart);
+    }
+  }
+  NS_WARNING(
+      "Unable to calculate DOMHighResTimeStamp for LastUserGestureTimeStamp");
+  return 0;
+}
+
 bool WindowContext::HasValidTransientUserGestureActivation() {
   MOZ_ASSERT(IsInProcess());
 
   if (GetUserActivationState() != UserActivation::State::FullActivated) {
-    MOZ_ASSERT(mUserGestureStart.IsNull(),
-               "mUserGestureStart should be null if the document hasn't ever "
-               "been activated by user gesture");
+    // mUserGestureStart should be null if the document hasn't ever been
+    // activated by user gesture
+    MOZ_ASSERT_IF(GetUserActivationState() == UserActivation::State::None,
+                  mUserGestureStart.IsNull());
     return false;
   }
 
@@ -476,7 +495,7 @@ bool WindowContext::HasValidTransientUserGestureActivation() {
 
 bool WindowContext::ConsumeTransientUserGestureActivation() {
   MOZ_ASSERT(IsInProcess());
-  MOZ_ASSERT(!IsCached());
+  MOZ_ASSERT(IsCurrent());
 
   if (!HasValidTransientUserGestureActivation()) {
     return false;
@@ -572,17 +591,17 @@ NS_IMPL_CYCLE_COLLECTION_TRACE_WRAPPERCACHE(WindowContext)
 namespace ipc {
 
 void IPDLParamTraits<dom::MaybeDiscarded<dom::WindowContext>>::Write(
-    IPC::Message* aMsg, IProtocol* aActor,
+    IPC::MessageWriter* aWriter, IProtocol* aActor,
     const dom::MaybeDiscarded<dom::WindowContext>& aParam) {
   uint64_t id = aParam.ContextId();
-  WriteIPDLParam(aMsg, aActor, id);
+  WriteIPDLParam(aWriter, aActor, id);
 }
 
 bool IPDLParamTraits<dom::MaybeDiscarded<dom::WindowContext>>::Read(
-    const IPC::Message* aMsg, PickleIterator* aIter, IProtocol* aActor,
+    IPC::MessageReader* aReader, IProtocol* aActor,
     dom::MaybeDiscarded<dom::WindowContext>* aResult) {
   uint64_t id = 0;
-  if (!ReadIPDLParam(aMsg, aIter, aActor, &id)) {
+  if (!ReadIPDLParam(aReader, aActor, &id)) {
     return false;
   }
 
@@ -597,24 +616,23 @@ bool IPDLParamTraits<dom::MaybeDiscarded<dom::WindowContext>>::Read(
 }
 
 void IPDLParamTraits<dom::WindowContext::IPCInitializer>::Write(
-    IPC::Message* aMessage, IProtocol* aActor,
+    IPC::MessageWriter* aWriter, IProtocol* aActor,
     const dom::WindowContext::IPCInitializer& aInit) {
   // Write actor ID parameters.
-  WriteIPDLParam(aMessage, aActor, aInit.mInnerWindowId);
-  WriteIPDLParam(aMessage, aActor, aInit.mOuterWindowId);
-  WriteIPDLParam(aMessage, aActor, aInit.mBrowsingContextId);
-  WriteIPDLParam(aMessage, aActor, aInit.mFields);
+  WriteIPDLParam(aWriter, aActor, aInit.mInnerWindowId);
+  WriteIPDLParam(aWriter, aActor, aInit.mOuterWindowId);
+  WriteIPDLParam(aWriter, aActor, aInit.mBrowsingContextId);
+  WriteIPDLParam(aWriter, aActor, aInit.mFields);
 }
 
 bool IPDLParamTraits<dom::WindowContext::IPCInitializer>::Read(
-    const IPC::Message* aMessage, PickleIterator* aIterator, IProtocol* aActor,
+    IPC::MessageReader* aReader, IProtocol* aActor,
     dom::WindowContext::IPCInitializer* aInit) {
   // Read actor ID parameters.
-  return ReadIPDLParam(aMessage, aIterator, aActor, &aInit->mInnerWindowId) &&
-         ReadIPDLParam(aMessage, aIterator, aActor, &aInit->mOuterWindowId) &&
-         ReadIPDLParam(aMessage, aIterator, aActor,
-                       &aInit->mBrowsingContextId) &&
-         ReadIPDLParam(aMessage, aIterator, aActor, &aInit->mFields);
+  return ReadIPDLParam(aReader, aActor, &aInit->mInnerWindowId) &&
+         ReadIPDLParam(aReader, aActor, &aInit->mOuterWindowId) &&
+         ReadIPDLParam(aReader, aActor, &aInit->mBrowsingContextId) &&
+         ReadIPDLParam(aReader, aActor, &aInit->mFields);
 }
 
 template struct IPDLParamTraits<dom::WindowContext::BaseTransaction>;
