@@ -20,13 +20,14 @@ void nsAuthSASL::Reset() { mSASLReady = false; }
 NS_IMPL_ISUPPORTS(nsAuthSASL, nsIAuthModule)
 
 NS_IMETHODIMP
-nsAuthSASL::Init(const char* serviceName, uint32_t serviceFlags,
-                 const char16_t* domain, const char16_t* username,
-                 const char16_t* password) {
+nsAuthSASL::Init(const nsACString& serviceName, uint32_t serviceFlags,
+                 const nsAString& domain, const nsAString& username,
+                 const nsAString& password) {
   nsresult rv;
 
-  NS_ASSERTION(username, "SASL requires a username");
-  NS_ASSERTION(!domain && !password, "unexpected credentials");
+  NS_ASSERTION(!username.IsEmpty(), "SASL requires a username");
+  NS_ASSERTION(domain.IsEmpty() && password.IsEmpty(),
+               "unexpected credentials");
 
   mUsername = username;
 
@@ -45,7 +46,7 @@ nsAuthSASL::Init(const char* serviceName, uint32_t serviceFlags,
 
   MOZ_ALWAYS_TRUE(mInnerModule = nsIAuthModule::CreateInstance(authType));
 
-  mInnerModule->Init(serviceName, serviceFlags, nullptr, nullptr, nullptr);
+  mInnerModule->Init(serviceName, serviceFlags, u""_ns, u""_ns, u""_ns);
 
   return NS_OK;
 }
