@@ -10,6 +10,8 @@
 #include "DrawTargetSkia.h"
 #include "DataSurfaceHelpers.h"
 #include "skia/include/core/SkData.h"
+#include "skia/include/core/SkImage.h"
+#include "skia/include/core/SkSurface.h"
 #include "mozilla/CheckedInt.h"
 
 namespace mozilla::gfx {
@@ -31,6 +33,15 @@ IntSize SourceSurfaceSkia::GetSize() const { return mSize; }
 
 SurfaceFormat SourceSurfaceSkia::GetFormat() const { return mFormat; }
 
+// This is only ever called by the DT destructor, which can only ever happen
+// from one place at a time. Therefore it doesn't need to hold the ChangeMutex
+// as mSurface is never read to directly and is just there to keep the object
+// alive, which itself is refcounted in a thread-safe manner.
+void SourceSurfaceSkia::GiveSurface(SkSurface* aSurface) {
+  mSurface.reset(aSurface);
+  mDrawTarget = nullptr;
+}
+
 sk_sp<SkImage> SourceSurfaceSkia::GetImage(Maybe<MutexAutoLock>* aLock) {
   // If we were provided a lock object, we can let the caller access
   // a shared SkImage and we know it won't go away while the lock is held.
@@ -148,7 +159,8 @@ uint8_t* SourceSurfaceSkia::GetData() {
   return reinterpret_cast<uint8_t*>(pixmap.writable_addr());
 }
 
-bool SourceSurfaceSkia::Map(MapType, MappedSurface* aMappedSurface) {
+bool SourceSurfaceSkia::Map(MapType, MappedSurface* aMappedSurface)
+    NO_THREAD_SAFETY_ANALYSIS {
   mChangeMutex.Lock();
   aMappedSurface->mData = GetData();
   aMappedSurface->mStride = Stride();
@@ -157,10 +169,14 @@ bool SourceSurfaceSkia::Map(MapType, MappedSurface* aMappedSurface) {
   if (!mIsMapped) {
     mChangeMutex.Unlock();
   }
+  // Static analysis will warn due to a conditional Unlock
+  PUSH_IGNORE_THREAD_SAFETY
   return isMapped;
+  POP_THREAD_SAFETY
 }
 
-void SourceSurfaceSkia::Unmap() {
+void SourceSurfaceSkia::Unmap() NO_THREAD_SAFETY_ANALYSIS {
+  mChangeMutex.AssertCurrentThreadOwns();
   MOZ_ASSERT(mIsMapped);
   mIsMapped = false;
   mChangeMutex.Unlock();
@@ -168,7 +184,7 @@ void SourceSurfaceSkia::Unmap() {
 
 void SourceSurfaceSkia::DrawTargetWillChange() {
   MutexAutoLock lock(mChangeMutex);
-  if (mDrawTarget) {
+  if (mDrawTarget.exchange(nullptr)) {
     // Raster snapshots do not use Skia's internal copy-on-write mechanism,
     // so we need to do an explicit copy here.
     // GPU snapshots, for which peekPixels is false, will already be dealt
@@ -181,7 +197,6 @@ void SourceSurfaceSkia::DrawTargetWillChange() {
         gfxCriticalError() << "Failed copying Skia raster snapshot";
       }
     }
-    mDrawTarget = nullptr;
   }
 }
 
