@@ -82,14 +82,19 @@ class CanvasEventRingBuffer final : public gfx::EventRingBuffer {
    * @param aReaderServices provides functions required by the reader
    * @returns true if initialization succeeds
    */
-  bool InitReader(const ipc::SharedMemoryBasic::Handle& aReadHandle,
-                  const CrossProcessSemaphoreHandle& aReaderSem,
-                  const CrossProcessSemaphoreHandle& aWriterSem,
+  bool InitReader(ipc::SharedMemoryBasic::Handle aReadHandle,
+                  CrossProcessSemaphoreHandle aReaderSem,
+                  CrossProcessSemaphoreHandle aWriterSem,
                   UniquePtr<ReaderServices> aReaderServices);
 
   bool good() const final { return mGood; }
 
-  void SetIsBad() final { mGood = false; }
+  bool WriterFailed() const { return mWrite->state == State::Failed; }
+
+  void SetIsBad() final {
+    mGood = false;
+    mRead->state = State::Failed;
+  }
 
   void write(const char* const aData, const size_t aSize) final;
 
@@ -178,7 +183,8 @@ class CanvasEventRingBuffer final : public gfx::EventRingBuffer {
      */
     AboutToWait,
     Waiting,
-    Stopped
+    Stopped,
+    Failed,
   };
 
   struct ReadFooter {
@@ -247,6 +253,9 @@ class CanvasDrawEventRecorder final : public gfx::DrawEventRecorderPrivate {
     aEvent.RecordToStream(mOutputStream);
   }
 
+  void StoreSourceSurfaceRecording(gfx::SourceSurface* aSurface,
+                                   const char* aReason) final;
+
   void RecordSourceSurfaceDestruction(void* aSurface) final;
 
   void Flush() final {}
