@@ -5,13 +5,21 @@
  * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
 
 #include "ImageContainer.h"
-#include <string.h>    // for memcpy, memset
-#include "GLImages.h"  // for SurfaceTextureImage
+
+#include <string.h>  // for memcpy, memset
+
+#include "GLImages.h"    // for SurfaceTextureImage
+#include "YCbCrUtils.h"  // for YCbCr conversions
 #include "gfx2DGlue.h"
 #include "gfxPlatform.h"  // for gfxPlatform
 #include "gfxUtils.h"     // for gfxUtils
 #include "libyuv.h"
-#include "mozilla/RefPtr.h"                 // for already_AddRefed
+#include "mozilla/CheckedInt.h"
+#include "mozilla/ProfilerLabels.h"
+#include "mozilla/RefPtr.h"  // for already_AddRefed
+#include "mozilla/StaticPrefs_layers.h"
+#include "mozilla/gfx/2D.h"
+#include "mozilla/gfx/gfxVars.h"
 #include "mozilla/ipc/CrossProcessMutex.h"  // for CrossProcessMutex, etc
 #include "mozilla/layers/CompositorTypes.h"
 #include "mozilla/layers/ImageBridgeChild.h"     // for ImageBridgeChild
@@ -19,24 +27,20 @@
 #include "mozilla/layers/ImageDataSerializer.h"  // for SurfaceDescriptorBuffer
 #include "mozilla/layers/LayersMessages.h"
 #include "mozilla/layers/SharedPlanarYCbCrImage.h"
-#include "mozilla/layers/SharedSurfacesChild.h"  // for SharedSurfacesAnimation
 #include "mozilla/layers/SharedRGBImage.h"
 #include "mozilla/layers/TextureClientRecycleAllocator.h"
-#include "mozilla/StaticPrefs_layers.h"
-#include "mozilla/gfx/gfxVars.h"
+#include "nsProxyRelease.h"
 #include "nsISupportsUtils.h"  // for NS_IF_ADDREF
-#include "YCbCrUtils.h"        // for YCbCr conversions
-#include "gfx2DGlue.h"
-#include "mozilla/gfx/2D.h"
-#include "mozilla/CheckedInt.h"
 
 #ifdef XP_MACOSX
+#  include "MacIOSurfaceImage.h"
 #  include "mozilla/gfx/QuartzSupport.h"
 #endif
 
 #ifdef XP_WIN
-#  include "gfxWindowsPlatform.h"
 #  include <d3d10_1.h>
+
+#  include "gfxWindowsPlatform.h"
 #  include "mozilla/gfx/DeviceManagerDx.h"
 #  include "mozilla/layers/D3D11YCbCrImage.h"
 #endif
@@ -85,10 +89,7 @@ UniquePtr<uint8_t[]> BufferRecycleBin::GetBuffer(uint32_t aSize) {
     return UniquePtr<uint8_t[]>(new (fallible) uint8_t[aSize]);
   }
 
-  uint32_t last = mRecycledBuffers.Length() - 1;
-  UniquePtr<uint8_t[]> result = std::move(mRecycledBuffers[last]);
-  mRecycledBuffers.RemoveElementAt(last);
-  return result;
+  return mRecycledBuffers.PopLastElement();
 }
 
 void BufferRecycleBin::ClearRecycledBuffers() {
@@ -173,13 +174,6 @@ void ImageContainer::EnsureImageClient() {
   }
 }
 
-SharedSurfacesAnimation* ImageContainer::EnsureSharedSurfacesAnimation() {
-  if (!mSharedAnimation) {
-    mSharedAnimation = new SharedSurfacesAnimation();
-  }
-  return mSharedAnimation;
-}
-
 ImageContainer::ImageContainer(Mode flag)
     : mRecursiveMutex("ImageContainer.mRecursiveMutex"),
       mGenerationCounter(++sGenerationCounter),
@@ -218,9 +212,27 @@ ImageContainer::~ImageContainer() {
       imageBridge->ForgetImageContainer(mAsyncContainerHandle);
     }
   }
-  if (mSharedAnimation) {
-    mSharedAnimation->Destroy();
+}
+
+Maybe<SurfaceDescriptor> Image::GetDesc() { return GetDescFromTexClient(); }
+
+Maybe<SurfaceDescriptor> Image::GetDescFromTexClient(
+    TextureClient* const tcOverride) {
+  RefPtr<TextureClient> tc = tcOverride;
+  if (!tcOverride) {
+    tc = GetTextureClient(nullptr);
   }
+  if (!tc) {
+    return {};
+  }
+
+  const auto& tcd = tc->GetInternalData();
+
+  SurfaceDescriptor ret;
+  if (!tcd->Serialize(ret)) {
+    return {};
+  }
+  return Some(ret);
 }
 
 RefPtr<PlanarYCbCrImage> ImageContainer::CreatePlanarYCbCrImage() {
@@ -238,10 +250,13 @@ RefPtr<PlanarYCbCrImage> ImageContainer::CreatePlanarYCbCrImage() {
 RefPtr<SharedRGBImage> ImageContainer::CreateSharedRGBImage() {
   RecursiveMutexAutoLock lock(mRecursiveMutex);
   EnsureImageClient();
-  if (!mImageClient || !mImageClient->AsImageClientSingle()) {
-    return nullptr;
+  if (mImageClient && mImageClient->AsImageClientSingle()) {
+    return new SharedRGBImage(mImageClient);
   }
-  return new SharedRGBImage(mImageClient);
+  if (mRecycleAllocator) {
+    return new SharedRGBImage(mRecycleAllocator);
+  }
+  return nullptr;
 }
 
 void ImageContainer::SetCurrentImageInternal(
@@ -288,7 +303,7 @@ void ImageContainer::SetCurrentImageInternal(
     }
   }
 
-  mCurrentImages.SwapElements(newImages);
+  mCurrentImages = std::move(newImages);
 }
 
 void ImageContainer::ClearImagesFromImageBridge() {
@@ -426,16 +441,7 @@ void ImageContainer::EnsureRecycleAllocatorForRDD(
     return;
   }
 
-  bool useRecycleAllocator =
-      StaticPrefs::layers_recycle_allocator_rdd_AtStartup();
-#ifdef XP_MACOSX
-  // Disable RecycleAllocator for RDD on MacOS without WebRender.
-  // Recycling caused rendering artifact on a MacOS PC with OpenGL compositor.
-  if (!gfxVars::UseWebRender()) {
-    useRecycleAllocator = false;
-  }
-#endif
-  if (!useRecycleAllocator) {
+  if (!StaticPrefs::layers_recycle_allocator_rdd_AtStartup()) {
     return;
   }
 
@@ -465,32 +471,126 @@ D3D11YCbCrRecycleAllocator* ImageContainer::GetD3D11YCbCrRecycleAllocator(
 }
 #endif
 
+#ifdef XP_MACOSX
+MacIOSurfaceRecycleAllocator*
+ImageContainer::GetMacIOSurfaceRecycleAllocator() {
+  if (!mMacIOSurfaceRecycleAllocator) {
+    mMacIOSurfaceRecycleAllocator = new MacIOSurfaceRecycleAllocator();
+  }
+
+  return mMacIOSurfaceRecycleAllocator;
+}
+#endif
+
+// -
+// https://searchfox.org/mozilla-central/source/dom/media/ipc/RemoteImageHolder.cpp#46
+
+Maybe<PlanarYCbCrData> PlanarYCbCrData::From(
+    const SurfaceDescriptorBuffer& sdb) {
+  if (sdb.desc().type() != BufferDescriptor::TYCbCrDescriptor) {
+    return {};
+  }
+  const YCbCrDescriptor& yuvDesc = sdb.desc().get_YCbCrDescriptor();
+
+  Maybe<Range<uint8_t>> buffer;
+  const MemoryOrShmem& memOrShmem = sdb.data();
+  switch (memOrShmem.type()) {
+    case MemoryOrShmem::Tuintptr_t:
+      gfxCriticalError() << "PlanarYCbCrData::From SurfaceDescriptorBuffer "
+                            "w/uintptr_t unsupported.";
+      break;
+    case MemoryOrShmem::TShmem:
+      buffer.emplace(memOrShmem.get_Shmem().Range<uint8_t>());
+      break;
+    default:
+      MOZ_ASSERT(false, "Unknown MemoryOrShmem type");
+      break;
+  }
+  if (!buffer) {
+    return {};
+  }
+
+  PlanarYCbCrData yuvData;
+  yuvData.mYStride = AssertedCast<int32_t>(yuvDesc.yStride());
+  yuvData.mCbCrStride = AssertedCast<int32_t>(yuvDesc.cbCrStride());
+  // default mYSkip, mCbSkip, mCrSkip because not held in YCbCrDescriptor
+  yuvData.mYSkip = yuvData.mCbSkip = yuvData.mCrSkip = 0;
+  yuvData.mPictureRect = yuvDesc.display();
+  yuvData.mStereoMode = yuvDesc.stereoMode();
+  yuvData.mColorDepth = yuvDesc.colorDepth();
+  yuvData.mYUVColorSpace = yuvDesc.yUVColorSpace();
+  yuvData.mColorRange = yuvDesc.colorRange();
+  yuvData.mChromaSubsampling = yuvDesc.chromaSubsampling();
+
+  const auto GetPlanePtr = [&](const uint32_t beginOffset,
+                               const gfx::IntSize size,
+                               const int32_t stride) -> uint8_t* {
+    if (size.width > stride) return nullptr;
+    auto bytesNeeded = CheckedInt<uintptr_t>(stride) *
+                       size.height;  // Don't accept `stride*(h-1)+w`.
+    bytesNeeded += beginOffset;
+    if (!bytesNeeded.isValid() || bytesNeeded.value() > buffer->length()) {
+      gfxCriticalError()
+          << "PlanarYCbCrData::From asked for out-of-bounds plane data.";
+      return nullptr;
+    }
+    return (buffer->begin() + beginOffset).get();
+  };
+  yuvData.mYChannel =
+      GetPlanePtr(yuvDesc.yOffset(), yuvDesc.ySize(), yuvData.mYStride);
+  yuvData.mCbChannel =
+      GetPlanePtr(yuvDesc.cbOffset(), yuvDesc.cbCrSize(), yuvData.mCbCrStride);
+  yuvData.mCrChannel =
+      GetPlanePtr(yuvDesc.crOffset(), yuvDesc.cbCrSize(), yuvData.mCbCrStride);
+
+  if (yuvData.mYSkip || yuvData.mCbSkip || yuvData.mCrSkip ||
+      yuvDesc.ySize().width < 0 || yuvDesc.ySize().height < 0 ||
+      yuvDesc.cbCrSize().width < 0 || yuvDesc.cbCrSize().height < 0 ||
+      yuvData.mYStride < 0 || yuvData.mCbCrStride < 0 || !yuvData.mYChannel ||
+      !yuvData.mCbChannel || !yuvData.mCrChannel) {
+    gfxCriticalError() << "Unusual PlanarYCbCrData: " << yuvData.mYSkip << ","
+                       << yuvData.mCbSkip << "," << yuvData.mCrSkip << ", "
+                       << yuvDesc.ySize().width << "," << yuvDesc.ySize().height
+                       << ", " << yuvDesc.cbCrSize().width << ","
+                       << yuvDesc.cbCrSize().height << ", " << yuvData.mYStride
+                       << "," << yuvData.mCbCrStride << ", "
+                       << yuvData.mYChannel << "," << yuvData.mCbChannel << ","
+                       << yuvData.mCrChannel;
+    return {};
+  }
+
+  return Some(yuvData);
+}
+
+// -
+
 PlanarYCbCrImage::PlanarYCbCrImage()
     : Image(nullptr, ImageFormat::PLANAR_YCBCR),
       mOffscreenFormat(SurfaceFormat::UNKNOWN),
       mBufferSize(0) {}
 
 nsresult PlanarYCbCrImage::BuildSurfaceDescriptorBuffer(
-    SurfaceDescriptorBuffer& aSdBuffer) {
+    SurfaceDescriptorBuffer& aSdBuffer,
+    const std::function<MemoryOrShmem(uint32_t)>& aAllocate) {
   const PlanarYCbCrData* pdata = GetData();
   MOZ_ASSERT(pdata, "must have PlanarYCbCrData");
   MOZ_ASSERT(pdata->mYSkip == 0 && pdata->mCbSkip == 0 && pdata->mCrSkip == 0,
              "YCbCrDescriptor doesn't hold skip values");
-  MOZ_ASSERT(pdata->mPicX == 0 && pdata->mPicY == 0,
-             "YCbCrDescriptor doesn't hold picx or picy");
 
+  auto ySize = pdata->YDataSize();
+  auto cbcrSize = pdata->CbCrDataSize();
   uint32_t yOffset;
   uint32_t cbOffset;
   uint32_t crOffset;
-  ImageDataSerializer::ComputeYCbCrOffsets(
-      pdata->mYStride, pdata->mYSize.height, pdata->mCbCrStride,
-      pdata->mCbCrSize.height, yOffset, cbOffset, crOffset);
+  ImageDataSerializer::ComputeYCbCrOffsets(pdata->mYStride, ySize.height,
+                                           pdata->mCbCrStride, cbcrSize.height,
+                                           yOffset, cbOffset, crOffset);
 
-  aSdBuffer.desc() = YCbCrDescriptor(
-      pdata->mYSize, pdata->mYStride, pdata->mCbCrSize, pdata->mCbCrStride,
-      yOffset, cbOffset, crOffset, pdata->mStereoMode, pdata->mColorDepth,
-      pdata->mYUVColorSpace, pdata->mColorRange,
-      /*hasIntermediateBuffer*/ false);
+  uint32_t bufferSize = ImageDataSerializer::ComputeYCbCrBufferSize(
+      ySize, pdata->mYStride, cbcrSize, pdata->mCbCrStride, yOffset, cbOffset,
+      crOffset);
+
+  aSdBuffer.data() = aAllocate(bufferSize);
 
   uint8_t* buffer = nullptr;
   const MemoryOrShmem& memOrShmem = aSdBuffer.data();
@@ -502,19 +602,24 @@ nsresult PlanarYCbCrImage::BuildSurfaceDescriptorBuffer(
       buffer = memOrShmem.get_Shmem().get<uint8_t>();
       break;
     default:
-      MOZ_ASSERT(false, "Unknown MemoryOrShmem type");
+      buffer = nullptr;
+      break;
   }
-  MOZ_ASSERT(buffer, "no valid buffer available to copy image data");
   if (!buffer) {
-    return NS_ERROR_INVALID_ARG;
+    return NS_ERROR_OUT_OF_MEMORY;
   }
 
-  CopyPlane(buffer + yOffset, pdata->mYChannel, pdata->mYSize, pdata->mYStride,
+  aSdBuffer.desc() = YCbCrDescriptor(
+      pdata->mPictureRect, ySize, pdata->mYStride, cbcrSize, pdata->mCbCrStride,
+      yOffset, cbOffset, crOffset, pdata->mStereoMode, pdata->mColorDepth,
+      pdata->mYUVColorSpace, pdata->mColorRange, pdata->mChromaSubsampling);
+
+  CopyPlane(buffer + yOffset, pdata->mYChannel, ySize, pdata->mYStride,
             pdata->mYSkip);
-  CopyPlane(buffer + cbOffset, pdata->mCbChannel, pdata->mCbCrSize,
-            pdata->mCbCrStride, pdata->mCbSkip);
-  CopyPlane(buffer + crOffset, pdata->mCrChannel, pdata->mCbCrSize,
-            pdata->mCbCrStride, pdata->mCrSkip);
+  CopyPlane(buffer + cbOffset, pdata->mCbChannel, cbcrSize, pdata->mCbCrStride,
+            pdata->mCbSkip);
+  CopyPlane(buffer + crOffset, pdata->mCrChannel, cbcrSize, pdata->mCbCrStride,
+            pdata->mCrSkip);
   return NS_OK;
 }
 
@@ -575,9 +680,11 @@ static void CopyPlane(uint8_t* aDst, const uint8_t* aSrc,
 bool RecyclingPlanarYCbCrImage::CopyData(const Data& aData) {
   // update buffer size
   // Use uint32_t throughout to match AllocateBuffer's param and mBufferSize
+  auto ySize = aData.YDataSize();
+  auto cbcrSize = aData.CbCrDataSize();
   const auto checkedSize =
-      CheckedInt<uint32_t>(aData.mCbCrStride) * aData.mCbCrSize.height * 2 +
-      CheckedInt<uint32_t>(aData.mYStride) * aData.mYSize.height;
+      CheckedInt<uint32_t>(aData.mCbCrStride) * cbcrSize.height * 2 +
+      CheckedInt<uint32_t>(aData.mYStride) * ySize.height;
 
   if (!checkedSize.isValid()) return false;
 
@@ -592,20 +699,19 @@ bool RecyclingPlanarYCbCrImage::CopyData(const Data& aData) {
 
   mData = aData;
   mData.mYChannel = mBuffer.get();
-  mData.mCbChannel = mData.mYChannel + mData.mYStride * mData.mYSize.height;
-  mData.mCrChannel =
-      mData.mCbChannel + mData.mCbCrStride * mData.mCbCrSize.height;
+  mData.mCbChannel = mData.mYChannel + mData.mYStride * ySize.height;
+  mData.mCrChannel = mData.mCbChannel + mData.mCbCrStride * cbcrSize.height;
   mData.mYSkip = mData.mCbSkip = mData.mCrSkip = 0;
 
-  CopyPlane(mData.mYChannel, aData.mYChannel, aData.mYSize, aData.mYStride,
+  CopyPlane(mData.mYChannel, aData.mYChannel, ySize, aData.mYStride,
             aData.mYSkip);
-  CopyPlane(mData.mCbChannel, aData.mCbChannel, aData.mCbCrSize,
-            aData.mCbCrStride, aData.mCbSkip);
-  CopyPlane(mData.mCrChannel, aData.mCrChannel, aData.mCbCrSize,
-            aData.mCbCrStride, aData.mCrSkip);
+  CopyPlane(mData.mCbChannel, aData.mCbChannel, cbcrSize, aData.mCbCrStride,
+            aData.mCbSkip);
+  CopyPlane(mData.mCrChannel, aData.mCrChannel, cbcrSize, aData.mCbCrStride,
+            aData.mCrSkip);
 
-  mSize = aData.mPicSize;
-  mOrigin = gfx::IntPoint(aData.mPicX, aData.mPicY);
+  mSize = aData.mPictureRect.Size();
+  mOrigin = aData.mPictureRect.TopLeft();
   return true;
 }
 
@@ -616,8 +722,8 @@ gfxImageFormat PlanarYCbCrImage::GetOffscreenFormat() const {
 
 bool PlanarYCbCrImage::AdoptData(const Data& aData) {
   mData = aData;
-  mSize = aData.mPicSize;
-  mOrigin = gfx::IntPoint(aData.mPicX, aData.mPicY);
+  mSize = aData.mPictureRect.Size();
+  mOrigin = aData.mPictureRect.TopLeft();
   return true;
 }
 
@@ -656,13 +762,20 @@ already_AddRefed<gfx::SourceSurface> PlanarYCbCrImage::GetAsSourceSurface() {
   return surface.forget();
 }
 
+PlanarYCbCrImage::~PlanarYCbCrImage() {
+  NS_ReleaseOnMainThread("PlanarYCbCrImage::mSourceSurface",
+                         mSourceSurface.forget());
+}
+
 NVImage::NVImage() : Image(nullptr, ImageFormat::NV_IMAGE), mBufferSize(0) {}
 
-NVImage::~NVImage() = default;
+NVImage::~NVImage() {
+  NS_ReleaseOnMainThread("NVImage::mSourceSurface", mSourceSurface.forget());
+}
 
 IntSize NVImage::GetSize() const { return mSize; }
 
-IntRect NVImage::GetPictureRect() const { return mData.GetPictureRect(); }
+IntRect NVImage::GetPictureRect() const { return mData.mPictureRect; }
 
 already_AddRefed<SourceSurface> NVImage::GetAsSourceSurface() {
   if (mSourceSurface) {
@@ -672,31 +785,30 @@ already_AddRefed<SourceSurface> NVImage::GetAsSourceSurface() {
 
   // Convert the current NV12 or NV21 data to YUV420P so that we can follow the
   // logics in PlanarYCbCrImage::GetAsSourceSurface().
-  const int bufferLength = mData.mYSize.height * mData.mYStride +
-                           mData.mCbCrSize.height * mData.mCbCrSize.width * 2;
-  auto* buffer = new uint8_t[bufferLength];
+  auto ySize = mData.YDataSize();
+  auto cbcrSize = mData.CbCrDataSize();
+  const int bufferLength =
+      ySize.height * mData.mYStride + cbcrSize.height * cbcrSize.width * 2;
+  UniquePtr<uint8_t[]> buffer(new uint8_t[bufferLength]);
 
   Data aData = mData;
-  aData.mCbCrStride = aData.mCbCrSize.width;
+  aData.mCbCrStride = cbcrSize.width;
   aData.mCbSkip = 0;
   aData.mCrSkip = 0;
-  aData.mYChannel = buffer;
-  aData.mCbChannel = aData.mYChannel + aData.mYSize.height * aData.mYStride;
-  aData.mCrChannel =
-      aData.mCbChannel + aData.mCbCrSize.height * aData.mCbCrStride;
+  aData.mYChannel = buffer.get();
+  aData.mCbChannel = aData.mYChannel + ySize.height * aData.mYStride;
+  aData.mCrChannel = aData.mCbChannel + cbcrSize.height * aData.mCbCrStride;
 
   if (mData.mCbChannel < mData.mCrChannel) {  // NV12
     libyuv::NV12ToI420(mData.mYChannel, mData.mYStride, mData.mCbChannel,
                        mData.mCbCrStride, aData.mYChannel, aData.mYStride,
                        aData.mCbChannel, aData.mCbCrStride, aData.mCrChannel,
-                       aData.mCbCrStride, aData.mYSize.width,
-                       aData.mYSize.height);
+                       aData.mCbCrStride, ySize.width, ySize.height);
   } else {  // NV21
     libyuv::NV21ToI420(mData.mYChannel, mData.mYStride, mData.mCrChannel,
                        mData.mCbCrStride, aData.mYChannel, aData.mYStride,
                        aData.mCbChannel, aData.mCbCrStride, aData.mCrChannel,
-                       aData.mCbCrStride, aData.mYSize.width,
-                       aData.mYSize.height);
+                       aData.mCbCrStride, ySize.width, ySize.height);
   }
 
   // The logics in PlanarYCbCrImage::GetAsSourceSurface().
@@ -726,9 +838,6 @@ already_AddRefed<SourceSurface> NVImage::GetAsSourceSurface() {
 
   mSourceSurface = surface;
 
-  // Release the temporary buffer.
-  delete[] buffer;
-
   return surface.forget();
 }
 
@@ -745,8 +854,8 @@ bool NVImage::SetData(const Data& aData) {
   // Calculate buffer size
   // Use uint32_t throughout to match AllocateBuffer's param and mBufferSize
   const auto checkedSize =
-      CheckedInt<uint32_t>(aData.mYSize.height) * aData.mYStride +
-      CheckedInt<uint32_t>(aData.mCbCrSize.height) * aData.mCbCrStride;
+      CheckedInt<uint32_t>(aData.YDataSize().height) * aData.mYStride +
+      CheckedInt<uint32_t>(aData.CbCrDataSize().height) * aData.mCbCrStride;
 
   if (!checkedSize.isValid()) return false;
 
@@ -768,7 +877,7 @@ bool NVImage::SetData(const Data& aData) {
   mData.mCrChannel = mData.mYChannel + (aData.mCrChannel - aData.mYChannel);
 
   // Update mSize.
-  mSize = aData.mPicSize;
+  mSize = aData.mPictureRect.Size();
 
   // Copy the input data into mBuffer.
   // This copies the y-channel and the interleaving CbCr-channel.
@@ -779,8 +888,8 @@ bool NVImage::SetData(const Data& aData) {
 
 const NVImage::Data* NVImage::GetData() const { return &mData; }
 
-UniquePtr<uint8_t> NVImage::AllocateBuffer(uint32_t aSize) {
-  UniquePtr<uint8_t> buffer(new uint8_t[aSize]);
+UniquePtr<uint8_t[]> NVImage::AllocateBuffer(uint32_t aSize) {
+  UniquePtr<uint8_t[]> buffer(new uint8_t[aSize]);
   return buffer;
 }
 
@@ -797,7 +906,10 @@ SourceSurfaceImage::SourceSurfaceImage(gfx::SourceSurface* aSourceSurface)
       mSourceSurface(aSourceSurface),
       mTextureFlags(TextureFlags::DEFAULT) {}
 
-SourceSurfaceImage::~SourceSurfaceImage() = default;
+SourceSurfaceImage::~SourceSurfaceImage() {
+  NS_ReleaseOnMainThread("SourceSurfaceImage::mSourceSurface",
+                         mSourceSurface.forget());
+}
 
 TextureClient* SourceSurfaceImage::GetTextureClient(
     KnowsCompositor* aKnowsCompositor) {
@@ -805,29 +917,28 @@ TextureClient* SourceSurfaceImage::GetTextureClient(
     return nullptr;
   }
 
-  auto entry = mTextureClients.LookupForAdd(aKnowsCompositor->GetSerial());
-  if (entry) {
-    return entry.Data();
-  }
+  return mTextureClients.WithEntryHandle(
+      aKnowsCompositor->GetSerial(), [&](auto&& entry) -> TextureClient* {
+        if (entry) {
+          return entry->get();
+        }
 
-  RefPtr<TextureClient> textureClient;
-  RefPtr<SourceSurface> surface = GetAsSourceSurface();
-  MOZ_ASSERT(surface);
-  if (surface) {
-    // gfx::BackendType::NONE means default to content backend
-    textureClient = TextureClient::CreateFromSurface(
-        aKnowsCompositor, surface, BackendSelector::Content, mTextureFlags,
-        ALLOC_DEFAULT);
-  }
-  if (textureClient) {
-    textureClient->SyncWithObject(aKnowsCompositor->GetSyncObject());
-    entry.OrInsert([&textureClient]() { return textureClient; });
-    return textureClient;
-  }
+        RefPtr<TextureClient> textureClient;
+        RefPtr<SourceSurface> surface = GetAsSourceSurface();
+        MOZ_ASSERT(surface);
+        if (surface) {
+          // gfx::BackendType::NONE means default to content backend
+          textureClient = TextureClient::CreateFromSurface(
+              aKnowsCompositor, surface, BackendSelector::Content,
+              mTextureFlags, ALLOC_DEFAULT);
+        }
+        if (textureClient) {
+          textureClient->SyncWithObject(aKnowsCompositor->GetSyncObject());
+          return entry.Insert(std::move(textureClient)).get();
+        }
 
-  // Remove the speculatively added entry.
-  entry.OrRemove();
-  return nullptr;
+        return nullptr;
+      });
 }
 
 ImageContainer::ProducerID ImageContainer::AllocateProducerID() {
