@@ -17,7 +17,6 @@
 #include "mozilla/EventForwards.h"  // for Modifiers
 #include "mozilla/layers/APZThreadUtils.h"
 #include "mozilla/layers/MatrixMessage.h"        // for MatrixMessage
-#include "mozilla/layers/RepaintRequest.h"       // for RepaintRequest
 #include "mozilla/layers/ScrollableLayerGuid.h"  // for ScrollableLayerGuid, etc
 #include "nsISupportsImpl.h"
 
@@ -27,6 +26,8 @@ class Runnable;
 
 namespace layers {
 
+struct RepaintRequest;
+
 class GeckoContentController {
  public:
   using APZStateChange = GeckoContentController_APZStateChange;
@@ -39,8 +40,7 @@ class GeckoContentController {
    *  MatrixMessage for each layers id in the current APZ tree, along with the
    * corresponding transform.
    */
-  virtual void NotifyLayerTransforms(
-      const nsTArray<MatrixMessage>& aTransforms) = 0;
+  virtual void NotifyLayerTransforms(nsTArray<MatrixMessage>&& aTransforms) = 0;
 
   /**
    * Requests a paint of the given RepaintRequest |aRequest| from Gecko.
@@ -74,6 +74,7 @@ class GeckoContentController {
    *        of the pinch.
    * @param aGuid The guid of the APZ that is detecting the pinch. This is
    *        generally the root APZC for the layers id.
+   * @param aFocusPoint The focus point of the pinch event.
    * @param aSpanChange For the START or END event, this is always 0.
    *        For a SCALE event, this is the difference in span between the
    *        previous state and the new state.
@@ -81,6 +82,7 @@ class GeckoContentController {
    */
   virtual void NotifyPinchGesture(PinchGestureInput::PinchGestureType aType,
                                   const ScrollableLayerGuid& aGuid,
+                                  const LayoutDevicePoint& aFocusPoint,
                                   LayoutDeviceCoord aSpanChange,
                                   Modifiers aModifiers) = 0;
 
@@ -145,9 +147,11 @@ class GeckoContentController {
 
   virtual void CancelAutoscroll(const ScrollableLayerGuid& aGuid) = 0;
 
-  virtual void UpdateOverscrollVelocity(float aX, float aY,
+  virtual void UpdateOverscrollVelocity(const ScrollableLayerGuid& aGuid,
+                                        float aX, float aY,
                                         bool aIsRootContent) {}
-  virtual void UpdateOverscrollOffset(float aX, float aY, bool aIsRootContent) {
+  virtual void UpdateOverscrollOffset(const ScrollableLayerGuid& aGuid,
+                                      float aX, float aY, bool aIsRootContent) {
   }
 
   GeckoContentController() = default;
@@ -162,6 +166,8 @@ class GeckoContentController {
    */
   virtual bool IsRemote() { return false; }
 
+  virtual PresShell* GetTopLevelPresShell() const { return nullptr; };
+
  protected:
   // Protected destructor, to discourage deletion outside of Release():
   virtual ~GeckoContentController() = default;
