@@ -5,49 +5,48 @@
  * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
 
 #include "ImageBridgeChild.h"
-#include <vector>                        // for vector
-#include "ImageBridgeParent.h"           // for ImageBridgeParent
-#include "ImageContainer.h"              // for ImageContainer
-#include "Layers.h"                      // for Layer, etc
-#include "ShadowLayers.h"                // for ShadowLayerForwarder
-#include "base/platform_thread.h"        // for PlatformThread
-#include "base/process.h"                // for ProcessId
-#include "base/task.h"                   // for NewRunnableFunction, etc
-#include "base/thread.h"                 // for Thread
-#include "mozilla/Assertions.h"          // for MOZ_ASSERT, etc
-#include "mozilla/Monitor.h"             // for Monitor, MonitorAutoLock
-#include "mozilla/ReentrantMonitor.h"    // for ReentrantMonitor, etc
-#include "mozilla/ipc/MessageChannel.h"  // for MessageChannel, etc
-#include "mozilla/ipc/Transport.h"       // for Transport
-#include "mozilla/gfx/gfxVars.h"
+
+#include <vector>  // for vector
+
+#include "ImageBridgeParent.h"  // for ImageBridgeParent
+#include "ImageContainer.h"     // for ImageContainer
+#include "Layers.h"             // for Layer, etc
+#include "SynchronousTask.h"
+#include "mozilla/Assertions.h"        // for MOZ_ASSERT, etc
+#include "mozilla/Monitor.h"           // for Monitor, MonitorAutoLock
+#include "mozilla/ReentrantMonitor.h"  // for ReentrantMonitor, etc
+#include "mozilla/StaticMutex.h"
+#include "mozilla/StaticPtr.h"  // for StaticRefPtr
+#include "mozilla/dom/ContentChild.h"
 #include "mozilla/gfx/Point.h"  // for IntSize
-#include "mozilla/layers/AsyncCanvasRenderer.h"
-#include "mozilla/media/MediaSystemResourceManager.h"  // for MediaSystemResourceManager
-#include "mozilla/media/MediaSystemResourceManagerChild.h"  // for MediaSystemResourceManagerChild
+#include "mozilla/gfx/gfxVars.h"
+#include "mozilla/ipc/Endpoint.h"
+#include "mozilla/ipc/MessageChannel.h"         // for MessageChannel, etc
 #include "mozilla/layers/CompositableClient.h"  // for CompositableChild, etc
 #include "mozilla/layers/CompositorThread.h"
 #include "mozilla/layers/ISurfaceAllocator.h"  // for ISurfaceAllocator
 #include "mozilla/layers/ImageClient.h"        // for ImageClient
 #include "mozilla/layers/LayersMessages.h"     // for CompositableOperation
 #include "mozilla/layers/TextureClient.h"      // for TextureClient
-#include "mozilla/dom/ContentChild.h"
+#include "mozilla/layers/TextureClient.h"
+#include "mozilla/media/MediaSystemResourceManager.h"  // for MediaSystemResourceManager
+#include "mozilla/media/MediaSystemResourceManagerChild.h"  // for MediaSystemResourceManagerChild
 #include "mozilla/mozalloc.h"  // for operator new, etc
-#include "mtransport/runnable_utils.h"
+#include "transport/runnable_utils.h"
 #include "nsContentUtils.h"
 #include "nsISupportsImpl.h"         // for ImageContainer::AddRef, etc
 #include "nsTArray.h"                // for AutoTArray, nsTArray, etc
 #include "nsTArrayForwardDeclare.h"  // for AutoTArray
-#include "nsThread.h"
-#include "nsThreadUtils.h"  // for NS_IsMainThread
-#include "mozilla/StaticMutex.h"
-#include "mozilla/StaticPtr.h"  // for StaticRefPtr
-#include "mozilla/layers/TextureClient.h"
-#include "SynchronousTask.h"
+#include "nsThreadUtils.h"           // for NS_IsMainThread
 
 #if defined(XP_WIN)
 #  include "mozilla/gfx/DeviceManagerDx.h"
 #endif
 
+#ifdef MOZ_WIDGET_ANDROID
+#  include "mozilla/layers/AndroidHardwareBuffer.h"
+#endif
+
 namespace mozilla {
 namespace ipc {
 class Shmem;
@@ -55,8 +54,6 @@ class Shmem;
 
 namespace layers {
 
-using base::ProcessId;
-using base::Thread;
 using namespace mozilla::ipc;
 using namespace mozilla::gfx;
 using namespace mozilla::media;
@@ -115,6 +112,15 @@ void ImageBridgeChild::UseTextures(
     }
 
     bool readLocked = t.mTextureClient->OnForwardedToHost();
+
+    auto fenceFd = t.mTextureClient->GetInternalData()->GetAcquireFence();
+    if (fenceFd.IsValid()) {
+      mTxn->AddNoSwapEdit(CompositableOperation(
+          aCompositable->GetIPCHandle(),
+          OpDeliverAcquireFence(nullptr, t.mTextureClient->GetIPDLActor(),
+                                fenceFd)));
+    }
+
     textures.AppendElement(
         TimedTexture(nullptr, t.mTextureClient->GetIPDLActor(), t.mTimeStamp,
                      t.mPictureRect, t.mFrameID, t.mProducerID, readLocked));
@@ -126,17 +132,21 @@ void ImageBridgeChild::UseTextures(
                                             OpUseTexture(textures)));
 }
 
-void ImageBridgeChild::UseComponentAlphaTextures(
-    CompositableClient* aCompositable, TextureClient* aTextureOnBlack,
-    TextureClient* aTextureOnWhite) {
-  MOZ_CRASH("should not be called");
-}
-
 void ImageBridgeChild::HoldUntilCompositableRefReleasedIfNecessary(
     TextureClient* aClient) {
   if (!aClient) {
     return;
   }
+
+#ifdef MOZ_WIDGET_ANDROID
+  auto bufferId = aClient->GetInternalData()->GetBufferId();
+  if (bufferId.isSome()) {
+    MOZ_ASSERT(aClient->GetFlags() & TextureFlags::WAIT_HOST_USAGE_END);
+    AndroidHardwareBufferManager::Get()->HoldUntilNotifyNotUsed(
+        bufferId.ref(), GetFwdTransactionId(), /* aUsesImageBridge */ true);
+  }
+#endif
+
   // Wait ReleaseCompositableRef only when TextureFlags::RECYCLE or
   // TextureFlags::WAIT_HOST_USAGE_END is set on ImageBridge.
   bool waitNotifyNotUsed =
@@ -168,11 +178,8 @@ void ImageBridgeChild::CancelWaitForNotifyNotUsed(uint64_t aTextureId) {
 }
 
 // Singleton
-static StaticMutex sImageBridgeSingletonLock;
+static StaticMutex sImageBridgeSingletonLock MOZ_UNANNOTATED;
 static StaticRefPtr<ImageBridgeChild> sImageBridgeChildSingleton;
-// sImageBridgeChildThread cannot be converted to use a generic
-// nsISerialEventTarget (which may be backed by a threadpool) until bug 1634846
-// is addressed. Therefore we keep it as an nsIThread here.
 static StaticRefPtr<nsIThread> sImageBridgeChildThread;
 
 // dispatched function
@@ -233,14 +240,6 @@ void ImageBridgeChild::CreateImageClientSync(SynchronousTask* aTask,
   *result = CreateImageClientNow(aType, aImageContainer);
 }
 
-// dispatched function
-void ImageBridgeChild::CreateCanvasClientSync(
-    SynchronousTask* aTask, CanvasClient::CanvasClientType aType,
-    TextureFlags aFlags, RefPtr<CanvasClient>* const outResult) {
-  AutoCompleteTask complete(aTask);
-  *outResult = CreateCanvasClientNow(aType, aFlags);
-}
-
 ImageBridgeChild::ImageBridgeChild(uint32_t aNamespace)
     : mNamespace(aNamespace),
       mCanSend(false),
@@ -285,8 +284,7 @@ void ImageBridgeChild::Connect(CompositableClient* aCompositable,
 
   CompositableHandle handle(id);
   aCompositable->InitIPDL(handle);
-  SendNewCompositable(handle, aCompositable->GetTextureInfo(),
-                      GetCompositorBackendType());
+  SendNewCompositable(handle, aCompositable->GetTextureInfo());
 }
 
 void ImageBridgeChild::ForgetImageContainer(const CompositableHandle& aHandle) {
@@ -333,45 +331,6 @@ void ImageBridgeChild::UpdateImageClient(RefPtr<ImageContainer> aContainer) {
   EndTransaction();
 }
 
-void ImageBridgeChild::UpdateAsyncCanvasRendererSync(
-    SynchronousTask* aTask, AsyncCanvasRenderer* aWrapper) {
-  AutoCompleteTask complete(aTask);
-
-  UpdateAsyncCanvasRendererNow(aWrapper);
-}
-
-void ImageBridgeChild::UpdateAsyncCanvasRenderer(
-    AsyncCanvasRenderer* aWrapper) {
-  aWrapper->GetCanvasClient()->UpdateAsync(aWrapper);
-
-  if (InImageBridgeChildThread()) {
-    UpdateAsyncCanvasRendererNow(aWrapper);
-    return;
-  }
-
-  SynchronousTask task("UpdateAsyncCanvasRenderer Lock");
-
-  RefPtr<Runnable> runnable = WrapRunnable(
-      RefPtr<ImageBridgeChild>(this),
-      &ImageBridgeChild::UpdateAsyncCanvasRendererSync, &task, aWrapper);
-  GetThread()->Dispatch(runnable.forget());
-
-  task.Wait();
-}
-
-void ImageBridgeChild::UpdateAsyncCanvasRendererNow(
-    AsyncCanvasRenderer* aWrapper) {
-  MOZ_ASSERT(aWrapper);
-
-  if (!CanSend()) {
-    return;
-  }
-
-  BeginTransaction();
-  aWrapper->GetCanvasClient()->Updated();
-  EndTransaction();
-}
-
 void ImageBridgeChild::FlushAllImagesSync(SynchronousTask* aTask,
                                           ImageClient* aClient,
                                           ImageContainer* aContainer) {
@@ -435,10 +394,6 @@ void ImageBridgeChild::EndTransaction() {
     cset.AppendElements(&mTxn->mOperations.front(), mTxn->mOperations.size());
   }
 
-  if (!IsSameProcess()) {
-    ShadowLayerForwarder::PlatformSyncBeforeUpdate();
-  }
-
   if (!SendUpdate(cset, mTxn->mDestroyedActors, GetFwdTransactionId())) {
     NS_WARNING("could not send async texture transaction");
     return;
@@ -461,10 +416,11 @@ bool ImageBridgeChild::InitForContent(Endpoint<PImageBridgeChild>&& aEndpoint,
 
   RefPtr<ImageBridgeChild> child = new ImageBridgeChild(aNamespace);
 
-  RefPtr<Runnable> runnable = NewRunnableMethod<Endpoint<PImageBridgeChild>&&>(
-      "layers::ImageBridgeChild::Bind", child, &ImageBridgeChild::Bind,
-      std::move(aEndpoint));
-  child->GetThread()->Dispatch(runnable.forget());
+  child->GetThread()->Dispatch(NS_NewRunnableFunction(
+      "layers::ImageBridgeChild::Bind",
+      [child, endpoint = std::move(aEndpoint)]() mutable {
+        child->Bind(std::move(endpoint));
+      }));
 
   // Assign this after so other threads can't post messages before we connect to
   // IPDL.
@@ -501,8 +457,7 @@ void ImageBridgeChild::Bind(Endpoint<PImageBridgeChild>&& aEndpoint) {
 }
 
 void ImageBridgeChild::BindSameProcess(RefPtr<ImageBridgeParent> aParent) {
-  ipc::MessageChannel* parentChannel = aParent->GetIPCChannel();
-  Open(parentChannel, aParent->GetThread(), mozilla::ipc::ChildSide);
+  Open(aParent, aParent->GetThread(), mozilla::ipc::ChildSide);
 
   // This reference is dropped in DeallocPImageBridgeChild.
   this->AddRef();
@@ -516,7 +471,10 @@ void ImageBridgeChild::ShutDown() {
 
   ShutdownSingleton();
 
-  sImageBridgeChildThread = nullptr;
+  if (sImageBridgeChildThread) {
+    sImageBridgeChildThread->Shutdown();
+    sImageBridgeChildThread = nullptr;
+  }
 }
 
 /* static */
@@ -597,9 +555,11 @@ void ImageBridgeChild::InitWithGPUProcess(
 
   RefPtr<ImageBridgeChild> child = new ImageBridgeChild(aNamespace);
 
-  child->GetThread()->Dispatch(NewRunnableMethod<Endpoint<PImageBridgeChild>&&>(
-      "layers::ImageBridgeChild::Bind", child, &ImageBridgeChild::Bind,
-      std::move(aEndpoint)));
+  child->GetThread()->Dispatch(NS_NewRunnableFunction(
+      "layers::ImageBridgeChild::Bind",
+      [child, endpoint = std::move(aEndpoint)]() mutable {
+        child->Bind(std::move(endpoint));
+      }));
 
   // Assign this after so other threads can't post messages before we connect to
   // IPDL.
@@ -628,58 +588,7 @@ void ImageBridgeChild::IdentifyCompositorTextureHost(
 
 void ImageBridgeChild::UpdateTextureFactoryIdentifier(
     const TextureFactoryIdentifier& aIdentifier) {
-  // ImageHost is incompatible between WebRender enabled and WebRender disabled.
-  // Then drop all ImageContainers' ImageClients during disabling WebRender.
-  bool disablingWebRender =
-      GetCompositorBackendType() == LayersBackend::LAYERS_WR &&
-      aIdentifier.mParentBackend != LayersBackend::LAYERS_WR;
-
-  // Do not update TextureFactoryIdentifier if aIdentifier is going to disable
-  // WebRender, but gecko is still using WebRender. Since gecko uses different
-  // incompatible ImageHost and TextureHost between WebRender and non-WebRender.
-  //
-  // Even when WebRender is still in use, if non-accelerated widget is opened,
-  // aIdentifier disables WebRender at ImageBridgeChild.
-  if (disablingWebRender && gfxVars::UseWebRender()) {
-    return;
-  }
-
-  // D3DTexture might become obsolte. To prevent to use obsoleted D3DTexture,
-  // drop all ImageContainers' ImageClients.
-
-  // During re-creating GPU process, there was a period that ImageBridgeChild
-  // was re-created, but ImageBridgeChild::UpdateTextureFactoryIdentifier() was
-  // not called yet. In the period, if ImageBridgeChild::CreateImageClient() is
-  // called, ImageBridgeParent creates incompatible ImageHost than
-  // WebRenderImageHost.
-  bool initializingWebRender =
-      GetCompositorBackendType() != LayersBackend::LAYERS_WR &&
-      aIdentifier.mParentBackend == LayersBackend::LAYERS_WR;
-
-  bool needsDrop = disablingWebRender || initializingWebRender;
-
-#if defined(XP_WIN)
-  RefPtr<ID3D11Device> device = gfx::DeviceManagerDx::Get()->GetImageDevice();
-  needsDrop |= !!mImageDevice && mImageDevice != device &&
-               GetCompositorBackendType() == LayersBackend::LAYERS_D3D11;
-  mImageDevice = device;
-#endif
-
   IdentifyTextureHost(aIdentifier);
-  if (needsDrop) {
-    nsTArray<RefPtr<ImageContainerListener> > listeners;
-    {
-      MutexAutoLock lock(mContainerMapLock);
-      for (const auto& entry : mImageContainerListeners) {
-        listeners.AppendElement(entry.second);
-      }
-    }
-    // Drop ImageContainer's ImageClient whithout holding mContainerMapLock to
-    // avoid deadlock.
-    for (auto container : listeners) {
-      container->DropImageClient();
-    }
-  }
 }
 
 RefPtr<ImageClient> ImageBridgeChild::CreateImageClient(
@@ -718,37 +627,6 @@ RefPtr<ImageClient> ImageBridgeChild::CreateImageClientNow(
   return client;
 }
 
-already_AddRefed<CanvasClient> ImageBridgeChild::CreateCanvasClient(
-    CanvasClient::CanvasClientType aType, TextureFlags aFlag) {
-  if (InImageBridgeChildThread()) {
-    return CreateCanvasClientNow(aType, aFlag);
-  }
-
-  SynchronousTask task("CreateCanvasClient Lock");
-
-  // RefPtrs on arguments are not needed since this dispatches synchronously.
-  RefPtr<CanvasClient> result = nullptr;
-  RefPtr<Runnable> runnable = WrapRunnable(
-      RefPtr<ImageBridgeChild>(this), &ImageBridgeChild::CreateCanvasClientSync,
-      &task, aType, aFlag, &result);
-  GetThread()->Dispatch(runnable.forget());
-
-  task.Wait();
-
-  return result.forget();
-}
-
-already_AddRefed<CanvasClient> ImageBridgeChild::CreateCanvasClientNow(
-    CanvasClient::CanvasClientType aType, TextureFlags aFlag) {
-  RefPtr<CanvasClient> client =
-      CanvasClient::CreateCanvasClient(aType, this, aFlag);
-  MOZ_ASSERT(client, "failed to create CanvasClient");
-  if (client) {
-    client->Connect();
-  }
-  return client.forget();
-}
-
 bool ImageBridgeChild::AllocUnsafeShmem(
     size_t aSize, ipc::SharedMemory::SharedMemoryType aType,
     ipc::Shmem* aShmem) {
@@ -849,7 +727,7 @@ bool ImageBridgeChild::DeallocShmem(ipc::Shmem& aShmem) {
 }
 
 PTextureChild* ImageBridgeChild::AllocPTextureChild(
-    const SurfaceDescriptor&, const ReadLockDescriptor&, const LayersBackend&,
+    const SurfaceDescriptor&, ReadLockDescriptor&, const LayersBackend&,
     const TextureFlags&, const uint64_t& aSerial,
     const wr::MaybeExternalImageId& aExternalImageId) {
   MOZ_ASSERT(CanSend());
@@ -884,6 +762,21 @@ mozilla::ipc::IPCResult ImageBridgeChild::RecvParentAsyncMessages(
         NotifyNotUsed(op.TextureId(), op.fwdTransactionId());
         break;
       }
+      case AsyncParentMessageData::TOpDeliverReleaseFence: {
+#ifdef MOZ_WIDGET_ANDROID
+        const OpDeliverReleaseFence& op = message.get_OpDeliverReleaseFence();
+        ipc::FileDescriptor fenceFd;
+        if (op.fenceFd().isSome()) {
+          fenceFd = *op.fenceFd();
+        }
+        AndroidHardwareBufferManager::Get()->NotifyNotUsed(
+            std::move(fenceFd), op.bufferId(), op.fwdTransactionId(),
+            op.usesImageBridge());
+#else
+        MOZ_ASSERT_UNREACHABLE("unexpected to be called");
+#endif
+        break;
+      }
       default:
         NS_ERROR("unknown AsyncParentMessageData type");
         return IPC_FAIL_NO_REASON(this);
@@ -925,12 +818,13 @@ mozilla::ipc::IPCResult ImageBridgeChild::RecvReportFramesDropped(
 }
 
 PTextureChild* ImageBridgeChild::CreateTexture(
-    const SurfaceDescriptor& aSharedData, const ReadLockDescriptor& aReadLock,
+    const SurfaceDescriptor& aSharedData, ReadLockDescriptor&& aReadLock,
     LayersBackend aLayersBackend, TextureFlags aFlags, uint64_t aSerial,
-    wr::MaybeExternalImageId& aExternalImageId, nsIEventTarget* aTarget) {
+    wr::MaybeExternalImageId& aExternalImageId) {
   MOZ_ASSERT(CanSend());
-  return SendPTextureConstructor(aSharedData, aReadLock, aLayersBackend, aFlags,
-                                 aSerial, aExternalImageId);
+  return SendPTextureConstructor(aSharedData, std::move(aReadLock),
+                                 aLayersBackend, aFlags, aSerial,
+                                 aExternalImageId);
 }
 
 static bool IBCAddOpDestroy(CompositableTransaction* aTxn,
