@@ -183,12 +183,10 @@ void gfxASurface::Init(cairo_surface_t* surface, bool existingSurface) {
     mFloatingRefs = 0;
   } else {
     mFloatingRefs = 1;
-#ifdef MOZ_TREE_CAIRO
     if (cairo_surface_get_content(surface) != CAIRO_CONTENT_COLOR) {
       cairo_surface_set_subpixel_antialiasing(
           surface, CAIRO_SUBPIXEL_ANTIALIASING_DISABLED);
     }
-#endif
   }
 }
 
@@ -251,24 +249,6 @@ void gfxASurface::Finish() {
   cairo_surface_finish(mSurface);
 }
 
-already_AddRefed<gfxASurface> gfxASurface::CreateSimilarSurface(
-    gfxContentType aContent, const IntSize& aSize) {
-  if (!mSurface || !mSurfaceValid) {
-    return nullptr;
-  }
-
-  cairo_surface_t* surface = cairo_surface_create_similar(
-      mSurface, cairo_content_t(int(aContent)), aSize.width, aSize.height);
-  if (cairo_surface_status(surface)) {
-    cairo_surface_destroy(surface);
-    return nullptr;
-  }
-
-  RefPtr<gfxASurface> result = Wrap(surface, aSize);
-  cairo_surface_destroy(surface);
-  return result.forget();
-}
-
 already_AddRefed<gfxImageSurface> gfxASurface::CopyToARGB32ImageSurface() {
   if (!mSurface || !mSurfaceValid) {
     return nullptr;
