@@ -45,11 +45,14 @@
 #include "nsDirectoryServiceDefs.h"
 #include "nsAppDirectoryServiceDefs.h"
 #include "nsIMemory.h"
+#include "nsMemory.h"
+#include "nsPresContext.h"
 #include "gfxFontConstants.h"
 
 #include "mozilla/EndianUtils.h"
 #include "mozilla/Preferences.h"
 #include "mozilla/scache/StartupCache.h"
+#include "mozilla/Telemetry.h"
 #include <fcntl.h>
 #include <sys/mman.h>
 #include <sys/stat.h>
@@ -172,27 +175,29 @@ gfxFont* FT2FontEntry::CreateFontInstance(const gfxFontStyle* aStyle) {
     return nullptr;
   }
 
-  // If variations are present, we will not use our cached mFTFace
-  // but always create a new one as it will have custom variation
-  // coordinates applied.
-  if ((!mVariationSettings.IsEmpty() ||
-       (aStyle && !aStyle->variationSettings.IsEmpty())) &&
-      (face->GetFace()->face_flags & FT_FACE_FLAG_MULTIPLE_MASTERS)) {
-    // Create a separate FT_Face because we need to apply custom
-    // variation settings to it.
-    RefPtr<SharedFTFace> varFace;
-    if (!mFilename.IsEmpty() && mFilename[0] == '/') {
-      varFace =
-          Factory::NewSharedFTFace(nullptr, mFilename.get(), mFTFontIndex);
-    } else {
-      varFace = face->GetData()->CloneFace(mFTFontIndex);
-    }
-    if (varFace) {
-      // Resolve variations from entry (descriptor) and style (property)
-      AutoTArray<gfxFontVariation, 8> settings;
-      GetVariationsForStyle(settings, aStyle ? *aStyle : gfxFontStyle());
-      gfxFT2FontBase::SetupVarCoords(GetMMVar(), settings, varFace->GetFace());
-      face = std::move(varFace);
+  if (face->GetFace()->face_flags & FT_FACE_FLAG_MULTIPLE_MASTERS) {
+    // Resolve variations from entry (descriptor) and style (property)
+    AutoTArray<gfxFontVariation, 8> settings;
+    GetVariationsForStyle(settings, aStyle ? *aStyle : gfxFontStyle());
+
+    // If variations are present, we will not use our cached mFTFace
+    // but always create a new one as it will have custom variation
+    // coordinates applied.
+    if (!settings.IsEmpty()) {
+      // Create a separate FT_Face because we need to apply custom
+      // variation settings to it.
+      RefPtr<SharedFTFace> varFace;
+      if (!mFilename.IsEmpty() && mFilename[0] == '/') {
+        varFace =
+            Factory::NewSharedFTFace(nullptr, mFilename.get(), mFTFontIndex);
+      } else {
+        varFace = face->GetData()->CloneFace(mFTFontIndex);
+      }
+      if (varFace) {
+        gfxFT2FontBase::SetupVarCoords(GetMMVar(), settings,
+                                       varFace->GetFace());
+        face = std::move(varFace);
+      }
     }
   }
 
@@ -302,6 +307,10 @@ static void SetPropertiesFromFace(gfxFontEntry* aFontEntry,
       (style & 2) ? FontSlantStyle::Italic() : FontSlantStyle::Normal());
   aFontEntry->mWeightRange = WeightRange(FontWeight(int(os2weight)));
   aFontEntry->mStretchRange = StretchRange(FontStretch(stretch));
+
+  // For variable fonts, update the style/weight/stretch attributes if the
+  // corresponding variation axes are present.
+  aFontEntry->SetupVariationRanges();
 }
 
 // Used to create the font entry for installed faces on the device,
@@ -346,21 +355,22 @@ FT2FontEntry* gfxFT2Font::GetFontEntry() {
 // properly.
 
 nsresult FT2FontEntry::ReadCMAP(FontInfoData* aFontInfoData) {
-  if (mCharacterMap) {
+  if (mCharacterMap || mShmemCharacterMap) {
     return NS_OK;
   }
 
   RefPtr<gfxCharacterMap> charmap = new gfxCharacterMap();
 
   nsresult rv = NS_ERROR_NOT_AVAILABLE;
-  hb_blob_t* cmapBlob = GetFontTable(TTAG_cmap);
-  if (cmapBlob) {
+  uint32_t uvsOffset = 0;
+  if (hb_blob_t* cmapBlob = GetFontTable(TTAG_cmap)) {
     unsigned int length;
     const char* data = hb_blob_get_data(cmapBlob, &length);
     rv = gfxFontUtils::ReadCMAP((const uint8_t*)data, length, *charmap,
-                                mUVSOffset);
+                                uvsOffset);
     hb_blob_destroy(cmapBlob);
   }
+  mUVSOffset.exchange(uvsOffset);
 
   if (NS_SUCCEEDED(rv) && !mIsDataUserFont && !HasGraphiteTables()) {
     // For downloadable fonts, trust the author and don't
@@ -396,39 +406,119 @@ nsresult FT2FontEntry::ReadCMAP(FontInfoData* aFontInfoData) {
   }
 #endif
 
-  mHasCmapTable = NS_SUCCEEDED(rv);
-  if (mHasCmapTable) {
+  bool setCharMap = true;
+  if (NS_SUCCEEDED(rv)) {
     gfxPlatformFontList* pfl = gfxPlatformFontList::PlatformFontList();
-    mCharacterMap = pfl->FindCharMap(charmap);
+    fontlist::FontList* sharedFontList = pfl->SharedFontList();
+    if (!IsUserFont() && mShmemFace) {
+      mShmemFace->SetCharacterMap(sharedFontList, charmap);  // async
+      if (TrySetShmemCharacterMap()) {
+        setCharMap = false;
+      }
+    } else {
+      charmap = pfl->FindCharMap(charmap);
+    }
+    mHasCmapTable = true;
   } else {
     // if error occurred, initialize to null cmap
-    mCharacterMap = new gfxCharacterMap();
+    charmap = new gfxCharacterMap();
+    mHasCmapTable = false;
   }
+  if (setCharMap) {
+    if (mCharacterMap.compareExchange(nullptr, charmap.get())) {
+      // We forget rather than addref because we don't use the charmap below.
+      Unused << charmap.forget();
+    }
+  }
+
   return rv;
 }
 
+hb_face_t* FT2FontEntry::CreateHBFace() const {
+  hb_face_t* result = nullptr;
+
+  if (mFilename[0] == '/') {
+    // An absolute path means a normal file in the filesystem, so we can use
+    // hb_blob_create_from_file to read it.
+    hb_blob_t* fileBlob = hb_blob_create_from_file(mFilename.get());
+    if (hb_blob_get_length(fileBlob) > 0) {
+      result = hb_face_create(fileBlob, mFTFontIndex);
+    }
+    hb_blob_destroy(fileBlob);
+  } else {
+    // A relative path means an omnijar resource, which we may need to
+    // decompress to a temporary buffer.
+    RefPtr<nsZipArchive> reader = Omnijar::GetReader(Omnijar::Type::GRE);
+    nsZipItem* item = reader->GetItem(mFilename.get());
+    MOZ_ASSERT(item, "failed to find zip entry");
+    if (item) {
+      // TODO(jfkthame):
+      // Check whether the item is compressed; if not, we could just get a
+      // pointer without needing to allocate a buffer and copy the data.
+      // (Currently this configuration isn't used for Gecko on Android.)
+      uint32_t length = item->RealSize();
+      uint8_t* buffer = static_cast<uint8_t*>(malloc(length));
+      if (buffer) {
+        nsZipCursor cursor(item, reader, buffer, length);
+        cursor.Copy(&length);
+        MOZ_ASSERT(length == item->RealSize(), "error reading font");
+        if (length == item->RealSize()) {
+          hb_blob_t* blob =
+              hb_blob_create((const char*)buffer, length,
+                             HB_MEMORY_MODE_READONLY, buffer, free);
+          result = hb_face_create(blob, mFTFontIndex);
+          hb_blob_destroy(blob);
+        }
+      }
+    }
+  }
+
+  return result;
+}
+
+bool FT2FontEntry::HasFontTable(uint32_t aTableTag) {
+  if (mAvailableTables.Count() > 0) {
+    return mAvailableTables.Contains(aTableTag);
+  }
+
+  // If we haven't created a FreeType face already, try to avoid that by
+  // reading the available table tags via harfbuzz and caching in a hashset.
+  if (!mFTFace && !mFilename.IsEmpty()) {
+    hb_face_t* face = CreateHBFace();
+    if (face) {
+      // Read table tags in batches; 32 should be enough for most fonts in a
+      // single operation.
+      const unsigned TAG_BUF_LENGTH = 32;
+      hb_tag_t tags[TAG_BUF_LENGTH];
+      unsigned int startOffset = 0;
+      unsigned int totalTables = 0;
+      do {
+        unsigned int count = TAG_BUF_LENGTH;
+        // Updates count to the number of table tags actually retrieved
+        totalTables = hb_face_get_table_tags(face, startOffset, &count, tags);
+        startOffset += count;
+        while (count-- > 0) {
+          mAvailableTables.Insert(tags[count]);
+        }
+      } while (startOffset < totalTables);
+      hb_face_destroy(face);
+    } else {
+      // Failed to create the HarfBuzz face! The font is probably broken.
+      // Put a dummy entry in mAvailableTables so that we don't bother
+      // re-trying here.
+      mAvailableTables.Insert(uint32_t(-1));
+    }
+    return mAvailableTables.Contains(aTableTag);
+  }
+
+  RefPtr<SharedFTFace> face = GetFTFace();
+  return gfxFT2FontEntryBase::FaceHasTable(face, aTableTag);
+}
+
 nsresult FT2FontEntry::CopyFontTable(uint32_t aTableTag,
                                      nsTArray<uint8_t>& aBuffer) {
   RefPtr<SharedFTFace> face = GetFTFace();
-  if (!face) {
-    return NS_ERROR_FAILURE;
-  }
-
-  FT_Error status;
-  FT_ULong len = 0;
-  status = FT_Load_Sfnt_Table(face->GetFace(), aTableTag, 0, nullptr, &len);
-  if (status != FT_Err_Ok || len == 0) {
-    return NS_ERROR_FAILURE;
-  }
-
-  if (!aBuffer.SetLength(len, fallible)) {
-    return NS_ERROR_OUT_OF_MEMORY;
-  }
-  uint8_t* buf = aBuffer.Elements();
-  status = FT_Load_Sfnt_Table(face->GetFace(), aTableTag, 0, buf, &len);
-  NS_ENSURE_TRUE(status == FT_Err_Ok, NS_ERROR_FAILURE);
-
-  return NS_OK;
+  return gfxFT2FontEntryBase::CopyFaceTable(face, aTableTag, aBuffer);
 }
 
 hb_blob_t* FT2FontEntry::GetFontTable(uint32_t aTableTag) {
@@ -444,47 +534,10 @@ hb_blob_t* FT2FontEntry::GetFontTable(uint32_t aTableTag) {
   // If the FT_Face hasn't been instantiated, try to read table directly
   // via harfbuzz API to avoid expensive FT_Face creation.
   if (!mFTFace && !mFilename.IsEmpty()) {
-    hb_blob_t* result = nullptr;
-    if (mFilename[0] == '/') {
-      // An absolute path means a normal file in the filesystem, so we can use
-      // hb_blob_create_from_file to read it.
-      hb_blob_t* fileBlob = hb_blob_create_from_file(mFilename.get());
-      if (hb_blob_get_length(fileBlob) > 0) {
-        hb_face_t* face = hb_face_create(fileBlob, mFTFontIndex);
-        result = hb_face_reference_table(face, aTableTag);
-        hb_face_destroy(face);
-      }
-      hb_blob_destroy(fileBlob);
-    } else {
-      // A relative path means an omnijar resource, which we may need to
-      // decompress to a temporary buffer.
-      RefPtr<nsZipArchive> reader = Omnijar::GetReader(Omnijar::Type::GRE);
-      nsZipItem* item = reader->GetItem(mFilename.get());
-      MOZ_ASSERT(item, "failed to find zip entry");
-      if (item) {
-        // TODO(jfkthame):
-        // Check whether the item is compressed; if not, we could just get a
-        // pointer without needing to allocate a buffer and copy the data.
-        // (Currently this configuration isn't used for Gecko on Android.)
-        uint32_t length = item->RealSize();
-        uint8_t* buffer = static_cast<uint8_t*>(malloc(length));
-        if (buffer) {
-          nsZipCursor cursor(item, reader, buffer, length);
-          cursor.Copy(&length);
-          MOZ_ASSERT(length == item->RealSize(), "error reading font");
-          if (length == item->RealSize()) {
-            hb_blob_t* blob =
-                hb_blob_create((const char*)buffer, length,
-                               HB_MEMORY_MODE_READONLY, buffer, free);
-            hb_face_t* face = hb_face_create(blob, mFTFontIndex);
-            result = hb_face_reference_table(face, aTableTag);
-            hb_face_destroy(face);
-            hb_blob_destroy(blob);
-          }
-        }
-      }
-    }
-    if (result) {
+    hb_face_t* face = CreateHBFace();
+    if (face) {
+      hb_blob_t* result = hb_face_reference_table(face, aTableTag);
+      hb_face_destroy(face);
       return result;
     }
   }
@@ -566,6 +619,7 @@ void FT2FontEntry::AddSizeOfIncludingThis(MallocSizeOf aMallocSizeOf,
  */
 
 void FT2FontFamily::AddFacesToFontList(nsTArray<FontListEntry>* aFontList) {
+  AutoReadLock lock(mLock);
   for (int i = 0, n = mAvailableFonts.Length(); i < n; ++i) {
     const FT2FontEntry* fe =
         static_cast<const FT2FontEntry*>(mAvailableFonts[i].get());
@@ -597,30 +651,30 @@ void gfxFT2FontList::CollectInitData(const FontListEntry& aFLE,
                                      StandardFile aStdFile) {
   nsAutoCString key(aFLE.familyName());
   BuildKeyNameFromFontName(key);
-  auto faceList = mFaceInitData.Get(key);
-  if (!faceList) {
-    faceList = new nsTArray<fontlist::Face::InitData>;
-    mFaceInitData.Put(key, faceList);
-    mFamilyInitData.AppendElement(
-        fontlist::Family::InitData{key, aFLE.familyName()});
-  }
-  uint32_t faceIndex = faceList->Length();
-  faceList->AppendElement(
-      fontlist::Face::InitData{aFLE.filepath(), aFLE.index(), false,
-                               WeightRange::FromScalar(aFLE.weightRange()),
-                               StretchRange::FromScalar(aFLE.stretchRange()),
-                               SlantStyleRange::FromScalar(aFLE.styleRange())});
+  mFaceInitData
+      .LookupOrInsertWith(
+          key,
+          [&] {
+            mFamilyInitData.AppendElement(
+                fontlist::Family::InitData{key, aFLE.familyName()});
+            return MakeUnique<nsTArray<fontlist::Face::InitData>>();
+          })
+      ->AppendElement(fontlist::Face::InitData{
+          aFLE.filepath(), aFLE.index(), false,
+          WeightRange::FromScalar(aFLE.weightRange()),
+          StretchRange::FromScalar(aFLE.stretchRange()),
+          SlantStyleRange::FromScalar(aFLE.styleRange())});
   nsAutoCString psname(aPSName), fullname(aFullName);
   if (!psname.IsEmpty()) {
     ToLowerCase(psname);
-    mLocalNameTable.Put(psname,
-                        fontlist::LocalFaceRec::InitData(key, faceIndex));
+    mLocalNameTable.InsertOrUpdate(
+        psname, fontlist::LocalFaceRec::InitData(key, aFLE.filepath()));
   }
   if (!fullname.IsEmpty()) {
     ToLowerCase(fullname);
     if (fullname != psname) {
-      mLocalNameTable.Put(fullname,
-                          fontlist::LocalFaceRec::InitData(key, faceIndex));
+      mLocalNameTable.InsertOrUpdate(
+          fullname, fontlist::LocalFaceRec::InitData(key, aFLE.filepath()));
     }
   }
 }
@@ -663,7 +717,7 @@ class FontNameCache {
   size_t EntryCount() const { return mMap.EntryCount(); }
 
   void DropStaleEntries() {
-    for (auto iter = mMap.Iter(); !iter.Done(); iter.Next()) {
+    for (auto iter = mMap.ConstIter(); !iter.Done(); iter.Next()) {
       auto entry = static_cast<FNCMapEntry*>(iter.Get());
       if (!entry->mFileExists) {
         iter.Remove();
@@ -678,7 +732,7 @@ class FontNameCache {
 
     LOG(("Writing FontNameCache:"));
     nsAutoCString buf;
-    for (auto iter = mMap.Iter(); !iter.Done(); iter.Next()) {
+    for (auto iter = mMap.ConstIter(); !iter.Done(); iter.Next()) {
       auto entry = static_cast<FNCMapEntry*>(iter.Get());
       MOZ_ASSERT(entry->mFileExists);
       buf.Append(entry->mFilename);
@@ -691,7 +745,7 @@ class FontNameCache {
       buf.Append(kFileSep);
     }
 
-    LOG(("putting FontNameCache to " CACHE_KEY ", length %u",
+    LOG(("putting FontNameCache to " CACHE_KEY ", length %zu",
          buf.Length() + 1));
     mCache->PutBuffer(CACHE_KEY, UniquePtr<char[]>(ToNewCString(buf)),
                       buf.Length() + 1);
@@ -807,14 +861,14 @@ class FontNameCache {
 
   PLDHashTableOps mOps;
 
-  typedef struct : public PLDHashEntryHdr {
+  struct FNCMapEntry : public PLDHashEntryHdr {
    public:
     nsCString mFilename;
     uint32_t mTimestamp;
     uint32_t mFilesize;
     nsCString mFaces;
     bool mFileExists;
-  } FNCMapEntry;
+  };
 
   static PLDHashNumber StringHash(const void* key) {
     return HashString(reinterpret_cast<const char*>(key));
@@ -892,6 +946,7 @@ gfxFT2FontList::gfxFT2FontList() : mJarModifiedTime(0) {
 }
 
 gfxFT2FontList::~gfxFT2FontList() {
+  AutoLock lock(mLock);
   if (mObserver) {
     mObserver->Remove();
   }
@@ -1041,7 +1096,7 @@ void FT2FontEntry::AppendToFaceList(nsCString& aFaceList,
 }
 
 void FT2FontEntry::CheckForBrokenFont(gfxFontFamily* aFamily) {
-  // note if the family is in the "bad underline" blacklist
+  // note if the family is in the "bad underline" blocklist
   if (aFamily->IsBadUnderlineFamily()) {
     mIsBadUnderlineFont = true;
   }
@@ -1116,7 +1171,9 @@ void gfxFT2FontList::AppendFacesFromFontFile(const nsCString& aFileName,
     CollectFunc unshared =
         [](const FontListEntry& aFLE, const nsCString& aPSName,
            const nsCString& aFullName, StandardFile aStdFile) {
-          PlatformFontList()->AppendFaceFromFontListEntry(aFLE, aStdFile);
+          auto* pfl = PlatformFontList();
+          pfl->mLock.AssertCurrentThreadIn();
+          pfl->AppendFaceFromFontListEntry(aFLE, aStdFile);
         };
     CollectFunc shared = [](const FontListEntry& aFLE, const nsCString& aPSName,
                             const nsCString& aFullName, StandardFile aStdFile) {
@@ -1244,17 +1301,17 @@ void gfxFT2FontList::AddFaceToList(const nsCString& aEntryName, uint32_t aIndex,
                         visibility);
       CollectInitData(fle, psname, fullname, aStdFile);
     } else {
-      RefPtr<gfxFontFamily> family = mFontFamilies.GetWeak(familyKey);
-      if (!family) {
-        family = new FT2FontFamily(familyName, visibility);
-        mFontFamilies.Put(familyKey, RefPtr{family});
-        if (mSkipSpaceLookupCheckFamilies.Contains(familyKey)) {
-          family->SetSkipSpaceFeatureCheck(true);
-        }
-        if (mBadUnderlineFamilyNames.ContainsSorted(familyKey)) {
-          family->SetBadUnderlineFamily();
-        }
-      }
+      RefPtr<gfxFontFamily> family =
+          mFontFamilies.LookupOrInsertWith(familyKey, [&] {
+            auto family = MakeRefPtr<FT2FontFamily>(familyName, visibility);
+            if (mSkipSpaceLookupCheckFamilies.Contains(familyKey)) {
+              family->SetSkipSpaceFeatureCheck(true);
+            }
+            if (mBadUnderlineFamilyNames.ContainsSorted(familyKey)) {
+              family->SetBadUnderlineFamily();
+            }
+            return family;
+          });
       family->AddFontEntry(fe);
       fe->CheckForBrokenFont(family);
     }
@@ -1287,7 +1344,9 @@ void gfxFT2FontList::AppendFacesFromOmnijarEntry(nsZipArchive* aArchive,
       CollectFunc unshared =
           [](const FontListEntry& aFLE, const nsCString& aPSName,
              const nsCString& aFullName, StandardFile aStdFile) {
-            PlatformFontList()->AppendFaceFromFontListEntry(aFLE, aStdFile);
+            auto* pfl = PlatformFontList();
+            pfl->mLock.AssertCurrentThreadIn();
+            pfl->AppendFaceFromFontListEntry(aFLE, aStdFile);
           };
       CollectFunc shared = [](const FontListEntry& aFLE,
                               const nsCString& aPSName,
@@ -1330,17 +1389,15 @@ void gfxFT2FontList::AppendFacesFromOmnijarEntry(nsZipArchive* aArchive,
 // Called on each family after all fonts are added to the list;
 // if aSortFaces is true this will sort faces to give priority to "standard"
 // font files.
-static void FinalizeFamilyMemberList(nsCStringHashKey::KeyType aKey,
-                                     RefPtr<gfxFontFamily>& aFamily,
-                                     bool aSortFaces) {
-  gfxFontFamily* family = aFamily.get();
+void FT2FontFamily::FinalizeMemberList(bool aSortFaces) {
+  AutoWriteLock lock(mLock);
 
-  family->SetHasStyles(true);
+  SetHasStyles(true);
 
   if (aSortFaces) {
-    family->SortAvailableFonts();
+    SortAvailableFonts();
   }
-  family->CheckForSimpleFamily();
+  CheckForSimpleFamily();
 }
 
 void gfxFT2FontList::FindFonts() {
@@ -1368,6 +1425,20 @@ void gfxFT2FontList::FindFonts() {
 
   static bool firstTime = true;
 
+  nsAutoCString androidFontsRoot = [&] {
+    // ANDROID_ROOT is the root of the android system, typically /system;
+    // font files are in /$ANDROID_ROOT/fonts/
+    nsAutoCString root;
+    char* androidRoot = PR_GetEnv("ANDROID_ROOT");
+    if (androidRoot) {
+      root = androidRoot;
+    } else {
+      root = "/system"_ns;
+    }
+    root.AppendLiteral("/fonts");
+    return root;
+  }();
+
   if (firstTime) {
     if (jni::GetAPIVersion() >= 29) {
       void* handle = dlopen("libandroid.so", RTLD_LAZY | RTLD_LOCAL);
@@ -1399,12 +1470,19 @@ void gfxFT2FontList::FindFonts() {
     if (iter) {
       void* font = systemFontIterator_next(iter);
       while (font) {
-        nsAutoCString path(font_getFontFilePath(font));
+        nsDependentCString path(font_getFontFilePath(font));
         AppendFacesFromFontFile(path, mFontNameCache.get(), kStandard);
         font_close(font);
         font = systemFontIterator_next(iter);
       }
 
+      // We don't yet support COLRv1 fonts (bug 1740525). Newer android versions
+      // have COLRv1 emoji font, and a legacy and hidden CBDT font we
+      // understand, so try to find NotoColorEmojiLegacy.ttf explicitly for now.
+      nsAutoCString legacyEmojiFont(androidFontsRoot);
+      legacyEmojiFont.Append("/NotoColorEmojiLegacy.ttf");
+      AppendFacesFromFontFile(legacyEmojiFont, mFontNameCache.get(), kStandard);
+
       systemFontIterator_close(iter);
     } else {
       useSystemFontAPI = false;
@@ -1414,18 +1492,7 @@ void gfxFT2FontList::FindFonts() {
   if (!useSystemFontAPI)
 #endif
   {
-    // ANDROID_ROOT is the root of the android system, typically /system;
-    // font files are in /$ANDROID_ROOT/fonts/
-    nsCString root;
-    char* androidRoot = PR_GetEnv("ANDROID_ROOT");
-    if (androidRoot) {
-      root = androidRoot;
-    } else {
-      root = NS_LITERAL_CSTRING("/system");
-    }
-    root.AppendLiteral("/fonts");
-
-    FindFontsInDir(root, mFontNameCache.get());
+    FindFontsInDir(androidFontsRoot, mFontNameCache.get());
   }
 
   // Look for fonts stored in omnijar, unless we're on a low-memory
@@ -1433,10 +1500,14 @@ void gfxFT2FontList::FindFonts() {
   // (Prefs may disable this, or force-enable it even with low memory.)
   bool lowmem;
   nsCOMPtr<nsIMemory> mem = nsMemory::GetGlobalMemoryService();
-  if ((NS_SUCCEEDED(mem->IsLowMemoryPlatform(&lowmem)) && !lowmem &&
-       Preferences::GetBool("gfx.bundled_fonts.enabled")) ||
-      Preferences::GetBool("gfx.bundled_fonts.force-enabled")) {
+  if (StaticPrefs::gfx_bundled_fonts_activate_AtStartup() > 0 ||
+      (StaticPrefs::gfx_bundled_fonts_activate_AtStartup() < 0 &&
+       NS_SUCCEEDED(mem->IsLowMemoryPlatform(&lowmem)) && !lowmem)) {
+    TimeStamp start = TimeStamp::Now();
     FindFontsInOmnijar(mFontNameCache.get());
+    TimeStamp end = TimeStamp::Now();
+    Telemetry::Accumulate(Telemetry::FONTLIST_BUNDLEDFONTS_ACTIVATE,
+                          (end - start).ToMilliseconds());
   }
 
   // Look for downloaded fonts in a profile-agnostic "fonts" directory.
@@ -1447,7 +1518,7 @@ void gfxFT2FontList::FindFonts() {
     nsresult rv = dirSvc->Get(NS_XPCOM_CURRENT_PROCESS_DIR, NS_GET_IID(nsIFile),
                               getter_AddRefs(appDir));
     if (NS_SUCCEEDED(rv)) {
-      appDir->AppendNative(NS_LITERAL_CSTRING("fonts"));
+      appDir->AppendNative("fonts"_ns);
       nsCString localPath;
       if (NS_SUCCEEDED(appDir->GetNativePath(localPath))) {
         FindFontsInDir(localPath, mFontNameCache.get());
@@ -1459,8 +1530,7 @@ void gfxFT2FontList::FindFonts() {
   nsCOMPtr<nsIFile> localDir;
   nsresult rv = NS_GetSpecialDirectory(NS_APP_USER_PROFILE_LOCAL_50_DIR,
                                        getter_AddRefs(localDir));
-  if (NS_SUCCEEDED(rv) &&
-      NS_SUCCEEDED(localDir->Append(NS_LITERAL_STRING("fonts")))) {
+  if (NS_SUCCEEDED(rv) && NS_SUCCEEDED(localDir->Append(u"fonts"_ns))) {
     nsCString localPath;
     rv = localDir->GetNativePath(localPath);
     if (NS_SUCCEEDED(rv)) {
@@ -1510,6 +1580,9 @@ void gfxFT2FontList::FindFontsInDir(const nsCString& aDir,
                                          "MTLmr3m.ttf",
                                          "MTLc3m.ttf",
                                          "NanumGothic.ttf",
+                                         "NotoColorEmoji.ttf",
+                                         "NotoColorEmojiFlags.ttf",
+                                         "NotoColorEmojiLegacy.ttf",
                                          "DroidSansJapanese.ttf",
                                          "DroidSansFallback.ttf"};
 
@@ -1532,7 +1605,7 @@ void gfxFT2FontList::FindFontsInDir(const nsCString& aDir,
         isStdFont = strcmp(sStandardFonts[i], ent->d_name) == 0;
       }
 
-      nsCString s(aDir);
+      nsAutoCString s(aDir);
       s.Append('/');
       s.Append(ent->d_name);
 
@@ -1554,39 +1627,40 @@ void gfxFT2FontList::AppendFaceFromFontListEntry(const FontListEntry& aFLE,
     nsAutoCString key(aFLE.familyName());
     BuildKeyNameFromFontName(key);
     fe->mStandardFace = (aStdFile == kStandard);
-    RefPtr<gfxFontFamily> family = mFontFamilies.GetWeak(key);
-    if (!family) {
-      family = new FT2FontFamily(aFLE.familyName(), aFLE.visibility());
-      mFontFamilies.Put(key, RefPtr{family});
+    RefPtr<gfxFontFamily> family = mFontFamilies.LookupOrInsertWith(key, [&] {
+      auto family =
+          MakeRefPtr<FT2FontFamily>(aFLE.familyName(), aFLE.visibility());
       if (mSkipSpaceLookupCheckFamilies.Contains(key)) {
         family->SetSkipSpaceFeatureCheck(true);
       }
       if (mBadUnderlineFamilyNames.ContainsSorted(key)) {
         family->SetBadUnderlineFamily();
       }
-    }
+      return family;
+    });
     family->AddFontEntry(fe);
 
     fe->CheckForBrokenFont(family);
   }
 }
 
-void gfxFT2FontList::ReadSystemFontList(nsTArray<FontListEntry>* aList) {
-  for (auto iter = mFontFamilies.Iter(); !iter.Done(); iter.Next()) {
-    auto family = static_cast<FT2FontFamily*>(iter.Data().get());
-    family->AddFacesToFontList(aList);
+void gfxFT2FontList::ReadSystemFontList(dom::SystemFontList* aList) {
+  AutoLock lock(mLock);
+  for (const auto& entry : mFontFamilies) {
+    auto family = static_cast<FT2FontFamily*>(entry.GetData().get());
+    family->AddFacesToFontList(&aList->entries());
   }
 }
 
 static void LoadSkipSpaceLookupCheck(
-    nsTHashtable<nsCStringHashKey>& aSkipSpaceLookupCheck) {
+    nsTHashSet<nsCString>& aSkipSpaceLookupCheck) {
   AutoTArray<nsCString, 5> skiplist;
   gfxFontUtils::GetPrefsFontList(
       "font.whitelist.skip_default_features_space_check", skiplist);
   uint32_t numFonts = skiplist.Length();
   for (uint32_t i = 0; i < numFonts; i++) {
     ToLowerCase(skiplist[i]);
-    aSkipSpaceLookupCheck.PutEntry(skiplist[i]);
+    aSkipSpaceLookupCheck.Insert(skiplist[i]);
   }
 }
 
@@ -1600,10 +1674,9 @@ nsresult gfxFT2FontList::InitFontListForPlatform() {
 
     // Finalize the families by sorting faces into standard order
     // and marking "simple" families.
-    for (auto iter = mFontFamilies.Iter(); !iter.Done(); iter.Next()) {
-      nsCStringHashKey::KeyType key = iter.Key();
-      RefPtr<gfxFontFamily>& family = iter.Data();
-      FinalizeFamilyMemberList(key, family, /* aSortFaces */ true);
+    for (const auto& entry : mFontFamilies) {
+      auto* family = static_cast<FT2FontFamily*>(entry.GetData().get());
+      family->FinalizeMemberList(/* aSortFaces */ true);
     }
 
     return NS_OK;
@@ -1612,7 +1685,7 @@ nsresult gfxFT2FontList::InitFontListForPlatform() {
   // Content process: use font list passed from the chrome process via
   // the GetXPCOMProcessAttributes message.
   auto& fontList = dom::ContentChild::GetSingleton()->SystemFontList();
-  for (FontListEntry& fle : fontList) {
+  for (FontListEntry& fle : fontList.entries()) {
     // We don't need to identify "standard" font files here,
     // as the faces are already sorted.
     AppendFaceFromFontListEntry(fle, kUnknown);
@@ -1620,16 +1693,15 @@ nsresult gfxFT2FontList::InitFontListForPlatform() {
 
   // We don't need to sort faces (because they were already sorted by the
   // chrome process, so we just maintain the existing order)
-  for (auto iter = mFontFamilies.Iter(); !iter.Done(); iter.Next()) {
-    nsCStringHashKey::KeyType key = iter.Key();
-    RefPtr<gfxFontFamily>& family = iter.Data();
-    FinalizeFamilyMemberList(key, family, /* aSortFaces */ false);
+  for (const auto& entry : mFontFamilies) {
+    auto* family = static_cast<FT2FontFamily*>(entry.GetData().get());
+    family->FinalizeMemberList(/* aSortFaces */ false);
   }
 
   LOG(("got font list from chrome process: %" PRIdPTR " faces in %" PRIu32
        " families",
-       fontList.Length(), mFontFamilies.Count()));
-  fontList.Clear();
+       fontList.entries().Length(), mFontFamilies.Count()));
+  fontList.entries().Clear();
 
   return NS_OK;
 }
@@ -1646,9 +1718,6 @@ void gfxFT2FontList::InitSharedFontListForPlatform() {
   // mFaceInitData (shared font list).
   FindFonts();
 
-  ApplyWhitelist(mFamilyInitData);
-  mFamilyInitData.Sort();
-
   mozilla::fontlist::FontList* list = SharedFontList();
   list->SetFamilyNames(mFamilyInitData);
 
@@ -1667,40 +1736,40 @@ gfxFontEntry* gfxFT2FontList::CreateFontEntry(fontlist::Face* aFace,
                                               const fontlist::Family* aFamily) {
   fontlist::FontList* list = SharedFontList();
   nsAutoCString desc(aFace->mDescriptor.AsString(list));
-  FontListEntry fle(aFamily->DisplayName().AsString(list), desc, desc,
-                    aFace->mWeight.AsScalar(), aFace->mStretch.AsScalar(),
-                    aFace->mStyle.AsScalar(), aFace->mIndex,
-                    aFamily->Visibility());
-  FT2FontEntry* fe = FT2FontEntry::CreateFontEntry(fle);
-
-  fe->mFixedPitch = aFace->mFixedPitch;
-  fe->mIsBadUnderlineFont = aFamily->IsBadUnderlineFamily();
-  fe->mShmemFace = aFace;
-  fe->mFamilyName = aFamily->DisplayName().AsString(list);
-
+  FT2FontEntry* fe =
+      FT2FontEntry::CreateFontEntry(desc, desc.get(), aFace->mIndex, nullptr);
+  fe->InitializeFrom(aFace, aFamily);
   fe->CheckForBrokenFont(aFamily->Key().AsString(list));
-
   return fe;
 }
 
 // called for each family name, based on the assumption that the
 // first part of the full name is the family name
 
-gfxFontEntry* gfxFT2FontList::LookupLocalFont(const nsACString& aFontName,
+gfxFontEntry* gfxFT2FontList::LookupLocalFont(nsPresContext* aPresContext,
+                                              const nsACString& aFontName,
                                               WeightRange aWeightForEntry,
                                               StretchRange aStretchForEntry,
                                               SlantStyleRange aStyleForEntry) {
+  AutoLock lock(mLock);
+
   if (SharedFontList()) {
-    return LookupInSharedFaceNameList(aFontName, aWeightForEntry,
+    return LookupInSharedFaceNameList(aPresContext, aFontName, aWeightForEntry,
                                       aStretchForEntry, aStyleForEntry);
   }
+
   // walk over list of names
   FT2FontEntry* fontEntry = nullptr;
+  FontVisibility level =
+      aPresContext ? aPresContext->GetFontVisibility() : FontVisibility::User;
+
+  for (const RefPtr<gfxFontFamily>& fontFamily : mFontFamilies.Values()) {
+    if (!IsVisibleToCSS(*fontFamily, level)) {
+      continue;
+    }
 
-  for (auto iter = mFontFamilies.Iter(); !iter.Done(); iter.Next()) {
     // Check family name, based on the assumption that the
     // first part of the full name is the family name
-    RefPtr<gfxFontFamily>& fontFamily = iter.Data();
 
     // does the family name match up to the length of the family name?
     const nsCString& family = fontFamily->Name();
@@ -1710,17 +1779,11 @@ gfxFontEntry* gfxFT2FontList::LookupLocalFont(const nsACString& aFontName,
 
     // if so, iterate over faces in this family to see if there is a match
     if (family.Equals(fullNameFamily, nsCaseInsensitiveCStringComparator)) {
-      nsTArray<RefPtr<gfxFontEntry> >& fontList = fontFamily->GetFontList();
-      int index, len = fontList.Length();
-      for (index = 0; index < len; index++) {
-        gfxFontEntry* fe = fontList[index];
-        if (!fe) {
-          continue;
-        }
-        if (fe->Name().Equals(aFontName, nsCaseInsensitiveCStringComparator)) {
-          fontEntry = static_cast<FT2FontEntry*>(fe);
-          goto searchDone;
-        }
+      gfxFontEntry* fe =
+          fontFamily->FindFont(aFontName, nsCaseInsensitiveCStringComparator);
+      if (fe) {
+        fontEntry = static_cast<FT2FontEntry*>(fe);
+        goto searchDone;
       }
     }
   }
@@ -1753,12 +1816,13 @@ searchDone:
 }
 
 FontFamily gfxFT2FontList::GetDefaultFontForPlatform(
-    const gfxFontStyle* aStyle) {
+    nsPresContext* aPresContext, const gfxFontStyle* aStyle,
+    nsAtom* aLanguage) {
   FontFamily ff;
 #if defined(MOZ_WIDGET_ANDROID)
-  ff = FindFamily(NS_LITERAL_CSTRING("Roboto"));
+  ff = FindFamily(aPresContext, "Roboto"_ns);
   if (ff.IsNull()) {
-    ff = FindFamily(NS_LITERAL_CSTRING("Droid Sans"));
+    ff = FindFamily(aPresContext, "Droid Sans"_ns);
   }
 #endif
   /* TODO: what about Qt or other platforms that may use this? */
