@@ -6,52 +6,65 @@
 #ifndef GFX_FONTENTRY_H
 #define GFX_FONTENTRY_H
 
-#include "gfxTypes.h"
-#include "nsString.h"
-#include "gfxFontConstants.h"
-#include "gfxFontFeatures.h"
+#include <math.h>
+#include <new>
+#include <utility>
+#include "ThebesRLBoxTypes.h"
 #include "gfxFontUtils.h"
 #include "gfxFontVariations.h"
-#include "nsTArray.h"
-#include "nsTHashtable.h"
-#include "mozilla/HashFunctions.h"
-#include "mozilla/MemoryReporting.h"
-#include "MainThreadUtils.h"
-#include "nsUnicodeScriptCodes.h"
-#include "nsDataHashtable.h"
+#include "gfxRect.h"
+#include "gfxTypes.h"
 #include "harfbuzz/hb.h"
+#include "mozilla/AlreadyAddRefed.h"
+#include "mozilla/Assertions.h"
 #include "mozilla/FontPropertyTypes.h"
-#include "mozilla/gfx/2D.h"
+#include "mozilla/MemoryReporting.h"
+#include "mozilla/Mutex.h"
+#include "mozilla/RefPtr.h"
+#include "mozilla/RWLock.h"
+#include "mozilla/TypedEnumBits.h"
 #include "mozilla/UniquePtr.h"
-#include "mozilla/WeakPtr.h"
-#include "ThebesRLBoxTypes.h"
-#include <math.h>
+#include "mozilla/intl/UnicodeScriptCodes.h"
+#include "nsTHashMap.h"
+#include "nsDebug.h"
+#include "nsHashKeys.h"
+#include "nsISupports.h"
+#include "nsStringFwd.h"
+#include "nsTArray.h"
+#include "nscore.h"
 
-typedef struct gr_face gr_face;
-typedef struct FT_MM_Var_ FT_MM_Var;
-
-#ifdef DEBUG
-#  include <stdio.h>
-#endif
-
-struct gfxFontStyle;
+class FontInfoData;
 class gfxContext;
 class gfxFont;
 class gfxFontFamily;
-class gfxUserFontData;
+class gfxPlatformFontList;
 class gfxSVGGlyphs;
-class FontInfoData;
-struct FontListSizes;
+class gfxUserFontData;
 class nsAtom;
+struct FontListSizes;
+struct gfxFontFeature;
+struct gfxFontStyle;
+enum class eFontPresentation : uint8_t;
+
+namespace IPC {
+template <class P>
+struct ParamTraits;
+}
 
 namespace mozilla {
 class SVGContextPaint;
 namespace fontlist {
-struct Family;
 struct Face;
+struct Family;
 }  // namespace fontlist
+namespace gfx {
+struct DeviceColor;
+}
 }  // namespace mozilla
 
+typedef struct gr_face gr_face;
+typedef struct FT_MM_Var_ FT_MM_Var;
+
 #define NO_FONT_LANGUAGE_OVERRIDE 0
 
 class gfxCharacterMap : public gfxSparseBitSet {
@@ -119,7 +132,7 @@ struct gfxFontFeatureInfo {
 class gfxFontEntry {
  public:
   typedef mozilla::gfx::DrawTarget DrawTarget;
-  typedef mozilla::unicode::Script Script;
+  typedef mozilla::intl::Script Script;
   typedef mozilla::FontWeight FontWeight;
   typedef mozilla::FontSlantStyle FontSlantStyle;
   typedef mozilla::FontStretch FontStretch;
@@ -132,6 +145,10 @@ class gfxFontEntry {
 
   explicit gfxFontEntry(const nsACString& aName, bool aIsStandardFace = false);
 
+  gfxFontEntry() = delete;
+  gfxFontEntry(const gfxFontEntry&) = delete;
+  gfxFontEntry& operator=(const gfxFontEntry&) = delete;
+
   // Create a new entry that refers to the same font as this, but without
   // additional state that may have been set up (such as family name).
   // (This is only to be used for system fonts in the platform font list,
@@ -212,15 +229,16 @@ class gfxFontEntry {
 
   inline bool HasCharacter(uint32_t ch) {
     if (mShmemCharacterMap) {
-      return mShmemCharacterMap->test(ch);
+      return GetShmemCharacterMap()->test(ch);
     }
     if (mCharacterMap) {
       if (mShmemFace && TrySetShmemCharacterMap()) {
         // Forget our temporary local copy, now we can use the shared cmap
-        mCharacterMap = nullptr;
-        return mShmemCharacterMap->test(ch);
+        auto* oldCmap = mCharacterMap.exchange(nullptr);
+        NS_IF_RELEASE(oldCmap);
+        return GetShmemCharacterMap()->test(ch);
       }
-      if (mCharacterMap->test(ch)) {
+      if (GetCharacterMap()->test(ch)) {
         return true;
       }
     }
@@ -228,7 +246,7 @@ class gfxFontEntry {
   }
 
   virtual bool SkipDuringSystemFallback() { return false; }
-  nsresult InitializeUVSMap();
+  void EnsureUVSMapInitialized();
   uint16_t GetUVSGlyph(uint32_t aCh, uint32_t aVS);
 
   // All concrete gfxFontEntry subclasses (except gfxUserFontEntry) need
@@ -239,7 +257,7 @@ class gfxFontEntry {
   // can be safely dereferenced.
   virtual nsresult ReadCMAP(FontInfoData* aFontInfoData = nullptr);
 
-  bool TryGetSVGData(gfxFont* aFont);
+  bool TryGetSVGData(const gfxFont* aFont);
   bool HasSVGGlyph(uint32_t aGlyphId);
   bool GetSVGGlyphExtents(DrawTarget* aDrawTarget, uint32_t aGlyphId,
                           gfxFloat aSize, gfxRect* aResult);
@@ -254,6 +272,19 @@ class gfxFontEntry {
                           const mozilla::gfx::DeviceColor& aDefaultColor,
                           nsTArray<uint16_t>& layerGlyphs,
                           nsTArray<mozilla::gfx::DeviceColor>& layerColors);
+  bool HasColorLayersForGlyph(uint32_t aGlyphId) {
+    MOZ_ASSERT(GetCOLR());
+    return gfxFontUtils::HasColorLayersForGlyph(GetCOLR(), aGlyphId);
+  }
+
+  bool HasColorBitmapTable() {
+    if (!mCheckedForColorBitmapTables) {
+      mHasColorBitmapTable = HasFontTable(TRUETYPE_TAG('C', 'B', 'D', 'T')) ||
+                             HasFontTable(TRUETYPE_TAG('s', 'b', 'i', 'x'));
+      mCheckedForColorBitmapTables = true;
+    }
+    return mHasColorBitmapTable;
+  }
 
   // Access to raw font table data (needed for Harfbuzz):
   // returns a pointer to data owned by the fontEntry or the OS,
@@ -406,6 +437,8 @@ class gfxFontEntry {
 
   bool HasBoldVariableWeight();
   bool HasItalicVariation();
+  bool HasOpticalSize();
+
   void CheckForVariationAxes();
 
   // Set up the entry's weight/stretch/style ranges according to axes found
@@ -436,25 +469,44 @@ class gfxFontEntry {
   nsCString mName;
   nsCString mFamilyName;
 
-  RefPtr<gfxCharacterMap> mCharacterMap;
+  // These are mutable so that we can take a read lock within a const method.
+  mutable mozilla::RWLock mLock;
+  mutable mozilla::Mutex mFeatureInfoLock;
+
+  mozilla::Atomic<gfxCharacterMap*> mCharacterMap;  // strong ref
+  gfxCharacterMap* GetCharacterMap() const { return mCharacterMap; }
 
   mozilla::fontlist::Face* mShmemFace = nullptr;
-  const SharedBitSet* mShmemCharacterMap = nullptr;
 
-  mozilla::UniquePtr<uint8_t[]> mUVSData;
+  mozilla::Atomic<const SharedBitSet*> mShmemCharacterMap;
+  const SharedBitSet* GetShmemCharacterMap() const {
+    return mShmemCharacterMap;
+  }
+
+  mozilla::Atomic<const uint8_t*> mUVSData;
+  const uint8_t* GetUVSData() const { return mUVSData; }
+
   mozilla::UniquePtr<gfxUserFontData> mUserFontData;
-  mozilla::UniquePtr<gfxSVGGlyphs> mSVGGlyphs;
+
+  mozilla::Atomic<gfxSVGGlyphs*> mSVGGlyphs;
+  gfxSVGGlyphs* GetSVGGlyphs() const { return mSVGGlyphs; }
+
   // list of gfxFonts that are using SVG glyphs
-  nsTArray<gfxFont*> mFontsUsingSVGGlyphs;
+  nsTArray<const gfxFont*> mFontsUsingSVGGlyphs GUARDED_BY(mLock);
   nsTArray<gfxFontFeature> mFeatureSettings;
   nsTArray<gfxFontVariation> mVariationSettings;
-  mozilla::UniquePtr<nsDataHashtable<nsUint32HashKey, bool>> mSupportedFeatures;
-  mozilla::UniquePtr<nsDataHashtable<nsUint32HashKey, hb_set_t*>>
-      mFeatureInputs;
 
-  // Color Layer font support
-  hb_blob_t* mCOLR = nullptr;
-  hb_blob_t* mCPAL = nullptr;
+  mozilla::UniquePtr<nsTHashMap<nsUint32HashKey, bool>> mSupportedFeatures
+      GUARDED_BY(mFeatureInfoLock);
+  mozilla::UniquePtr<nsTHashMap<nsUint32HashKey, hb_set_t*>> mFeatureInputs
+      GUARDED_BY(mFeatureInfoLock);
+
+  // Color Layer font support. These tables are inert once loaded, so we don't
+  // need to hold a lock when reading them.
+  mozilla::Atomic<hb_blob_t*> mCOLR;
+  mozilla::Atomic<hb_blob_t*> mCPAL;
+  hb_blob_t* GetCOLR() const { return mCOLR; }
+  hb_blob_t* GetCPAL() const { return mCPAL; }
 
   // bitvector of substitution space features per script, one each
   // for default and non-default features
@@ -462,7 +514,7 @@ class gfxFontEntry {
   uint32_t
       mNonDefaultSubSpaceFeatures[(int(Script::NUM_SCRIPT_CODES) + 31) / 32];
 
-  uint32_t mUVSOffset = 0;
+  mozilla::Atomic<uint32_t> mUVSOffset;
 
   uint32_t mLanguageOverride = NO_FONT_LANGUAGE_OVERRIDE;
 
@@ -470,6 +522,15 @@ class gfxFontEntry {
   StretchRange mStretchRange = StretchRange(FontStretch::Normal());
   SlantStyleRange mStyleRange = SlantStyleRange(FontSlantStyle::Normal());
 
+  // Font metrics overrides (as multiples of used font size); negative values
+  // indicate no override to be applied.
+  float mAscentOverride = -1.0;
+  float mDescentOverride = -1.0;
+  float mLineGapOverride = -1.0;
+
+  // Scaling factor to be applied to the font size.
+  float mSizeAdjust = 1.0;
+
   // For user fonts (only), we need to record whether or not weight/stretch/
   // slant variations should be clamped to the range specified in the entry
   // properties. When the @font-face rule omitted one or more of these
@@ -494,13 +555,13 @@ class gfxFontEntry {
     // properties to the variation axes (though they can still be
     // explicitly set using font-variation-settings).
     eNonCSSWeight = (1 << 5),
-    eNonCSSStretch = (1 << 6)
+    eNonCSSStretch = (1 << 6),
+
+    // Whether the font has an 'opsz' axis.
+    eOpticalSize = (1 << 7)
   };
   RangeFlags mRangeFlags = RangeFlags::eNoFlags;
 
-  // NOTE that there are currently exactly 24 one-bit flags defined here,
-  // so together with the 8-bit RangeFlags above, this packs neatly to a
-  // 32-bit boundary. Worth considering if further flags are wanted.
   bool mFixedPitch : 1;
   bool mIsBadUnderlineFont : 1;
   bool mIsUserFontContainer : 1;  // userfont entry
@@ -509,33 +570,31 @@ class gfxFontEntry {
   bool mStandardFace : 1;
   bool mIgnoreGDEF : 1;
   bool mIgnoreGSUB : 1;
-  bool mSVGInitialized : 1;
-  bool mHasSpaceFeaturesInitialized : 1;
-  bool mHasSpaceFeatures : 1;
-  bool mHasSpaceFeaturesKerning : 1;
-  bool mHasSpaceFeaturesNonKerning : 1;
   bool mSkipDefaultFeatureSpaceCheck : 1;
-  bool mGraphiteSpaceContextualsInitialized : 1;
-  bool mHasGraphiteSpaceContextuals : 1;
-  bool mSpaceGlyphIsInvisible : 1;
-  bool mSpaceGlyphIsInvisibleInitialized : 1;
-  bool mHasGraphiteTables : 1;
-  bool mCheckedForGraphiteTables : 1;
-  bool mHasCmapTable : 1;
-  bool mGrFaceInitialized : 1;
-  bool mCheckedForColorGlyph : 1;
-  bool mCheckedForVariationAxes : 1;
+
+  mozilla::Atomic<bool> mSVGInitialized;
+  mozilla::Atomic<bool> mHasSpaceFeaturesInitialized;
+  mozilla::Atomic<bool> mHasSpaceFeatures;
+  mozilla::Atomic<bool> mHasSpaceFeaturesKerning;
+  mozilla::Atomic<bool> mHasSpaceFeaturesNonKerning;
+  mozilla::Atomic<bool> mGraphiteSpaceContextualsInitialized;
+  mozilla::Atomic<bool> mHasGraphiteSpaceContextuals;
+  mozilla::Atomic<bool> mSpaceGlyphIsInvisible;
+  mozilla::Atomic<bool> mSpaceGlyphIsInvisibleInitialized;
+  mozilla::Atomic<bool> mHasGraphiteTables;
+  mozilla::Atomic<bool> mCheckedForGraphiteTables;
+  mozilla::Atomic<bool> mHasCmapTable;
+  mozilla::Atomic<bool> mGrFaceInitialized;
+  mozilla::Atomic<bool> mCheckedForColorGlyph;
+  mozilla::Atomic<bool> mCheckedForVariationAxes;
+  mozilla::Atomic<bool> mHasColorBitmapTable;
+  mozilla::Atomic<bool> mCheckedForColorBitmapTables;
 
  protected:
   friend class gfxPlatformFontList;
-  friend class gfxMacPlatformFontList;
-  friend class gfxUserFcFontEntry;
   friend class gfxFontFamily;
-  friend class gfxSingleFaceMacFontFamily;
   friend class gfxUserFontEntry;
 
-  gfxFontEntry();
-
   // Protected destructor, to discourage deletion outside of Release():
   virtual ~gfxFontEntry();
 
@@ -556,7 +615,7 @@ class gfxFontEntry {
   // Helper for HasTrackingTable; check/parse the table and cache pointers
   // to the subtables we need. Returns false on failure, in which case the
   // table is unusable.
-  bool ParseTrakTable();
+  bool ParseTrakTable() REQUIRES(mLock);
 
   // lookup the cmap in cached font data
   virtual already_AddRefed<gfxCharacterMap> GetCMAPFromFontInfo(
@@ -570,6 +629,11 @@ class gfxFontEntry {
   // hasn't handled our SetCharacterMap message yet).
   bool TrySetShmemCharacterMap();
 
+  // Helper for gfxPlatformFontList::CreateFontEntry methods: set properties
+  // of the gfxFontEntry based on shared Face and Family records.
+  void InitializeFrom(mozilla::fontlist::Face* aFace,
+                      const mozilla::fontlist::Family* aFamily);
+
   // Shaper-specific face objects, shared by all instantiations of the same
   // physical font, regardless of size.
   // Usually, only one of these will actually be created for any given font
@@ -580,7 +644,7 @@ class gfxFontEntry {
   // in its destructor. The font entry has only this non-owning reference to
   // the face; when the face is deleted, it will tell the font entry to forget
   // it, so that a new face will be created next time it is needed.
-  hb_face_t* mHBFace = nullptr;
+  mozilla::Atomic<hb_face_t*> mHBFace;
 
   static hb_blob_t* HBGetTable(hb_face_t* face, uint32_t aTag, void* aUserData);
 
@@ -589,6 +653,7 @@ class gfxFontEntry {
 
   // All libGraphite functionality is sandboxed in an rlbox sandbox. This
   // contains data for the sandbox instance.
+  // Currently graphite shaping is only supported on the main thread.
   struct GrSandboxData;
   GrSandboxData* mSandboxData = nullptr;
 
@@ -600,21 +665,24 @@ class gfxFontEntry {
 
   // For AAT font, a strong reference to the 'trak' table (if present).
   hb_blob_t* const kTrakTableUninitialized = (hb_blob_t*)(intptr_t(-1));
-  hb_blob_t* mTrakTable = kTrakTableUninitialized;
+  mozilla::Atomic<hb_blob_t*> mTrakTable;
+  hb_blob_t* GetTrakTable() const { return mTrakTable; }
   bool TrakTableInitialized() const {
     return mTrakTable != kTrakTableUninitialized;
   }
 
   // Cached pointers to tables within 'trak', initialized by ParseTrakTable.
-  const mozilla::AutoSwap_PRInt16* mTrakValues;
-  const mozilla::AutoSwap_PRInt32* mTrakSizeTable;
+  // This data is inert once loaded, so locking is not required to read it.
+  const mozilla::AutoSwap_PRInt16* mTrakValues = nullptr;
+  const mozilla::AutoSwap_PRInt32* mTrakSizeTable = nullptr;
 
   // number of current users of this entry's mGrFace
   nsrefcnt mGrFaceRefCnt = 0;
 
   static tainted_opaque_gr<const void*> GrGetTable(
       rlbox_sandbox_gr& sandbox, tainted_opaque_gr<const void*> aAppFaceHandle,
-      tainted_opaque_gr<unsigned int> aName, tainted_opaque_gr<size_t*> aLen);
+      tainted_opaque_gr<unsigned int> aName,
+      tainted_opaque_gr<unsigned int*> aLen);
   static void GrReleaseTable(rlbox_sandbox_gr& sandbox,
                              tainted_opaque_gr<const void*> aAppFaceHandle,
                              tainted_opaque_gr<const void*> aTableBuffer);
@@ -629,7 +697,7 @@ class gfxFontEntry {
   // kInvalidUPEM for non-sfnt font formats)
   uint16_t mUnitsPerEm = 0;
 
-  uint16_t mNumTrakSizes;
+  uint16_t mNumTrakSizes = 0;
 
  private:
   /**
@@ -720,10 +788,9 @@ class gfxFontEntry {
     hb_blob_t* mBlob;
   };
 
-  mozilla::UniquePtr<nsTHashtable<FontTableHashEntry>> mFontTableCache;
-
-  gfxFontEntry(const gfxFontEntry&);
-  gfxFontEntry& operator=(const gfxFontEntry&);
+  using FontTableCache = nsTHashtable<FontTableHashEntry>;
+  mozilla::Atomic<FontTableCache*> mFontTableCache;
+  FontTableCache* GetFontTableCache() const { return mFontTableCache; }
 };
 
 MOZ_MAKE_ENUM_CLASS_BITWISE_OPERATORS(gfxFontEntry::RangeFlags)
@@ -747,40 +814,25 @@ inline bool gfxFontEntry::SupportsBold() {
 
 // used when iterating over all fonts looking for a match for a given character
 struct GlobalFontMatch {
-  GlobalFontMatch(const uint32_t aCharacter, const gfxFontStyle& aStyle)
-      : mStyle(aStyle), mCh(aCharacter) {}
+  GlobalFontMatch(uint32_t aCharacter, uint32_t aNextCh,
+                  const gfxFontStyle& aStyle, eFontPresentation aPresentation)
+      : mStyle(aStyle),
+        mCh(aCharacter),
+        mNextCh(aNextCh),
+        mPresentation(aPresentation) {}
 
   RefPtr<gfxFontEntry> mBestMatch;       // current best match
   RefPtr<gfxFontFamily> mMatchedFamily;  // the family it belongs to
   mozilla::fontlist::Family* mMatchedSharedFamily = nullptr;
-  const gfxFontStyle& mStyle;       // style to match
-  const uint32_t mCh;               // codepoint to be matched
-  uint32_t mCount = 0;              // number of fonts matched
-  uint32_t mCmapsTested = 0;        // number of cmaps tested
-  float mMatchDistance = INFINITY;  // metric indicating closest match
+  const gfxFontStyle& mStyle;  // style to match
+  const uint32_t mCh;          // codepoint to be matched
+  const uint32_t mNextCh;      // following codepoint (or zero)
+  eFontPresentation mPresentation;
+  uint32_t mCount = 0;               // number of fonts matched
+  uint32_t mCmapsTested = 0;         // number of cmaps tested
+  double mMatchDistance = INFINITY;  // metric indicating closest match
 };
 
-// Installation status (base system / langpack / user-installed) may determine
-// whether the font is visible to CSS font-family or src:local() lookups.
-// (Exactly what these mean and how accurate they are may be vary across
-// platforms -- e.g. on Linux there is no clear "base" set of fonts.)
-enum class FontVisibility : uint8_t {
-  Unknown = 0,   // No categorization of families available on this system
-  Base = 1,      // Standard part of the base OS installation
-  LangPack = 2,  // From an optional OS component such as language support
-  User = 3,      // User-installed font (or installed by another app, etc)
-  Hidden = 4,    // Internal system font, should never exposed to users
-  Webfont = 5,   // Webfont defined by @font-face
-  Count = 6,     // Count of values, for IPC serialization
-};
-
-namespace IPC {
-template <>
-struct ParamTraits<FontVisibility>
-    : public ContiguousEnumSerializer<FontVisibility, FontVisibility::Unknown,
-                                      FontVisibility::Count> {};
-}  // namespace IPC
-
 class gfxFontFamily {
  public:
   // Used by stylo
@@ -788,19 +840,14 @@ class gfxFontFamily {
 
   gfxFontFamily(const nsACString& aName, FontVisibility aVisibility)
       : mName(aName),
+        mLock("gfxFontFamily lock"),
         mVisibility(aVisibility),
-        mOtherFamilyNamesInitialized(false),
-        mHasOtherFamilyNames(false),
-        mFaceNamesInitialized(false),
-        mHasStyles(false),
         mIsSimpleFamily(false),
         mIsBadUnderlineFamily(false),
-        mFamilyCharacterMapInitialized(false),
         mSkipDefaultFeatureSpaceCheck(false),
-        mCheckForFallbackFaces(false),
-        mCheckedForLegacyFamilyNames(false) {}
+        mCheckForFallbackFaces(false) {}
 
-  const nsCString& Name() { return mName; }
+  const nsCString& Name() const { return mName; }
 
   virtual void LocalizedName(nsACString& aLocalizedName);
   virtual bool HasOtherFamilyNames();
@@ -811,11 +858,31 @@ class gfxFontFamily {
   // faces in a large family into separate "styled families" because of
   // GDI's 4-faces-per-family limitation). If found, the styled family
   // name will be added to the font list's "other family names" table.
+  // Note that the caller must already hold the gfxPlatformFontList lock.
   bool CheckForLegacyFamilyNames(gfxPlatformFontList* aFontList);
 
-  nsTArray<RefPtr<gfxFontEntry>>& GetFontList() { return mAvailableFonts; }
+  // Callers must hold a read-lock for as long as they're using the list.
+  const nsTArray<RefPtr<gfxFontEntry>>& GetFontList() REQUIRES_SHARED(mLock) {
+    return mAvailableFonts;
+  }
+  void ReadLock() ACQUIRE_SHARED(mLock) { mLock.ReadLock(); }
+  void ReadUnlock() RELEASE_SHARED(mLock) { mLock.ReadUnlock(); }
+
+  uint32_t FontListLength() const {
+    mozilla::AutoReadLock lock(mLock);
+    return mAvailableFonts.Length();
+  }
 
   void AddFontEntry(RefPtr<gfxFontEntry> aFontEntry) {
+    mozilla::AutoWriteLock lock(mLock);
+    AddFontEntryLocked(aFontEntry);
+  }
+
+  void AddFontEntryLocked(RefPtr<gfxFontEntry> aFontEntry) REQUIRES(mLock) {
+    // Avoid potentially duplicating entries.
+    if (mAvailableFonts.Contains(aFontEntry)) {
+      return;
+    }
     // bug 589682 - set the IgnoreGDEF flag on entries for Italic faces
     // of Times New Roman, because of buggy table in those fonts
     if (aFontEntry->IsItalic() && !aFontEntry->IsUserFont() &&
@@ -834,19 +901,19 @@ class gfxFontFamily {
     // we need to ensure any null entries are removed, as well as clearing
     // the flag (which may be set again later).
     if (mIsSimpleFamily) {
-      for (size_t i = mAvailableFonts.Length() - 1; i-- > 0;) {
-        if (!mAvailableFonts[i]) {
-          mAvailableFonts.RemoveElementAt(i);
-        }
-      }
+      mAvailableFonts.RemoveElementsBy([](const auto& font) { return !font; });
       mIsSimpleFamily = false;
     }
   }
 
   // note that the styles for this family have been added
-  bool HasStyles() { return mHasStyles; }
+  bool HasStyles() const { return mHasStyles; }
   void SetHasStyles(bool aHasStyles) { mHasStyles = aHasStyles; }
 
+  void SetCheckedForLegacyFamilyNames(bool aChecked) {
+    mCheckedForLegacyFamilyNames = aChecked;
+  }
+
   // choose a specific face to match a style using CSS font matching
   // rules (weight matching occurs here).  may return a face that doesn't
   // precisely match (e.g. normal face when no italic face exists).
@@ -857,8 +924,10 @@ class gfxFontFamily {
                                     nsTArray<gfxFontEntry*>& aFontEntryList,
                                     bool aIgnoreSizeTolerance = false);
 
-  // checks for a matching font within the family
-  // used as part of the font fallback process
+  // Checks for a matching font within the family; used as part of the font
+  // fallback process.
+  // Note that when this is called, the caller must already be holding the
+  // gfxPlatformFontList lock.
   void FindFontForChar(GlobalFontMatch* aMatchData);
 
   // checks all fonts for a matching font within the family
@@ -870,52 +939,70 @@ class gfxFontFamily {
   // set when other family names have been read in
   void SetOtherFamilyNamesInitialized() { mOtherFamilyNamesInitialized = true; }
 
-  // read in other localized family names, fullnames and Postscript names
-  // for all faces and append to lookup tables
+  // Read in other localized family names, fullnames and Postscript names
+  // for all faces and append to lookup tables.
+  // Note that when this is called, the caller must already be holding the
+  // gfxPlatformFontList lock.
   virtual void ReadFaceNames(gfxPlatformFontList* aPlatformFontList,
                              bool aNeedFullnamePostscriptNames,
                              FontInfoData* aFontInfoData = nullptr);
 
-  // find faces belonging to this family (platform implementations override
-  // this; should be made pure virtual once all subclasses have been updated)
-  virtual void FindStyleVariations(FontInfoData* aFontInfoData = nullptr) {}
+  // Find faces belonging to this family (platform implementations override).
+  // This is a no-op in cases where the family is explicitly populated by other
+  // means, rather than being asked to find its faces via system API.
+  virtual void FindStyleVariationsLocked(FontInfoData* aFontInfoData = nullptr)
+      REQUIRES(mLock){};
+  void FindStyleVariations(FontInfoData* aFontInfoData = nullptr) {
+    if (mHasStyles) {
+      return;
+    }
+    mozilla::AutoWriteLock lock(mLock);
+    FindStyleVariationsLocked(aFontInfoData);
+  }
 
   // search for a specific face using the Postscript name
-  gfxFontEntry* FindFont(const nsACString& aPostscriptName);
+  gfxFontEntry* FindFont(const nsACString& aFontName,
+                         const nsCStringComparator& aCmp) const;
 
-  // read in cmaps for all the faces
+  // Read in cmaps for all the faces.
+  // Note that when this is called, the caller must already be holding the
+  // gfxPlatformFontList lock.
   void ReadAllCMAPs(FontInfoData* aFontInfoData = nullptr);
 
   bool TestCharacterMap(uint32_t aCh) {
     if (!mFamilyCharacterMapInitialized) {
       ReadAllCMAPs();
     }
+    mozilla::AutoReadLock lock(mLock);
     return mFamilyCharacterMap.test(aCh);
   }
 
-  void ResetCharacterMap() {
+  void ResetCharacterMap() REQUIRES(mLock) {
     mFamilyCharacterMap.reset();
     mFamilyCharacterMapInitialized = false;
   }
 
-  // mark this family as being in the "bad" underline offset blacklist
+  // mark this family as being in the "bad" underline offset blocklist
   void SetBadUnderlineFamily() {
+    mozilla::AutoWriteLock lock(mLock);
     mIsBadUnderlineFamily = true;
     if (mHasStyles) {
       SetBadUnderlineFonts();
     }
   }
 
+  virtual bool IsSingleFaceFamily() const { return false; }
+
   bool IsBadUnderlineFamily() const { return mIsBadUnderlineFamily; }
   bool CheckForFallbackFaces() const { return mCheckForFallbackFaces; }
 
   // sort available fonts to put preferred (standard) faces towards the end
-  void SortAvailableFonts();
+  void SortAvailableFonts() REQUIRES(mLock);
 
   // check whether the family fits into the simple 4-face model,
   // so we can use simplified style-matching;
   // if so set the mIsSimpleFamily flag (defaults to False before we've checked)
-  void CheckForSimpleFamily();
+  void CheckForSimpleFamily() REQUIRES(mLock);
 
   // For memory reporter
   virtual void AddSizeOfExcludingThis(mozilla::MallocSizeOf aMallocSizeOf,
@@ -957,32 +1044,34 @@ class gfxFontFamily {
                                    hb_blob_t* aNameTable,
                                    bool useFullName = false);
 
-  // set whether this font family is in "bad" underline offset blacklist.
-  void SetBadUnderlineFonts() {
-    uint32_t i, numFonts = mAvailableFonts.Length();
-    for (i = 0; i < numFonts; i++) {
-      if (mAvailableFonts[i]) {
-        mAvailableFonts[i]->mIsBadUnderlineFont = true;
+  // set whether this font family is in "bad" underline offset blocklist.
+  void SetBadUnderlineFonts() REQUIRES(mLock) {
+    for (auto& f : mAvailableFonts) {
+      if (f) {
+        f->mIsBadUnderlineFont = true;
       }
     }
   }
 
   nsCString mName;
-  nsTArray<RefPtr<gfxFontEntry>> mAvailableFonts;
-  gfxSparseBitSet mFamilyCharacterMap;
+  nsTArray<RefPtr<gfxFontEntry>> mAvailableFonts GUARDED_BY(mLock);
+  gfxSparseBitSet mFamilyCharacterMap GUARDED_BY(mLock);
+
+  mutable mozilla::RWLock mLock;
 
   FontVisibility mVisibility;
 
-  bool mOtherFamilyNamesInitialized : 1;
-  bool mHasOtherFamilyNames : 1;
-  bool mFaceNamesInitialized : 1;
-  bool mHasStyles : 1;
-  bool mIsSimpleFamily : 1;
+  mozilla::Atomic<bool> mOtherFamilyNamesInitialized;
+  mozilla::Atomic<bool> mFaceNamesInitialized;
+  mozilla::Atomic<bool> mHasStyles;
+  mozilla::Atomic<bool> mFamilyCharacterMapInitialized;
+  mozilla::Atomic<bool> mCheckedForLegacyFamilyNames;
+  mozilla::Atomic<bool> mHasOtherFamilyNames;
+
+  bool mIsSimpleFamily : 1 GUARDED_BY(mLock);
   bool mIsBadUnderlineFamily : 1;
-  bool mFamilyCharacterMapInitialized : 1;
   bool mSkipDefaultFeatureSpaceCheck : 1;
   bool mCheckForFallbackFaces : 1;  // check other faces for character
-  bool mCheckedForLegacyFamilyNames : 1;
 
   enum {
     // for "simple" families, the faces are stored in mAvailableFonts
