@@ -8,11 +8,13 @@
 #define MOZILLA_GFX_RENDERTEXTUREHOST_H
 
 #include "GLConsts.h"
+#include "GLTypes.h"
 #include "nsISupportsImpl.h"
 #include "mozilla/gfx/2D.h"
 #include "mozilla/layers/LayersSurfaces.h"
 #include "mozilla/RefPtr.h"
 #include "mozilla/webrender/webrender_ffi.h"  // for wr::ImageRendering
+#include "mozilla/webrender/WebRenderTypes.h"
 
 namespace mozilla {
 
@@ -22,9 +24,13 @@ class GLContext;
 
 namespace wr {
 
-class RenderDXGITextureHostOGL;
+class RenderAndroidSurfaceTextureHost;
+class RenderCompositor;
+class RenderDXGITextureHost;
+class RenderDXGIYCbCrTextureHost;
+class RenderMacIOSurfaceTextureHost;
 class RenderBufferTextureHost;
-class RenderTextureHostOGL;
+class RenderTextureHostSWGL;
 
 void ActivateBindAndTexParameteri(gl::GLContext* aGL, GLenum aActiveTexture,
                                   GLenum aBindTarget, GLuint aBindTexture,
@@ -37,32 +43,63 @@ class RenderTextureHost {
   RenderTextureHost();
 
   virtual wr::WrExternalImage Lock(uint8_t aChannelIndex, gl::GLContext* aGL,
-                                   wr::ImageRendering aRendering) = 0;
-  virtual void Unlock() = 0;
+                                   wr::ImageRendering aRendering);
+
+  virtual void Unlock() {}
+
+  virtual wr::WrExternalImage LockSWGL(uint8_t aChannelIndex, void* aContext,
+                                       RenderCompositor* aCompositor,
+                                       wr::ImageRendering aRendering);
+
+  virtual void UnlockSWGL() {}
+
   virtual void ClearCachedResources() {}
 
   // Called asynchronouly when corresponding TextureHost's mCompositableCount
   // becomes from 0 to 1. For now, it is used only for
-  // SurfaceTextureHost/RenderAndroidSurfaceTextureHostOGL.
+  // SurfaceTextureHost/RenderAndroidSurfaceTextureHost.
   virtual void PrepareForUse() {}
   // Called asynchronouly when corresponding TextureHost's is actually going to
   // be used by WebRender. For now, it is used only for
-  // SurfaceTextureHost/RenderAndroidSurfaceTextureHostOGL.
-  virtual void NofityForUse() {}
+  // SurfaceTextureHost/RenderAndroidSurfaceTextureHost.
+  virtual void NotifyForUse() {}
   // Called asynchronouly when corresponding TextureHost's mCompositableCount
   // becomes 0. For now, it is used only for
-  // SurfaceTextureHost/RenderAndroidSurfaceTextureHostOGL.
+  // SurfaceTextureHost/RenderAndroidSurfaceTextureHost.
   virtual void NotifyNotUsed() {}
+  // Returns true when RenderTextureHost needs SyncObjectHost::Synchronize()
+  // call, before its usage.
+  virtual bool SyncObjectNeeded() { return false; }
 
-  virtual RenderDXGITextureHostOGL* AsRenderDXGITextureHostOGL() {
+  virtual size_t Bytes() = 0;
+
+  virtual RenderDXGITextureHost* AsRenderDXGITextureHost() { return nullptr; }
+  virtual RenderDXGIYCbCrTextureHost* AsRenderDXGIYCbCrTextureHost() {
     return nullptr;
   }
 
+  virtual RenderMacIOSurfaceTextureHost* AsRenderMacIOSurfaceTextureHost() {
+    return nullptr;
+  }
+
+  virtual RenderAndroidSurfaceTextureHost* AsRenderAndroidSurfaceTextureHost() {
+    return nullptr;
+  }
+
+  virtual RenderTextureHostSWGL* AsRenderTextureHostSWGL() { return nullptr; }
+
  protected:
   virtual ~RenderTextureHost();
 
   bool IsFilterUpdateNecessary(wr::ImageRendering aRendering);
 
+  // Returns the UV coordinates to be used when sampling the texture, in pixels.
+  // For most implementations these will be (0, 0) and (size.x, size.y), but
+  // some texture types (such as RenderAndroidSurfaceTextureHost) require an
+  // additional transform to be applied to the coordinates.
+  virtual std::pair<gfx::Point, gfx::Point> GetUvCoords(
+      gfx::IntSize aTextureSize) const;
+
   wr::ImageRendering mCachedRendering;
 };
 
