@@ -4,7 +4,7 @@
 
 use api::BorderRadius;
 use api::units::*;
-use euclid::{Point2D, Rect, Size2D, Vector2D};
+use euclid::{Point2D, Rect, Box2D, Size2D, Vector2D, point2, point3};
 use euclid::{default, Transform2D, Transform3D, Scale};
 use malloc_size_of::{MallocShallowSizeOf, MallocSizeOf, MallocSizeOfOps};
 use plane_split::{Clipper, Polygon};
@@ -66,6 +66,9 @@ pub trait VecHelper<T> {
     /// Equivalent to `mem::replace(&mut vec, Vec::new())`
     fn take(&mut self) -> Self;
 
+    /// Call clear and return self (useful for chaining with calls that move the vector).
+    fn cleared(self) -> Self;
+
     /// Functionally equivalent to `mem::replace(&mut vec, Vec::new())` but tries
     /// to keep the allocation in the caller if it is empty or replace it with a
     /// pre-allocated vector.
@@ -99,6 +102,12 @@ impl<T> VecHelper<T> for Vec<T> {
         replace(self, Vec::new())
     }
 
+    fn cleared(mut self) -> Self {
+        self.clear();
+
+        self
+    }
+
     fn take_and_preallocate(&mut self) -> Self {
         let len = self.len();
         if len == 0 {
@@ -117,14 +126,22 @@ impl<T> VecHelper<T> for Vec<T> {
 // TODO(gw): We should try and incorporate F <-> T units here,
 //           but it's a bit tricky to do that now with the
 //           way the current spatial tree works.
-#[derive(Debug, Clone, Copy, MallocSizeOf)]
+#[derive(Debug, Clone, Copy, MallocSizeOf, PartialEq)]
 #[cfg_attr(feature = "capture", derive(Serialize))]
+#[cfg_attr(feature = "replay", derive(Deserialize))]
 pub struct ScaleOffset {
     pub scale: default::Vector2D<f32>,
     pub offset: default::Vector2D<f32>,
 }
 
 impl ScaleOffset {
+    pub fn new(sx: f32, sy: f32, tx: f32, ty: f32) -> Self {
+        ScaleOffset {
+            scale: Vector2D::new(sx, sy),
+            offset: Vector2D::new(tx, ty),
+        }
+    }
+
     pub fn identity() -> Self {
         ScaleOffset {
             scale: Vector2D::new(1.0, 1.0),
@@ -141,14 +158,12 @@ impl ScaleOffset {
         // To check that we have a pure scale / translation:
         // Every field must match an identity matrix, except:
         //  - Any value present in tx,ty
-        //  - Any non-neg value present in sx,sy (avoid negative for reflection/rotation)
+        //  - Any value present in sx,sy
 
-        if m.m11 < 0.0 ||
-           m.m12.abs() > NEARLY_ZERO ||
+        if m.m12.abs() > NEARLY_ZERO ||
            m.m13.abs() > NEARLY_ZERO ||
            m.m14.abs() > NEARLY_ZERO ||
            m.m21.abs() > NEARLY_ZERO ||
-           m.m22 < 0.0 ||
            m.m23.abs() > NEARLY_ZERO ||
            m.m24.abs() > NEARLY_ZERO ||
            m.m31.abs() > NEARLY_ZERO ||
@@ -173,6 +188,13 @@ impl ScaleOffset {
         }
     }
 
+    pub fn from_scale(scale: default::Vector2D<f32>) -> Self {
+        ScaleOffset {
+            scale,
+            offset: Vector2D::new(0.0, 0.0),
+        }
+    }
+
     pub fn inverse(&self) -> Self {
         ScaleOffset {
             scale: Vector2D::new(
@@ -220,29 +242,80 @@ impl ScaleOffset {
         }
     }
 
-    pub fn map_rect<F, T>(&self, rect: &Rect<f32, F>) -> Rect<f32, T> {
-        Rect::new(
-            Point2D::new(
-                rect.origin.x * self.scale.x + self.offset.x,
-                rect.origin.y * self.scale.y + self.offset.y,
-            ),
-            Size2D::new(
-                rect.size.width * self.scale.x,
-                rect.size.height * self.scale.y,
-            )
+    pub fn map_rect<F, T>(&self, rect: &Box2D<f32, F>) -> Box2D<f32, T> {
+        // TODO(gw): The logic below can return an unexpected result if the supplied
+        //           rect is invalid (has size < 0). Since Gecko currently supplied
+        //           invalid rects in some cases, adding a max(0) here ensures that
+        //           mapping an invalid rect retains the property that rect.is_empty()
+        //           will return true (the mapped rect output will have size 0 instead
+        //           of a negative size). In future we could catch / assert / fix
+        //           these invalid rects earlier, and assert here instead.
+
+        let w = rect.width().max(0.0);
+        let h = rect.height().max(0.0);
+
+        let mut x0 = rect.min.x * self.scale.x + self.offset.x;
+        let mut y0 = rect.min.y * self.scale.y + self.offset.y;
+
+        let mut sx = w * self.scale.x;
+        let mut sy = h * self.scale.y;
+        // Handle negative scale. Previously, branchless float math was used to find the
+        // min / max vertices and size. However, that sequence of operations was producind
+        // additional floating point accuracy on android emulator builds, causing one test
+        // to fail an assert. Instead, we retain the same math as previously, and adjust
+        // the origin / size if required.
+
+        if self.scale.x < 0.0 {
+            x0 += sx;
+            sx = -sx;
+        }
+        if self.scale.y < 0.0 {
+            y0 += sy;
+            sy = -sy;
+        }
+
+        Box2D::from_origin_and_size(
+            Point2D::new(x0, y0),
+            Size2D::new(sx, sy),
         )
     }
 
-    pub fn unmap_rect<F, T>(&self, rect: &Rect<f32, F>) -> Rect<f32, T> {
-        Rect::new(
-            Point2D::new(
-                (rect.origin.x - self.offset.x) / self.scale.x,
-                (rect.origin.y - self.offset.y) / self.scale.y,
-            ),
-            Size2D::new(
-                rect.size.width / self.scale.x,
-                rect.size.height / self.scale.y,
-            )
+    pub fn unmap_rect<F, T>(&self, rect: &Box2D<f32, F>) -> Box2D<f32, T> {
+        // TODO(gw): The logic below can return an unexpected result if the supplied
+        //           rect is invalid (has size < 0). Since Gecko currently supplied
+        //           invalid rects in some cases, adding a max(0) here ensures that
+        //           mapping an invalid rect retains the property that rect.is_empty()
+        //           will return true (the mapped rect output will have size 0 instead
+        //           of a negative size). In future we could catch / assert / fix
+        //           these invalid rects earlier, and assert here instead.
+
+        let w = rect.width().max(0.0);
+        let h = rect.height().max(0.0);
+
+        let mut x0 = (rect.min.x - self.offset.x) / self.scale.x;
+        let mut y0 = (rect.min.y - self.offset.y) / self.scale.y;
+
+        let mut sx = w / self.scale.x;
+        let mut sy = h / self.scale.y;
+
+        // Handle negative scale. Previously, branchless float math was used to find the
+        // min / max vertices and size. However, that sequence of operations was producind
+        // additional floating point accuracy on android emulator builds, causing one test
+        // to fail an assert. Instead, we retain the same math as previously, and adjust
+        // the origin / size if required.
+
+        if self.scale.x < 0.0 {
+            x0 += sx;
+            sx = -sx;
+        }
+        if self.scale.y < 0.0 {
+            y0 += sy;
+            sy = -sy;
+        }
+
+        Box2D::from_origin_and_size(
+            Point2D::new(x0, y0),
+            Size2D::new(sx, sy),
         )
     }
 
@@ -275,7 +348,7 @@ impl ScaleOffset {
     }
 
     pub fn to_transform<F, T>(&self) -> Transform3D<f32, F, T> {
-        Transform3D::row_major(
+        Transform3D::new(
             self.scale.x,
             0.0,
             0.0,
@@ -310,7 +383,7 @@ pub trait MatrixHelpers<Src, Dst> {
     /// transformed by this matrix to have scaling exceeding the supplied limit.
     fn exceeds_2d_scale(&self, limit: f64) -> bool;
     fn inverse_project(&self, target: &Point2D<f32, Dst>) -> Option<Point2D<f32, Src>>;
-    fn inverse_rect_footprint(&self, rect: &Rect<f32, Dst>) -> Option<Rect<f32, Src>>;
+    fn inverse_rect_footprint(&self, rect: &Box2D<f32, Dst>) -> Option<Box2D<f32, Src>>;
     fn transform_kind(&self) -> TransformedRectKind;
     fn is_simple_translation(&self) -> bool;
     fn is_simple_2d_translation(&self) -> bool;
@@ -324,6 +397,8 @@ pub trait MatrixHelpers<Src, Dst> {
     /// Turn Z transformation into identity. This is useful when crossing "flat"
     /// transform styled stacking contexts upon traversing the coordinate systems.
     fn flatten_z_output(&mut self);
+
+    fn cast_unit<NewSrc, NewDst>(&self) -> Transform3D<f32, NewSrc, NewDst>;
 }
 
 impl<Src, Dst> MatrixHelpers<Src, Dst> for Transform3D<f32, Src, Dst> {
@@ -374,22 +449,26 @@ impl<Src, Dst> MatrixHelpers<Src, Dst> for Transform3D<f32, Src, Dst> {
         self.m21 * self.m21 + self.m22 * self.m22 > limit2
     }
 
+    /// Find out a point in `Src` that would be projected into the `target`.
     fn inverse_project(&self, target: &Point2D<f32, Dst>) -> Option<Point2D<f32, Src>> {
-        let m: Transform2D<f32, Src, Dst>;
-        m = Transform2D::column_major(
-            self.m11 - target.x * self.m14,
-            self.m21 - target.x * self.m24,
-            self.m41 - target.x * self.m44,
-            self.m12 - target.y * self.m14,
-            self.m22 - target.y * self.m24,
-            self.m42 - target.y * self.m44,
+        // form the linear equation for the hyperplane intersection
+        let m = Transform2D::<f32, Src, Dst>::new(
+            self.m11 - target.x * self.m14, self.m12 - target.y * self.m14,
+            self.m21 - target.x * self.m24, self.m22 - target.y * self.m24,
+            self.m41 - target.x * self.m44, self.m42 - target.y * self.m44,
         );
-        m.inverse().map(|inv| Point2D::new(inv.m31, inv.m32))
+        let inv = m.inverse()?;
+        // we found the point, now check if it maps to the positive hemisphere
+        if inv.m31 * self.m14 + inv.m32 * self.m24 + self.m44 > 0.0 {
+            Some(Point2D::new(inv.m31, inv.m32))
+        } else {
+            None
+        }
     }
 
-    fn inverse_rect_footprint(&self, rect: &Rect<f32, Dst>) -> Option<Rect<f32, Src>> {
-        Some(Rect::from_points(&[
-            self.inverse_project(&rect.origin)?,
+    fn inverse_rect_footprint(&self, rect: &Box2D<f32, Dst>) -> Option<Box2D<f32, Src>> {
+        Some(Box2D::from_points(&[
+            self.inverse_project(&rect.top_left())?,
             self.inverse_project(&rect.top_right())?,
             self.inverse_project(&rect.bottom_left())?,
             self.inverse_project(&rect.bottom_right())?,
@@ -434,7 +513,7 @@ impl<Src, Dst> MatrixHelpers<Src, Dst> for Transform3D<f32, Src, Dst> {
      *  a  b  0  1
      */
     fn is_2d_scale_translation(&self) -> bool {
-        (self.m33 - 1.0).abs() < NEARLY_ZERO && 
+        (self.m33 - 1.0).abs() < NEARLY_ZERO &&
             (self.m44 - 1.0).abs() < NEARLY_ZERO &&
             self.m12.abs() < NEARLY_ZERO && self.m13.abs() < NEARLY_ZERO && self.m14.abs() < NEARLY_ZERO &&
             self.m21.abs() < NEARLY_ZERO && self.m23.abs() < NEARLY_ZERO && self.m24.abs() < NEARLY_ZERO &&
@@ -462,6 +541,16 @@ impl<Src, Dst> MatrixHelpers<Src, Dst> for Transform3D<f32, Src, Dst> {
         self.m23 = 0.0;
         self.m33 = 1.0;
         self.m43 = 0.0;
+        //Note: we used to zero out m3? as well, see "reftests/flatten-all-flat.yaml" test
+    }
+
+    fn cast_unit<NewSrc, NewDst>(&self) -> Transform3D<f32, NewSrc, NewDst> {
+        Transform3D::new(
+            self.m11, self.m12, self.m13, self.m14,
+            self.m21, self.m22, self.m23, self.m24,
+            self.m31, self.m32, self.m33, self.m34,
+            self.m41, self.m42, self.m43, self.m44,
+        )
     }
 }
 
@@ -486,7 +575,6 @@ where
     Self: Sized,
 {
     fn from_floats(x0: f32, y0: f32, x1: f32, y1: f32) -> Self;
-    fn is_well_formed_and_nonempty(&self) -> bool;
     fn snap(&self) -> Self;
 }
 
@@ -498,10 +586,6 @@ impl<U> RectHelpers<U> for Rect<f32, U> {
         )
     }
 
-    fn is_well_formed_and_nonempty(&self) -> bool {
-        self.size.width > 0.0 && self.size.height > 0.0
-    }
-
     fn snap(&self) -> Self {
         let origin = Point2D::new(
             (self.origin.x + 0.5).floor(),
@@ -517,6 +601,19 @@ impl<U> RectHelpers<U> for Rect<f32, U> {
     }
 }
 
+impl<U> RectHelpers<U> for Box2D<f32, U> {
+    fn from_floats(x0: f32, y0: f32, x1: f32, y1: f32) -> Self {
+        Box2D {
+            min: Point2D::new(x0, y0),
+            max: Point2D::new(x1, y1),
+        }
+    }
+
+    fn snap(&self) -> Self {
+        self.round()
+    }
+}
+
 pub trait VectorHelpers<U>
 where
     Self: Sized,
@@ -553,23 +650,23 @@ pub fn pack_as_float(value: u32) -> f32 {
 
 #[inline]
 fn extract_inner_rect_impl<U>(
-    rect: &Rect<f32, U>,
+    rect: &Box2D<f32, U>,
     radii: &BorderRadius,
     k: f32,
-) -> Option<Rect<f32, U>> {
+) -> Option<Box2D<f32, U>> {
     // `k` defines how much border is taken into account
     // We enforce the offsets to be rounded to pixel boundaries
     // by `ceil`-ing and `floor`-ing them
 
     let xl = (k * radii.top_left.width.max(radii.bottom_left.width)).ceil();
-    let xr = (rect.size.width - k * radii.top_right.width.max(radii.bottom_right.width)).floor();
+    let xr = (rect.width() - k * radii.top_right.width.max(radii.bottom_right.width)).floor();
     let yt = (k * radii.top_left.height.max(radii.top_right.height)).ceil();
     let yb =
-        (rect.size.height - k * radii.bottom_left.height.max(radii.bottom_right.height)).floor();
+        (rect.height() - k * radii.bottom_left.height.max(radii.bottom_right.height)).floor();
 
     if xl <= xr && yt <= yb {
-        Some(Rect::new(
-            Point2D::new(rect.origin.x + xl, rect.origin.y + yt),
+        Some(Box2D::from_origin_and_size(
+            Point2D::new(rect.min.x + xl, rect.min.y + yt),
             Size2D::new(xr - xl, yb - yt),
         ))
     } else {
@@ -580,20 +677,26 @@ fn extract_inner_rect_impl<U>(
 /// Return an aligned rectangle that is inside the clip region and doesn't intersect
 /// any of the bounding rectangles of the rounded corners.
 pub fn extract_inner_rect_safe<U>(
-    rect: &Rect<f32, U>,
+    rect: &Box2D<f32, U>,
     radii: &BorderRadius,
-) -> Option<Rect<f32, U>> {
+) -> Option<Box2D<f32, U>> {
     // value of `k==1.0` is used for extraction of the corner rectangles
     // see `SEGMENT_CORNER_*` in `clip_shared.glsl`
     extract_inner_rect_impl(rect, radii, 1.0)
 }
 
+#[cfg(test)]
+use euclid::vec3;
+
 #[cfg(test)]
 pub mod test {
     use super::*;
-    use euclid::default::{Point2D, Transform3D};
-    use euclid::Angle;
+    use euclid::default::{Point2D, Size2D, Transform3D};
+    use euclid::{Angle, approxeq::ApproxEq};
     use std::f32::consts::PI;
+    use crate::clip::{is_left_of_line, polygon_contains_point};
+    use crate::prim_store::PolygonKey;
+    use api::FillRule;
 
     #[test]
     fn inverse_project() {
@@ -601,31 +704,104 @@ pub mod test {
         let p0 = Point2D::new(1.0, 2.0);
         // an identical transform doesn't need any inverse projection
         assert_eq!(m0.inverse_project(&p0), Some(p0));
-        let m1 = Transform3D::create_rotation(0.0, 1.0, 0.0, Angle::radians(PI / 3.0));
+        let m1 = Transform3D::rotation(0.0, 1.0, 0.0, Angle::radians(-PI / 3.0));
         // rotation by 60 degrees would imply scaling of X component by a factor of 2
         assert_eq!(m1.inverse_project(&p0), Some(Point2D::new(2.0, 2.0)));
     }
 
+    #[test]
+    fn inverse_project_footprint() {
+        let m = Transform3D::new(
+            0.477499992, 0.135000005, -1.0, 0.000624999986,
+            -0.642787635, 0.766044438, 0.0, 0.0,
+            0.766044438, 0.642787635, 0.0, 0.0,
+            1137.10986, 113.71286, 402.0, 0.748749971,
+        );
+        let r = Box2D::from_size(Size2D::new(804.0, 804.0));
+        {
+            let points = &[
+                r.top_left(),
+                r.top_right(),
+                r.bottom_left(),
+                r.bottom_right(),
+            ];
+            let mi = m.inverse().unwrap();
+            // In this section, we do the forward and backward transformation
+            // to confirm that its bijective.
+            // We also do the inverse projection path, and confirm it functions the same way.
+            info!("Points:");
+            for p in points {
+                let pp = m.transform_point2d_homogeneous(*p);
+                let p3 = pp.to_point3d().unwrap();
+                let pi = mi.transform_point3d_homogeneous(p3);
+                let px = pi.to_point2d().unwrap();
+                let py = m.inverse_project(&pp.to_point2d().unwrap()).unwrap();
+                info!("\t{:?} -> {:?} -> {:?} -> ({:?} -> {:?}, {:?})", p, pp, p3, pi, px, py);
+                assert!(px.approx_eq_eps(p, &Point2D::new(0.001, 0.001)));
+                assert!(py.approx_eq_eps(p, &Point2D::new(0.001, 0.001)));
+            }
+        }
+        // project
+        let rp = project_rect(&m, &r, &Box2D::from_size(Size2D::new(1000.0, 1000.0))).unwrap();
+        info!("Projected {:?}", rp);
+        // one of the points ends up in the negative hemisphere
+        assert_eq!(m.inverse_project(&rp.min), None);
+        // inverse
+        if let Some(ri) = m.inverse_rect_footprint(&rp) {
+            // inverse footprint should be larger, since it doesn't know the original Z
+            assert!(ri.contains_box(&r), "Inverse {:?}", ri);
+        }
+    }
+
     fn validate_convert(xref: &LayoutTransform) {
         let so = ScaleOffset::from_transform(xref).unwrap();
         let xf = so.to_transform();
         assert!(xref.approx_eq(&xf));
     }
 
+    #[test]
+    fn negative_scale_map_unmap() {
+        let xref = LayoutTransform::scale(1.0, -1.0, 1.0)
+                        .pre_translate(LayoutVector3D::new(124.0, 38.0, 0.0));
+        let so = ScaleOffset::from_transform(&xref).unwrap();
+        let local_rect = Box2D {
+            min: LayoutPoint::new(50.0, -100.0),
+            max: LayoutPoint::new(250.0, 300.0),
+        };
+
+        let mapped_rect = so.map_rect::<LayoutPixel, DevicePixel>(&local_rect);
+        let xf_rect = project_rect(
+            &xref,
+            &local_rect,
+            &LayoutRect::max_rect(),
+        ).unwrap();
+
+        assert!(mapped_rect.min.x.approx_eq(&xf_rect.min.x));
+        assert!(mapped_rect.min.y.approx_eq(&xf_rect.min.y));
+        assert!(mapped_rect.max.x.approx_eq(&xf_rect.max.x));
+        assert!(mapped_rect.max.y.approx_eq(&xf_rect.max.y));
+
+        let unmapped_rect = so.unmap_rect::<DevicePixel, LayoutPixel>(&mapped_rect);
+        assert!(unmapped_rect.min.x.approx_eq(&local_rect.min.x));
+        assert!(unmapped_rect.min.y.approx_eq(&local_rect.min.y));
+        assert!(unmapped_rect.max.x.approx_eq(&local_rect.max.x));
+        assert!(unmapped_rect.max.y.approx_eq(&local_rect.max.y));
+    }
+
     #[test]
     fn scale_offset_convert() {
-        let xref = LayoutTransform::create_translation(130.0, 200.0, 0.0);
+        let xref = LayoutTransform::translation(130.0, 200.0, 0.0);
         validate_convert(&xref);
 
-        let xref = LayoutTransform::create_scale(13.0, 8.0, 1.0);
+        let xref = LayoutTransform::scale(13.0, 8.0, 1.0);
         validate_convert(&xref);
 
-        let xref = LayoutTransform::create_scale(0.5, 0.5, 1.0)
+        let xref = LayoutTransform::scale(0.5, 0.5, 1.0)
                         .pre_translate(LayoutVector3D::new(124.0, 38.0, 0.0));
         validate_convert(&xref);
 
-        let xref = LayoutTransform::create_translation(50.0, 240.0, 0.0)
-                        .pre_transform(&LayoutTransform::create_scale(30.0, 11.0, 1.0));
+        let xref = LayoutTransform::scale(30.0, 11.0, 1.0)
+            .then_translate(vec3(50.0, 240.0, 0.0));
         validate_convert(&xref);
     }
 
@@ -642,23 +818,24 @@ pub mod test {
 
     #[test]
     fn scale_offset_inverse() {
-        let xref = LayoutTransform::create_translation(130.0, 200.0, 0.0);
+        let xref = LayoutTransform::translation(130.0, 200.0, 0.0);
         validate_inverse(&xref);
 
-        let xref = LayoutTransform::create_scale(13.0, 8.0, 1.0);
+        let xref = LayoutTransform::scale(13.0, 8.0, 1.0);
         validate_inverse(&xref);
 
-        let xref = LayoutTransform::create_scale(0.5, 0.5, 1.0)
-                        .pre_translate(LayoutVector3D::new(124.0, 38.0, 0.0));
+        let xref = LayoutTransform::translation(124.0, 38.0, 0.0).
+            then_scale(0.5, 0.5, 1.0);
+
         validate_inverse(&xref);
 
-        let xref = LayoutTransform::create_translation(50.0, 240.0, 0.0)
-                        .pre_transform(&LayoutTransform::create_scale(30.0, 11.0, 1.0));
+        let xref = LayoutTransform::scale(30.0, 11.0, 1.0)
+            .then_translate(vec3(50.0, 240.0, 0.0));
         validate_inverse(&xref);
     }
 
     fn validate_accumulate(x0: &LayoutTransform, x1: &LayoutTransform) {
-        let x = x0.pre_transform(x1);
+        let x = x1.then(&x0);
 
         let s0 = ScaleOffset::from_transform(x0).unwrap();
         let s1 = ScaleOffset::from_transform(x1).unwrap();
@@ -670,8 +847,8 @@ pub mod test {
 
     #[test]
     fn scale_offset_accumulate() {
-        let x0 = LayoutTransform::create_translation(130.0, 200.0, 0.0);
-        let x1 = LayoutTransform::create_scale(7.0, 3.0, 1.0);
+        let x0 = LayoutTransform::translation(130.0, 200.0, 0.0);
+        let x1 = LayoutTransform::scale(7.0, 3.0, 1.0);
 
         validate_accumulate(&x0, &x1);
     }
@@ -691,6 +868,83 @@ pub mod test {
         assert_eq!(origin, Point2D::new(1.0, 0.5));
         assert_eq!(m.transform_point2d(origin), Some(Point2D::zero()));
     }
+
+    #[test]
+    fn polygon_clip_is_left_of_point() {
+        // Define points of a line through (1, -3) and (-2, 6) to test against.
+        // If the triplet consisting of these two points and the test point
+        // form a counter-clockwise triangle, then the test point is on the
+        // left. The easiest way to visualize this is with an "ascending"
+        // line from low-Y to high-Y.
+        let p0_x = 1.0;
+        let p0_y = -3.0;
+        let p1_x = -2.0;
+        let p1_y = 6.0;
+
+        // Test some points to the left of the line.
+        assert!(is_left_of_line(-9.0, 0.0, p0_x, p0_y, p1_x, p1_y) > 0.0);
+        assert!(is_left_of_line(-1.0, 1.0, p0_x, p0_y, p1_x, p1_y) > 0.0);
+        assert!(is_left_of_line(1.0, -4.0, p0_x, p0_y, p1_x, p1_y) > 0.0);
+
+        // Test some points on the line.
+        assert!(is_left_of_line(-3.0, 9.0, p0_x, p0_y, p1_x, p1_y) == 0.0);
+        assert!(is_left_of_line(0.0, 0.0, p0_x, p0_y, p1_x, p1_y) == 0.0);
+        assert!(is_left_of_line(100.0, -300.0, p0_x, p0_y, p1_x, p1_y) == 0.0);
+
+        // Test some points to the right of the line.
+        assert!(is_left_of_line(0.0, 1.0, p0_x, p0_y, p1_x, p1_y) < 0.0);
+        assert!(is_left_of_line(-4.0, 13.0, p0_x, p0_y, p1_x, p1_y) < 0.0);
+        assert!(is_left_of_line(5.0, -12.0, p0_x, p0_y, p1_x, p1_y) < 0.0);
+    }
+
+    #[test]
+    fn polygon_clip_contains_point() {
+        // We define the points of a self-overlapping polygon, which we will
+        // use to create polygons with different windings and fill rules.
+        let p0 = LayoutPoint::new(4.0, 4.0);
+        let p1 = LayoutPoint::new(6.0, 4.0);
+        let p2 = LayoutPoint::new(4.0, 7.0);
+        let p3 = LayoutPoint::new(2.0, 1.0);
+        let p4 = LayoutPoint::new(8.0, 1.0);
+        let p5 = LayoutPoint::new(6.0, 7.0);
+
+        let poly_clockwise_nonzero = PolygonKey::new(
+            &[p5, p4, p3, p2, p1, p0].to_vec(), FillRule::Nonzero
+        );
+        let poly_clockwise_evenodd = PolygonKey::new(
+            &[p5, p4, p3, p2, p1, p0].to_vec(), FillRule::Evenodd
+        );
+        let poly_counter_clockwise_nonzero = PolygonKey::new(
+            &[p0, p1, p2, p3, p4, p5].to_vec(), FillRule::Nonzero
+        );
+        let poly_counter_clockwise_evenodd = PolygonKey::new(
+            &[p0, p1, p2, p3, p4, p5].to_vec(), FillRule::Evenodd
+        );
+
+        // We define a rect that provides a bounding clip area of
+        // the polygon.
+        let rect = LayoutRect::from_size(LayoutSize::new(10.0, 10.0));
+
+        // And we'll test three points of interest.
+        let p_inside_once = LayoutPoint::new(5.0, 3.0);
+        let p_inside_twice = LayoutPoint::new(5.0, 5.0);
+        let p_outside = LayoutPoint::new(9.0, 9.0);
+
+        // We should get the same results for both clockwise and
+        // counter-clockwise polygons.
+        // For nonzero polygons, the inside twice point is considered inside.
+        for poly_nonzero in vec![poly_clockwise_nonzero, poly_counter_clockwise_nonzero].iter() {
+            assert_eq!(polygon_contains_point(&p_inside_once, &rect, &poly_nonzero), true);
+            assert_eq!(polygon_contains_point(&p_inside_twice, &rect, &poly_nonzero), true);
+            assert_eq!(polygon_contains_point(&p_outside, &rect, &poly_nonzero), false);
+        }
+        // For evenodd polygons, the inside twice point is considered outside.
+        for poly_evenodd in vec![poly_clockwise_evenodd, poly_counter_clockwise_evenodd].iter() {
+            assert_eq!(polygon_contains_point(&p_inside_once, &rect, &poly_evenodd), true);
+            assert_eq!(polygon_contains_point(&p_inside_twice, &rect, &poly_evenodd), false);
+            assert_eq!(polygon_contains_point(&p_outside, &rect, &poly_evenodd), false);
+        }
+    }
 }
 
 pub trait MaxRect {
@@ -699,7 +953,7 @@ pub trait MaxRect {
 
 impl MaxRect for DeviceIntRect {
     fn max_rect() -> Self {
-        DeviceIntRect::new(
+        DeviceIntRect::from_origin_and_size(
             DeviceIntPoint::new(i32::MIN / 2, i32::MIN / 2),
             DeviceIntSize::new(i32::MAX, i32::MAX),
         )
@@ -723,9 +977,28 @@ impl<U> MaxRect for Rect<f32, U> {
     }
 }
 
+impl<U> MaxRect for Box2D<f32, U> {
+    fn max_rect() -> Self {
+        // Having an unlimited bounding box is fine up until we try
+        // to cast it to `i32`, where we get `-2147483648` for any
+        // values larger than or equal to 2^31.
+        //
+        // Note: clamping to i32::MIN and i32::MAX is not a solution,
+        // with explanation left as an exercise for the reader.
+        const MAX_COORD: f32 = 1.0e9;
+
+        Box2D::new(
+            Point2D::new(-MAX_COORD, -MAX_COORD),
+            Point2D::new(MAX_COORD, MAX_COORD),
+        )
+    }
+}
+
 /// An enum that tries to avoid expensive transformation matrix calculations
 /// when possible when dealing with non-perspective axis-aligned transformations.
 #[derive(Debug, MallocSizeOf)]
+#[cfg_attr(feature = "capture", derive(Serialize))]
+#[cfg_attr(feature = "replay", derive(Deserialize))]
 pub enum FastTransform<Src, Dst> {
     /// A simple offset, which can be used without doing any matrix math.
     Offset(Vector2D<f32, Src>),
@@ -780,7 +1053,7 @@ impl<Src, Dst> FastTransform<Src, Dst> {
     pub fn to_transform(&self) -> Cow<Transform3D<f32, Src, Dst>> {
         match *self {
             FastTransform::Offset(offset) => Cow::Owned(
-                Transform3D::create_translation(offset.x, offset.y, 0.0)
+                Transform3D::translation(offset.x, offset.y, 0.0)
             ),
             FastTransform::Transform { ref transform, .. } => Cow::Borrowed(transform),
         }
@@ -799,7 +1072,7 @@ impl<Src, Dst> FastTransform<Src, Dst> {
         }
     }
 
-    pub fn post_transform<NewDst>(&self, other: &FastTransform<Dst, NewDst>) -> FastTransform<Src, NewDst> {
+    pub fn then<NewDst>(&self, other: &FastTransform<Dst, NewDst>) -> FastTransform<Src, NewDst> {
         match *self {
             FastTransform::Offset(offset) => match *other {
                 FastTransform::Offset(other_offset) => {
@@ -817,15 +1090,15 @@ impl<Src, Dst> FastTransform<Src, Dst> {
                 FastTransform::Offset(other_offset) => {
                     FastTransform::with_transform(
                         transform
-                            .post_translate(other_offset.to_3d())
+                            .then_translate(other_offset.to_3d())
                             .with_destination::<NewDst>()
                     )
                 }
                 FastTransform::Transform { transform: ref other_transform, inverse: ref other_inverse, is_2d: other_is_2d } => {
                     FastTransform::Transform {
-                        transform: transform.post_transform(other_transform),
+                        transform: transform.then(other_transform),
                         inverse: inverse.as_ref().and_then(|self_inv|
-                            other_inverse.as_ref().map(|other_inv| self_inv.pre_transform(other_inv))
+                            other_inverse.as_ref().map(|other_inv| other_inv.then(self_inv))
                         ),
                         is_2d: is_2d & other_is_2d,
                     }
@@ -838,7 +1111,7 @@ impl<Src, Dst> FastTransform<Src, Dst> {
         &self,
         other: &FastTransform<NewSrc, Src>
     ) -> FastTransform<NewSrc, Dst> {
-        other.post_transform(self)
+        other.then(self)
     }
 
     pub fn pre_translate(&self, other_offset: Vector2D<f32, Src>) -> Self {
@@ -850,13 +1123,13 @@ impl<Src, Dst> FastTransform<Src, Dst> {
         }
     }
 
-    pub fn post_translate(&self, other_offset: Vector2D<f32, Dst>) -> Self {
+    pub fn then_translate(&self, other_offset: Vector2D<f32, Dst>) -> Self {
         match *self {
             FastTransform::Offset(offset) => {
                 FastTransform::Offset(offset + other_offset * Scale::<_, _, Src>::new(1.0))
             }
             FastTransform::Transform { ref transform, .. } => {
-                let transform = transform.post_translate(other_offset.to_3d());
+                let transform = transform.then_translate(other_offset.to_3d());
                 FastTransform::with_transform(transform)
             }
         }
@@ -884,6 +1157,24 @@ impl<Src, Dst> FastTransform<Src, Dst> {
         }
     }
 
+    #[inline(always)]
+    pub fn project_point2d(&self, point: Point2D<f32, Src>) -> Option<Point2D<f32, Dst>> {
+        match* self {
+            FastTransform::Offset(..) => self.transform_point2d(point),
+            FastTransform::Transform{ref transform, ..} => {
+                // Find a value for z that will transform to 0.
+
+                // The transformed value of z is computed as:
+                // z' = point.x * self.m13 + point.y * self.m23 + z * self.m33 + self.m43
+
+                // Solving for z when z' = 0 gives us:
+                let z = -(point.x * transform.m13 + point.y * transform.m23 + transform.m43) / transform.m33;
+
+                transform.transform_point3d(point3(point.x, point.y, z)).map(| p3 | point2(p3.x, p3.y))
+            }
+        }
+    }
+
     #[inline(always)]
     pub fn inverse(&self) -> Option<FastTransform<Dst, Src>> {
         match *self {
@@ -918,13 +1209,13 @@ pub type LayoutToWorldFastTransform = FastTransform<LayoutPixel, WorldPixel>;
 
 pub fn project_rect<F, T>(
     transform: &Transform3D<f32, F, T>,
-    rect: &Rect<f32, F>,
-    bounds: &Rect<f32, T>,
-) -> Option<Rect<f32, T>>
+    rect: &Box2D<f32, F>,
+    bounds: &Box2D<f32, T>,
+) -> Option<Box2D<f32, T>>
  where F: fmt::Debug
 {
     let homogens = [
-        transform.transform_point2d_homogeneous(rect.origin),
+        transform.transform_point2d_homogeneous(rect.top_left()),
         transform.transform_point2d_homogeneous(rect.top_right()),
         transform.transform_point2d_homogeneous(rect.bottom_left()),
         transform.transform_point2d_homogeneous(rect.bottom_right()),
@@ -934,11 +1225,11 @@ pub fn project_rect<F, T>(
     // Otherwise, it will be clamped to the screen bounds anyway.
     if homogens.iter().any(|h| h.w <= 0.0 || h.w.is_nan()) {
         let mut clipper = Clipper::new();
-        let polygon = Polygon::from_rect(*rect, 1);
+        let polygon = Polygon::from_rect(rect.to_rect(), 1);
 
         let planes = match Clipper::<_, _, usize>::frustum_planes(
             transform,
-            Some(*bounds),
+            Some(bounds.to_rect()),
         ) {
             Ok(planes) => planes,
             Err(..) => return None,
@@ -953,7 +1244,7 @@ pub fn project_rect<F, T>(
             return None
         }
 
-        Some(Rect::from_points(results
+        Some(Box2D::from_points(results
             .into_iter()
             // filter out parts behind the view plane
             .flat_map(|poly| &poly.points)
@@ -965,7 +1256,7 @@ pub fn project_rect<F, T>(
         ))
     } else {
         // we just checked for all the points to be in positive hemisphere, so `unwrap` is valid
-        Some(Rect::from_points(&[
+        Some(Box2D::from_points(&[
             homogens[0].to_point2d().unwrap(),
             homogens[1].to_point2d().unwrap(),
             homogens[2].to_point2d().unwrap(),
@@ -974,15 +1265,6 @@ pub fn project_rect<F, T>(
     }
 }
 
-pub fn raster_rect_to_device_pixels(
-    rect: RasterRect,
-    device_pixel_scale: DevicePixelScale,
-) -> DeviceRect {
-    let world_rect = rect * Scale::new(1.0);
-    let device_rect = world_rect * device_pixel_scale;
-    device_rect.round_out()
-}
-
 /// Run the first callback over all elements in the array. If the callback returns true,
 /// the element is removed from the array and moved to a second callback.
 ///
@@ -1082,6 +1364,56 @@ impl Recycler {
     }
 }
 
+/// Record the size of a data structure to preallocate a similar size
+/// at the next frame and avoid growing it too many time.
+#[derive(Copy, Clone, Debug)]
+pub struct Preallocator {
+    size: usize,
+}
+
+impl Preallocator {
+    pub fn new(initial_size: usize) -> Self {
+        Preallocator {
+            size: initial_size,
+        }
+    }
+
+    /// Record the size of a vector to preallocate it the next frame.
+    pub fn record_vec<T>(&mut self, vec: &Vec<T>) {
+        let len = vec.len();
+        if len > self.size {
+            self.size = len;
+        } else {
+            self.size = (self.size + len) / 2;
+        }
+    }
+
+    /// The size that we'll preallocate the vector with.
+    pub fn preallocation_size(&self) -> usize {
+        // Round up to multiple of 16 to avoid small tiny
+        // variations causing reallocations.
+        (self.size + 15) & !15
+    }
+
+    /// Preallocate vector storage.
+    ///
+    /// The preallocated amount depends on the length recorded in the last
+    /// record_vec call.
+    pub fn preallocate_vec<T>(&self, vec: &mut Vec<T>) {
+        let len = vec.len();
+        let cap = self.preallocation_size();
+        if len < cap {
+            vec.reserve(cap - len);
+        }
+    }
+}
+
+impl Default for Preallocator {
+    fn default() -> Self {
+        Self::new(0)
+    }
+}
+
 /// Arc wrapper to support measurement via MallocSizeOf.
 ///
 /// Memory reporting for Arcs is tricky because of the risk of double-counting.
@@ -1134,11 +1466,14 @@ impl<T: MallocSizeOf> MallocSizeOf for PrimaryArc<T> {
 /// modifications:
 ///
 /// * Removed `xMajor` parameter.
+/// * All arithmetics is done with double precision.
 pub fn scale_factors<Src, Dst>(
     mat: &Transform3D<f32, Src, Dst>
 ) -> (f32, f32) {
+    let m11 = mat.m11 as f64;
+    let m12 = mat.m12 as f64;
     // Determinant is just of the 2D component.
-    let det = mat.m11 * mat.m22 - mat.m12 * mat.m21;
+    let det = m11 * mat.m22 as f64 - m12 * mat.m21 as f64;
     if det == 0.0 {
         return (0.0, 0.0);
     }
@@ -1146,10 +1481,23 @@ pub fn scale_factors<Src, Dst>(
     // ignore mirroring
     let det = det.abs();
 
-    let major = (mat.m11 * mat.m11 + mat.m12 * mat.m12).sqrt();
+    let major = (m11 * m11 + m12 * m12).sqrt();
     let minor = if major != 0.0 { det / major } else { 0.0 };
 
-    (major, minor)
+    (major as f32, minor as f32)
+}
+
+#[test]
+fn scale_factors_large() {
+    // https://bugzilla.mozilla.org/show_bug.cgi?id=1748499
+    let mat = Transform3D::<f32, (), ()>::new(
+        1.6534229920333123e27, 3.673100922561787e27, 0.0, 0.0,
+        -3.673100922561787e27, 1.6534229920333123e27, 0.0, 0.0,
+        0.0, 0.0, 1.0, 0.0,
+        -828140552192.0, -1771307401216.0, 0.0, 1.0,
+    );
+    let (major, minor) = scale_factors(&mat);
+    assert!(major.is_normal() && minor.is_normal());
 }
 
 /// Clamp scaling factor to a power of two.
@@ -1218,3 +1566,65 @@ macro_rules! c_str {
         }
     }
 }
+
+/// This is inspired by the `weak-table` crate.
+/// It holds a Vec of weak pointers that are garbage collected as the Vec
+pub struct WeakTable {
+    inner: Vec<std::sync::Weak<Vec<u8>>>
+}
+
+impl WeakTable {
+    pub fn new() -> WeakTable {
+        WeakTable { inner: Vec::new() }
+    }
+    pub fn insert(&mut self, x: std::sync::Weak<Vec<u8>>) {
+        if self.inner.len() == self.inner.capacity() {
+            self.remove_expired();
+
+            // We want to make sure that we change capacity()
+            // even if remove_expired() removes some entries
+            // so that we don't repeatedly hit remove_expired()
+            if self.inner.len() * 3 < self.inner.capacity() {
+                // We use a different multiple for shrinking then
+                // expanding so that we we don't accidentally
+                // oscilate.
+                self.inner.shrink_to_fit();
+            } else {
+                // Otherwise double our size
+                self.inner.reserve(self.inner.len())
+            }
+        }
+        self.inner.push(x);
+    }
+
+    fn remove_expired(&mut self) {
+        self.inner.retain(|x| x.strong_count() > 0)
+    }
+
+    pub fn iter(&self) -> impl Iterator<Item = Arc<Vec<u8>>> + '_ {
+        self.inner.iter().filter_map(|x| x.upgrade())
+    }
+}
+
+#[test]
+fn weak_table() {
+    let mut tbl = WeakTable::new();
+    let mut things = Vec::new();
+    let target_count = 50;
+    for _ in 0..target_count {
+        things.push(Arc::new(vec![4]));
+    }
+    for i in &things {
+        tbl.insert(Arc::downgrade(i))
+    }
+    assert_eq!(tbl.inner.len(), target_count);
+    drop(things);
+    assert_eq!(tbl.iter().count(), 0);
+
+    // make sure that we shrink the table if it gets too big
+    // by adding a bunch of dead items
+    for _ in 0..target_count*2 {
+        tbl.insert(Arc::downgrade(&Arc::new(vec![5])))
+    }
+    assert!(tbl.inner.capacity() <= 4);
+}
