@@ -5,7 +5,8 @@
  * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
 
 #include "mozilla/Attributes.h"
-#include "mozilla/Utf8.h"  // mozilla::Utf8Unit
+#include "mozilla/ArrayUtils.h"  // mozilla::ArrayLength
+#include "mozilla/Utf8.h"        // mozilla::Utf8Unit
 
 #include <cstdarg>
 
@@ -23,14 +24,15 @@
 #include "js/CompilationAndEvaluation.h"
 #include "js/CompileOptions.h"         // JS::CompileOptions
 #include "js/friend/JSMEnvironment.h"  // JS::ExecuteInJSMEnvironment, JS::GetJSMEnvironmentOfScriptedCaller, JS::NewJSMEnvironment
-#include "js/Object.h"                 // JS::GetCompartment
+#include "js/loader/ModuleLoadRequest.h"
+#include "js/Object.h"  // JS::GetCompartment
 #include "js/Printf.h"
+#include "js/PropertyAndElement.h"  // JS_DefineFunctions, JS_DefineProperty, JS_Enumerate, JS_GetElement, JS_GetProperty, JS_GetPropertyById, JS_HasOwnProperty, JS_HasOwnPropertyById, JS_SetProperty, JS_SetPropertyById
 #include "js/PropertySpec.h"
 #include "js/SourceText.h"  // JS::SourceText
 #include "nsCOMPtr.h"
 #include "nsDirectoryServiceDefs.h"
 #include "nsDirectoryServiceUtils.h"
-#include "nsExceptionHandler.h"
 #include "nsIComponentManager.h"
 #include "mozilla/Module.h"
 #include "nsIFile.h"
@@ -48,6 +50,8 @@
 #include "nsReadableUtils.h"
 #include "nsXULAppAPI.h"
 #include "WrapperFactory.h"
+#include "JSMEnvironmentProxy.h"
+#include "ModuleEnvironmentProxy.h"
 
 #include "AutoMemMap.h"
 #include "ScriptPreloader-inl.h"
@@ -63,6 +67,7 @@
 #include "mozilla/ScriptPreloader.h"
 #include "mozilla/ScopeExit.h"
 #include "mozilla/dom/AutoEntryScript.h"
+#include "mozilla/dom/ReferrerPolicyBinding.h"
 #include "mozilla/dom/ScriptSettings.h"
 #include "mozilla/ResultExtensions.h"
 #include "mozilla/UniquePtrExtensions.h"
@@ -74,7 +79,8 @@ using namespace mozilla::loader;
 using namespace xpc;
 using namespace JS;
 
-#define JS_CACHE_PREFIX(aType) "jsloader/" aType
+#define JS_CACHE_PREFIX(aScopeType, aCompilationTarget) \
+  "jsloader/" aScopeType "/" aCompilationTarget
 
 /**
  * Buffer sizes for serialization and deserialization of scripts.
@@ -101,6 +107,39 @@ static LazyLogModule gJSCLLog("JSComponentLoader");
 #define ERROR_UNINITIALIZED_SYMBOL \
   "%s - Symbol '%s' accessed before initialization. Cyclic import?"
 
+static constexpr char JSMSuffix[] = ".jsm";
+static constexpr size_t JSMSuffixLength = mozilla::ArrayLength(JSMSuffix) - 1;
+static constexpr char MJSSuffix[] = ".sys.mjs";
+static constexpr size_t MJSSuffixLength = mozilla::ArrayLength(MJSSuffix) - 1;
+
+static bool IsJSM(const nsACString& aLocation) {
+  if (aLocation.Length() < JSMSuffixLength) {
+    return false;
+  }
+  const auto ext = Substring(aLocation, aLocation.Length() - JSMSuffixLength);
+  return ext == JSMSuffix;
+}
+
+static bool IsMJS(const nsACString& aLocation) {
+  if (aLocation.Length() < MJSSuffixLength) {
+    return false;
+  }
+  const auto ext = Substring(aLocation, aLocation.Length() - MJSSuffixLength);
+  return ext == MJSSuffix;
+}
+
+static void ToJSM(const nsACString& aLocation, nsAutoCString& aOut) {
+  MOZ_ASSERT(IsMJS(aLocation));
+  aOut = Substring(aLocation, 0, aLocation.Length() - MJSSuffixLength);
+  aOut += JSMSuffix;
+}
+
+static void ToMJS(const nsACString& aLocation, nsAutoCString& aOut) {
+  MOZ_ASSERT(IsJSM(aLocation));
+  aOut = Substring(aLocation, 0, aLocation.Length() - JSMSuffixLength);
+  aOut += MJSSuffix;
+}
+
 static bool Dump(JSContext* cx, unsigned argc, Value* vp) {
   if (!nsJSUtils::DumpEnabled()) {
     return true;
@@ -217,7 +256,9 @@ mozJSComponentLoader::mozJSComponentLoader()
 class MOZ_STACK_CLASS ComponentLoaderInfo {
  public:
   explicit ComponentLoaderInfo(const nsACString& aLocation)
-      : mLocation(aLocation) {}
+      : mLocation(&aLocation), mIsModule(false) {}
+  explicit ComponentLoaderInfo(nsIURI* aURI, bool aIsModule)
+      : mLocation(nullptr), mURI(aURI), mIsModule(aIsModule) {}
 
   nsIIOService* IOService() {
     MOZ_ASSERT(mIOService);
@@ -238,7 +279,8 @@ class MOZ_STACK_CLASS ComponentLoaderInfo {
   }
   nsresult EnsureURI() {
     BEGIN_ENSURE(URI, IOService);
-    return mIOService->NewURI(mLocation, nullptr, nullptr,
+    MOZ_ASSERT(mLocation);
+    return mIOService->NewURI(*mLocation, nullptr, nullptr,
                               getter_AddRefs(mURI));
   }
 
@@ -269,7 +311,10 @@ class MOZ_STACK_CLASS ComponentLoaderInfo {
     return ResolveURI(mURI, getter_AddRefs(mResolvedURI));
   }
 
-  const nsACString& Key() { return mLocation; }
+  const nsACString& Key() {
+    MOZ_ASSERT(mLocation);
+    return *mLocation;
+  }
 
   [[nodiscard]] nsresult GetLocation(nsCString& aLocation) {
     nsresult rv = EnsureURI();
@@ -277,12 +322,15 @@ class MOZ_STACK_CLASS ComponentLoaderInfo {
     return mURI->GetSpec(aLocation);
   }
 
+  bool IsModule() const { return mIsModule; }
+
  private:
-  const nsACString& mLocation;
+  const nsACString* mLocation;
   nsCOMPtr<nsIIOService> mIOService;
   nsCOMPtr<nsIURI> mURI;
   nsCOMPtr<nsIChannel> mScriptChannel;
   nsCOMPtr<nsIURI> mResolvedURI;
+  const bool mIsModule;
 };
 
 template <typename... Args>
@@ -309,7 +357,6 @@ mozJSComponentLoader::~mozJSComponentLoader() {
   MOZ_ASSERT(!mInitialized,
              "UnloadModules() was not explicitly called before cleaning up "
              "mozJSComponentLoader");
-
   if (mInitialized) {
     UnloadModules();
   }
@@ -319,25 +366,6 @@ mozJSComponentLoader::~mozJSComponentLoader() {
 
 StaticRefPtr<mozJSComponentLoader> mozJSComponentLoader::sSelf;
 
-// For terrible compatibility reasons, we need to consider both the global
-// lexical environment and the global of modules when searching for exported
-// symbols.
-static JSObject* ResolveModuleObjectProperty(JSContext* aCx,
-                                             HandleObject aModObj,
-                                             const char* name) {
-  if (JS_HasExtensibleLexicalEnvironment(aModObj)) {
-    RootedObject lexical(aCx, JS_ExtensibleLexicalEnvironment(aModObj));
-    bool found;
-    if (!JS_HasOwnProperty(aCx, lexical, name, &found)) {
-      return nullptr;
-    }
-    if (found) {
-      return lexical;
-    }
-  }
-  return aModObj;
-}
-
 const mozilla::Module* mozJSComponentLoader::LoadModule(FileLocation& aFile) {
   if (!NS_IsMainThread()) {
     MOZ_ASSERT(false, "Don't use JS components off the main thread");
@@ -464,6 +492,11 @@ void mozJSComponentLoader::InitStatics() {
 void mozJSComponentLoader::Unload() {
   if (sSelf) {
     sSelf->UnloadModules();
+
+    if (sSelf->mModuleLoader) {
+      sSelf->mModuleLoader->Shutdown();
+      sSelf->mModuleLoader = nullptr;
+    }
   }
 }
 
@@ -568,7 +601,9 @@ void mozJSComponentLoader::CreateLoaderGlobal(JSContext* aCx,
   auto backstagePass = MakeRefPtr<BackstagePass>();
   RealmOptions options;
 
-  options.creationOptions().setNewCompartmentInSystemZone();
+  options.creationOptions()
+      .setFreezeBuiltins(true)
+      .setNewCompartmentInSystemZone();
   xpc::SetPrefableRealmOptions(options);
 
   // Defer firing OnNewGlobalObject until after the __URI__ property has
@@ -594,6 +629,11 @@ void mozJSComponentLoader::CreateLoaderGlobal(JSContext* aCx,
   // about:memory may use that information
   xpc::SetLocationForGlobal(global, aLocation);
 
+  MOZ_ASSERT(!mModuleLoader);
+  RefPtr<ComponentScriptLoader> scriptLoader = new ComponentScriptLoader;
+  mModuleLoader = new ComponentModuleLoader(scriptLoader, backstagePass);
+  backstagePass->InitModuleLoader(mModuleLoader);
+
   aGlobal.set(global);
 }
 
@@ -616,59 +656,93 @@ JSObject* mozJSComponentLoader::GetSharedGlobal(JSContext* aCx) {
   return mLoaderGlobal;
 }
 
-JSObject* mozJSComponentLoader::PrepareObjectForLocation(
-    JSContext* aCx, nsIFile* aComponentFile, nsIURI* aURI, bool* aRealFile) {
-  nsAutoCString nativePath;
-  NS_ENSURE_SUCCESS(aURI->GetSpec(nativePath), nullptr);
+/* static */
+nsresult mozJSComponentLoader::LoadSingleModuleScript(
+    JSContext* aCx, nsIURI* aURI, MutableHandleScript aScriptOut) {
+  ComponentLoaderInfo info(aURI, true);
+  nsresult rv = info.EnsureResolvedURI();
+  NS_ENSURE_SUCCESS(rv, rv);
 
+  nsCOMPtr<nsIFile> sourceFile;
+  rv = GetSourceFile(info.ResolvedURI(), getter_AddRefs(sourceFile));
+  NS_ENSURE_SUCCESS(rv, rv);
+
+  bool realFile = LocationIsRealFile(aURI);
+
+  RootedScript script(aCx);
+  return GetScriptForLocation(aCx, info, sourceFile, realFile, aScriptOut);
+}
+
+/* static */
+nsresult mozJSComponentLoader::GetSourceFile(nsIURI* aResolvedURI,
+                                             nsIFile** aSourceFileOut) {
+  // Get the JAR if there is one.
+  nsCOMPtr<nsIJARURI> jarURI;
+  nsresult rv = NS_OK;
+  jarURI = do_QueryInterface(aResolvedURI, &rv);
+  nsCOMPtr<nsIFileURL> baseFileURL;
+  if (NS_SUCCEEDED(rv)) {
+    nsCOMPtr<nsIURI> baseURI;
+    while (jarURI) {
+      jarURI->GetJARFile(getter_AddRefs(baseURI));
+      jarURI = do_QueryInterface(baseURI, &rv);
+    }
+    baseFileURL = do_QueryInterface(baseURI, &rv);
+    NS_ENSURE_SUCCESS(rv, rv);
+  } else {
+    baseFileURL = do_QueryInterface(aResolvedURI, &rv);
+    NS_ENSURE_SUCCESS(rv, rv);
+  }
+
+  return baseFileURL->GetFile(aSourceFileOut);
+}
+
+/* static */
+bool mozJSComponentLoader::LocationIsRealFile(nsIURI* aURI) {
+  // We need to be extra careful checking for URIs pointing to files.
+  // EnsureFile may not always get called, especially on resource URIs so we
+  // need to call GetFile to make sure this is a valid file.
+  nsresult rv = NS_OK;
+  nsCOMPtr<nsIFileURL> fileURL = do_QueryInterface(aURI, &rv);
+  nsCOMPtr<nsIFile> testFile;
+  if (NS_SUCCEEDED(rv)) {
+    fileURL->GetFile(getter_AddRefs(testFile));
+  }
+
+  return bool(testFile);
+}
+
+JSObject* mozJSComponentLoader::PrepareObjectForLocation(
+    JSContext* aCx, nsIFile* aComponentFile, nsIURI* aURI, bool aRealFile) {
   RootedObject globalObj(aCx, GetSharedGlobal(aCx));
+  NS_ENSURE_TRUE(globalObj, nullptr);
+  JSAutoRealm ar(aCx, globalObj);
 
   // |thisObj| is the object we set properties on for a particular .jsm.
-  RootedObject thisObj(aCx, globalObj);
+  RootedObject thisObj(aCx, JS::NewJSMEnvironment(aCx));
   NS_ENSURE_TRUE(thisObj, nullptr);
 
-  JSAutoRealm ar(aCx, thisObj);
+  if (aRealFile) {
+    if (XRE_IsParentProcess()) {
+      RootedObject locationObj(aCx);
 
-  thisObj = JS::NewJSMEnvironment(aCx);
-  NS_ENSURE_TRUE(thisObj, nullptr);
+      nsresult rv = nsXPConnect::XPConnect()->WrapNative(
+          aCx, thisObj, aComponentFile, NS_GET_IID(nsIFile),
+          locationObj.address());
+      NS_ENSURE_SUCCESS(rv, nullptr);
+      NS_ENSURE_TRUE(locationObj, nullptr);
 
-  *aRealFile = false;
-
-  // need to be extra careful checking for URIs pointing to files
-  // EnsureFile may not always get called, especially on resource URIs
-  // so we need to call GetFile to make sure this is a valid file
-  {
-    // Create an extra scope so that ~nsCOMPtr will run before the returned
-    // JSObject* is placed on the stack, since otherwise a GC in the destructor
-    // would invalidate the return value.
-    nsresult rv = NS_OK;
-    nsCOMPtr<nsIFileURL> fileURL = do_QueryInterface(aURI, &rv);
-    nsCOMPtr<nsIFile> testFile;
-    if (NS_SUCCEEDED(rv)) {
-      fileURL->GetFile(getter_AddRefs(testFile));
-    }
-
-    if (testFile) {
-      *aRealFile = true;
-
-      if (XRE_IsParentProcess()) {
-        RootedObject locationObj(aCx);
-
-        rv = nsXPConnect::XPConnect()->WrapNative(aCx, thisObj, aComponentFile,
-                                                  NS_GET_IID(nsIFile),
-                                                  locationObj.address());
-        NS_ENSURE_SUCCESS(rv, nullptr);
-        NS_ENSURE_TRUE(locationObj, nullptr);
-
-        if (!JS_DefineProperty(aCx, thisObj, "__LOCATION__", locationObj, 0)) {
-          return nullptr;
-        }
+      if (!JS_DefineProperty(aCx, thisObj, "__LOCATION__", locationObj, 0)) {
+        return nullptr;
       }
     }
   }
 
   // Expose the URI from which the script was imported through a special
   // variable that we insert into the JSM.
+  nsAutoCString nativePath;
+  NS_ENSURE_SUCCESS(aURI->GetSpec(nativePath), nullptr);
+
   RootedString exposedUri(
       aCx, JS_NewStringCopyN(aCx, nativePath.get(), nativePath.Length()));
   NS_ENSURE_TRUE(exposedUri, nullptr);
@@ -722,126 +796,31 @@ nsresult mozJSComponentLoader::ObjectForLocation(
   jsapi.Init();
   JSContext* cx = jsapi.cx();
 
-  bool realFile = false;
   nsresult rv = aInfo.EnsureURI();
   NS_ENSURE_SUCCESS(rv, rv);
+
+  bool realFile = LocationIsRealFile(aInfo.URI());
+
   RootedObject obj(
-      cx, PrepareObjectForLocation(cx, aComponentFile, aInfo.URI(), &realFile));
+      cx, PrepareObjectForLocation(cx, aComponentFile, aInfo.URI(), realFile));
   NS_ENSURE_TRUE(obj, NS_ERROR_FAILURE);
   MOZ_ASSERT(!JS_IsGlobalObject(obj));
 
   JSAutoRealm ar(cx, obj);
 
   RootedScript script(cx);
-
-  nsAutoCString nativePath;
-  rv = aInfo.URI()->GetSpec(nativePath);
-  NS_ENSURE_SUCCESS(rv, rv);
-
-  // Before compiling the script, first check to see if we have it in
-  // the startupcache.  Note: as a rule, startupcache errors are not fatal
-  // to loading the script, since we can always slow-load.
-
-  bool writeToCache = false;
-  StartupCache* cache = StartupCache::GetSingleton();
-
-  aInfo.EnsureResolvedURI();
-
-  nsAutoCString cachePath(JS_CACHE_PREFIX("non-syntactic"));
-  rv = PathifyURI(aInfo.ResolvedURI(), cachePath);
-  NS_ENSURE_SUCCESS(rv, rv);
-
-  CompileOptions options(cx);
-  ScriptPreloader::FillCompileOptionsForCachedScript(options);
-  options.setFileAndLine(nativePath.get(), 1);
-  options.setForceStrictMode();
-  options.setNonSyntacticScope(true);
-
-  script =
-      ScriptPreloader::GetSingleton().GetCachedScript(cx, options, cachePath);
-  if (!script && cache) {
-    ReadCachedScript(cache, cachePath, cx, options, &script);
-  }
-
-  if (script) {
-    LOG(("Successfully loaded %s from startupcache\n", nativePath.get()));
-  } else if (cache) {
-    // This is ok, it just means the script is not yet in the
-    // cache. Could mean that the cache was corrupted and got removed,
-    // but either way we're going to write this out.
-    writeToCache = true;
-    // ReadCachedScript may have set a pending exception.
-    JS_ClearPendingException(cx);
-  }
-
-  if (!script) {
-    // The script wasn't in the cache , so compile it now.
-    LOG(("Slow loading %s\n", nativePath.get()));
-
-    // If we can no longer write to caches, we should stop using lazy sources
-    // and instead let normal syntax parsing occur. This can occur in content
-    // processes after the ScriptPreloader is flushed where we can read but no
-    // longer write.
-    if (!cache && !ScriptPreloader::GetSingleton().Active()) {
-      options.setSourceIsLazy(false);
-    }
-
-    if (realFile) {
-      AutoMemMap map;
-      MOZ_TRY(map.init(aComponentFile));
-
-      // Note: exceptions will get handled further down;
-      // don't early return for them here.
-      auto buf = map.get<char>();
-
-      JS::SourceText<mozilla::Utf8Unit> srcBuf;
-      if (srcBuf.init(cx, buf.get(), map.size(),
-                      JS::SourceOwnership::Borrowed)) {
-        script = Compile(cx, options, srcBuf);
-      } else {
-        MOZ_ASSERT(!script);
-      }
-    } else {
-      nsCString str;
-      MOZ_TRY_VAR(str, ReadScript(aInfo));
-
-      JS::SourceText<mozilla::Utf8Unit> srcBuf;
-      if (srcBuf.init(cx, str.get(), str.Length(),
-                      JS::SourceOwnership::Borrowed)) {
-        script = Compile(cx, options, srcBuf);
-      } else {
-        MOZ_ASSERT(!script);
-      }
-    }
+  rv = GetScriptForLocation(cx, aInfo, aComponentFile, realFile, &script,
+                            aLocation);
+  if (NS_FAILED(rv)) {
     // Propagate the exception, if one exists. Also, don't leave the stale
     // exception on this context.
-    if (!script && aPropagateExceptions && jsapi.HasException()) {
+    if (aPropagateExceptions && jsapi.HasException()) {
       if (!jsapi.StealException(aException)) {
         return NS_ERROR_OUT_OF_MEMORY;
       }
     }
-  }
 
-  if (!script) {
-    return NS_ERROR_FAILURE;
-  }
-
-  MOZ_ASSERT_IF(ScriptPreloader::GetSingleton().Active(), options.sourceIsLazy);
-  ScriptPreloader::GetSingleton().NoteScript(nativePath, cachePath, script);
-
-  if (writeToCache) {
-    MOZ_ASSERT(options.sourceIsLazy);
-
-    // We successfully compiled the script, so cache it.
-    rv = WriteCachedScript(cache, cachePath, cx, script);
-
-    // Don't treat failure to write as fatal, since we might be working
-    // with a read-only cache.
-    if (NS_SUCCEEDED(rv)) {
-      LOG(("Successfully wrote to cache\n"));
-    } else {
-      LOG(("Failed to write to cache\n"));
-    }
+    return rv;
   }
 
   // Assign aObject here so that it's available to recursive imports.
@@ -862,7 +841,7 @@ nsresult mozJSComponentLoader::ObjectForLocation(
     bool executeOk = false;
     if (JS_IsGlobalObject(obj)) {
       JS::RootedValue rval(cx);
-      executeOk = JS::CloneAndExecuteScript(aescx, script, &rval);
+      executeOk = JS_ExecuteScript(aescx, script, &rval);
     } else {
       executeOk = JS::ExecuteInJSMEnvironment(aescx, script, obj);
     }
@@ -879,12 +858,155 @@ nsresult mozJSComponentLoader::ObjectForLocation(
     }
   }
 
-  /* Freed when we remove from the table. */
-  *aLocation = ToNewCString(nativePath, mozilla::fallible);
-  if (!*aLocation) {
-    aObject.set(nullptr);
-    aTableScript.set(nullptr);
-    return NS_ERROR_OUT_OF_MEMORY;
+  return rv;
+}
+
+/* static */
+nsresult mozJSComponentLoader::GetScriptForLocation(
+    JSContext* aCx, ComponentLoaderInfo& aInfo, nsIFile* aComponentFile,
+    bool aUseMemMap, MutableHandleScript aScriptOut, char** aLocationOut) {
+  // JS compilation errors are returned via an exception on the context.
+  MOZ_ASSERT(!JS_IsExceptionPending(aCx));
+
+  aScriptOut.set(nullptr);
+
+  nsAutoCString nativePath;
+  nsresult rv = aInfo.URI()->GetSpec(nativePath);
+  NS_ENSURE_SUCCESS(rv, rv);
+
+  // Before compiling the script, first check to see if we have it in
+  // the preloader cache or the startupcache.  Note: as a rule, preloader cache
+  // errors and startupcache errors are not fatal to loading the script, since
+  // we can always slow-load.
+
+  bool storeIntoStartupCache = false;
+  StartupCache* cache = StartupCache::GetSingleton();
+
+  aInfo.EnsureResolvedURI();
+
+  nsAutoCString cachePath;
+  if (aInfo.IsModule()) {
+    rv = PathifyURI(JS_CACHE_PREFIX("non-syntactic", "module"),
+                    aInfo.ResolvedURI(), cachePath);
+  } else {
+    rv = PathifyURI(JS_CACHE_PREFIX("non-syntactic", "script"),
+                    aInfo.ResolvedURI(), cachePath);
+  }
+  NS_ENSURE_SUCCESS(rv, rv);
+
+  JS::DecodeOptions decodeOptions;
+  ScriptPreloader::FillDecodeOptionsForCachedStencil(decodeOptions);
+
+  RefPtr<JS::Stencil> stencil =
+      ScriptPreloader::GetSingleton().GetCachedStencil(aCx, decodeOptions,
+                                                       cachePath);
+
+  if (!stencil && cache) {
+    ReadCachedStencil(cache, cachePath, aCx, decodeOptions,
+                      getter_AddRefs(stencil));
+    if (!stencil) {
+      JS_ClearPendingException(aCx);
+
+      storeIntoStartupCache = true;
+    }
+  }
+
+  if (stencil) {
+    LOG(("Successfully loaded %s from cache\n", nativePath.get()));
+  } else {
+    // The script wasn't in the cache , so compile it now.
+    LOG(("Slow loading %s\n", nativePath.get()));
+
+    CompileOptions options(aCx);
+    ScriptPreloader::FillCompileOptionsForCachedStencil(options);
+    options.setFileAndLine(nativePath.get(), 1);
+    if (aInfo.IsModule()) {
+      options.setModule();
+      // Top level await is not supported in synchronously loaded modules.
+      options.topLevelAwait = false;
+
+      // Make all top-level `vars` available in `ModuleEnvironmentObject`.
+      options.deoptimizeModuleGlobalVars = true;
+    } else {
+      options.setForceStrictMode();
+      options.setNonSyntacticScope(true);
+    }
+
+    // If we can no longer write to caches, we should stop using lazy sources
+    // and instead let normal syntax parsing occur. This can occur in content
+    // processes after the ScriptPreloader is flushed where we can read but no
+    // longer write.
+    if (!storeIntoStartupCache && !ScriptPreloader::GetSingleton().Active()) {
+      options.setSourceIsLazy(false);
+    }
+
+    if (aUseMemMap) {
+      AutoMemMap map;
+      MOZ_TRY(map.init(aComponentFile));
+
+      // Note: exceptions will get handled further down;
+      // don't early return for them here.
+      auto buf = map.get<char>();
+
+      JS::SourceText<mozilla::Utf8Unit> srcBuf;
+      if (srcBuf.init(aCx, buf.get(), map.size(),
+                      JS::SourceOwnership::Borrowed)) {
+        stencil = CompileStencil(aCx, options, srcBuf, aInfo.IsModule());
+      }
+    } else {
+      nsCString str;
+      MOZ_TRY_VAR(str, ReadScript(aInfo));
+
+      JS::SourceText<mozilla::Utf8Unit> srcBuf;
+      if (srcBuf.init(aCx, str.get(), str.Length(),
+                      JS::SourceOwnership::Borrowed)) {
+        stencil = CompileStencil(aCx, options, srcBuf, aInfo.IsModule());
+      }
+    }
+
+#ifdef DEBUG
+    // The above shouldn't touch any options for instantiation.
+    JS::InstantiateOptions instantiateOptions(options);
+    instantiateOptions.assertDefault();
+#endif
+
+    if (!stencil) {
+      return NS_ERROR_FAILURE;
+    }
+  }
+
+  aScriptOut.set(InstantiateStencil(aCx, stencil, aInfo.IsModule()));
+  if (!aScriptOut) {
+    return NS_ERROR_FAILURE;
+  }
+
+  // ScriptPreloader::NoteScript needs to be called unconditionally, to
+  // reflect the usage into the next session's cache.
+  ScriptPreloader::GetSingleton().NoteStencil(nativePath, cachePath, stencil);
+
+  // Write to startup cache only when we didn't have any cache for the script
+  // and compiled it.
+  if (storeIntoStartupCache) {
+    MOZ_ASSERT(stencil);
+
+    // We successfully compiled the script, so cache it.
+    rv = WriteCachedStencil(cache, cachePath, aCx, stencil);
+
+    // Don't treat failure to write as fatal, since we might be working
+    // with a read-only cache.
+    if (NS_SUCCEEDED(rv)) {
+      LOG(("Successfully wrote to cache\n"));
+    } else {
+      LOG(("Failed to write to cache\n"));
+    }
+  }
+
+  /* Owned by ModuleEntry. Freed when we remove from the table. */
+  if (aLocationOut) {
+    *aLocationOut = ToNewCString(nativePath, mozilla::fallible);
+    if (!*aLocationOut) {
+      return NS_ERROR_OUT_OF_MEMORY;
+    }
   }
 
   return NS_OK;
@@ -912,6 +1034,36 @@ void mozJSComponentLoader::UnloadModules() {
   }
 }
 
+/* static */
+already_AddRefed<Stencil> mozJSComponentLoader::CompileStencil(
+    JSContext* aCx, const JS::CompileOptions& aOptions,
+    JS::SourceText<mozilla::Utf8Unit>& aSource, bool aIsModule) {
+  if (aIsModule) {
+    return CompileModuleScriptToStencil(aCx, aOptions, aSource);
+  }
+
+  return CompileGlobalScriptToStencil(aCx, aOptions, aSource);
+}
+
+/* static */
+JSScript* mozJSComponentLoader::InstantiateStencil(JSContext* aCx,
+                                                   JS::Stencil* aStencil,
+                                                   bool aIsModule) {
+  JS::InstantiateOptions instantiateOptions;
+
+  if (aIsModule) {
+    RootedObject module(aCx);
+    module = JS::InstantiateModuleStencil(aCx, instantiateOptions, aStencil);
+    if (!module) {
+      return nullptr;
+    }
+
+    return JS::GetModuleScript(module);
+  }
+
+  return JS::InstantiateGlobalStencil(aCx, instantiateOptions, aStencil);
+}
+
 nsresult mozJSComponentLoader::ImportInto(const nsACString& registryLocation,
                                           HandleValue targetValArg,
                                           JSContext* cx, uint8_t optionalArgc,
@@ -971,7 +1123,27 @@ nsresult mozJSComponentLoader::IsModuleLoaded(const nsACString& aLocation,
 
   mInitialized = true;
   ComponentLoaderInfo info(aLocation);
-  *retval = !!mImports.Get(info.Key());
+  if (mImports.Get(info.Key())) {
+    *retval = true;
+    return NS_OK;
+  }
+
+  if (IsJSM(aLocation) && mModuleLoader) {
+    nsAutoCString mjsLocation;
+    ToMJS(aLocation, mjsLocation);
+
+    ComponentLoaderInfo mjsInfo(mjsLocation);
+
+    nsresult rv = mjsInfo.EnsureURI();
+    NS_ENSURE_SUCCESS(rv, rv);
+
+    if (mModuleLoader->IsModuleFetched(mjsInfo.URI())) {
+      *retval = true;
+      return NS_OK;
+    }
+  }
+
+  *retval = false;
   return NS_OK;
 }
 
@@ -983,6 +1155,30 @@ void mozJSComponentLoader::GetLoadedModules(
   }
 }
 
+nsresult mozJSComponentLoader::GetLoadedESModules(
+    nsTArray<nsCString>& aLoadedModules) {
+  return mModuleLoader->GetFetchedModuleURLs(aLoadedModules);
+}
+
+nsresult mozJSComponentLoader::GetLoadedJSAndESModules(
+    nsTArray<nsCString>& aLoadedModules) {
+  GetLoadedModules(aLoadedModules);
+
+  nsTArray<nsCString> modules;
+  nsresult rv = GetLoadedESModules(modules);
+  NS_ENSURE_SUCCESS(rv, rv);
+
+  for (const auto& location : modules) {
+    if (IsMJS(location)) {
+      nsAutoCString jsmLocation;
+      ToJSM(location, jsmLocation);
+      aLoadedModules.AppendElement(jsmLocation);
+    }
+  }
+
+  return NS_OK;
+}
+
 void mozJSComponentLoader::GetLoadedComponents(
     nsTArray<nsCString>& aLoadedComponents) {
   aLoadedComponents.SetCapacity(mModules.Count());
@@ -1033,22 +1229,6 @@ nsresult mozJSComponentLoader::GetComponentLoadStack(
 #endif
 }
 
-static JSObject* ResolveModuleObjectPropertyById(JSContext* aCx,
-                                                 HandleObject aModObj,
-                                                 HandleId id) {
-  if (JS_HasExtensibleLexicalEnvironment(aModObj)) {
-    RootedObject lexical(aCx, JS_ExtensibleLexicalEnvironment(aModObj));
-    bool found;
-    if (!JS_HasOwnPropertyById(aCx, lexical, id, &found)) {
-      return nullptr;
-    }
-    if (found) {
-      return lexical;
-    }
-  }
-  return aModObj;
-}
-
 nsresult mozJSComponentLoader::ImportInto(const nsACString& aLocation,
                                           HandleObject targetObj, JSContext* cx,
                                           MutableHandleObject vp) {
@@ -1141,7 +1321,7 @@ nsresult mozJSComponentLoader::ExtractExports(
     symbolHolder = ResolveModuleObjectPropertyById(cx, aMod->obj, symbolId);
     if (!symbolHolder ||
         !JS_GetPropertyById(cx, symbolHolder, symbolId, &value)) {
-      RootedString symbolStr(cx, JSID_TO_STRING(symbolId));
+      RootedString symbolStr(cx, symbolId.toString());
       JS::UniqueChars bytes = JS_EncodeStringToUTF8(cx, symbolStr);
       if (!bytes) {
         return NS_ERROR_FAILURE;
@@ -1153,7 +1333,7 @@ nsresult mozJSComponentLoader::ExtractExports(
     // It's possible |value| is the uninitialized lexical MagicValue when
     // there's a cyclic import: const obj = ChromeUtils.import("parent.jsm").
     if (value.isMagic(JS_UNINITIALIZED_LEXICAL)) {
-      RootedString symbolStr(cx, JSID_TO_STRING(symbolId));
+      RootedString symbolStr(cx, symbolId.toString());
       JS::UniqueChars bytes = JS_EncodeStringToUTF8(cx, symbolStr);
       if (!bytes) {
         return NS_ERROR_FAILURE;
@@ -1167,7 +1347,7 @@ nsresult mozJSComponentLoader::ExtractExports(
     }
 
     if (!JS_SetPropertyById(cx, aExports, symbolId, value)) {
-      RootedString symbolStr(cx, JSID_TO_STRING(symbolId));
+      RootedString symbolStr(cx, symbolId.toString());
       JS::UniqueChars bytes = JS_EncodeStringToUTF8(cx, symbolStr);
       if (!bytes) {
         return NS_ERROR_FAILURE;
@@ -1179,8 +1359,7 @@ nsresult mozJSComponentLoader::ExtractExports(
     if (i == 0) {
       logBuffer.AssignLiteral("Installing symbols [ ");
     }
-    JS::UniqueChars bytes =
-        JS_EncodeStringToLatin1(cx, JSID_TO_STRING(symbolId));
+    JS::UniqueChars bytes = JS_EncodeStringToLatin1(cx, symbolId.toString());
     if (!!bytes) {
       logBuffer.Append(bytes.get());
     }
@@ -1202,6 +1381,11 @@ nsresult mozJSComponentLoader::ExtractExports(
   return NS_OK;
 }
 
+/* static */
+bool mozJSComponentLoader::IsTrustedScheme(nsIURI* aURI) {
+  return aURI->SchemeIs("resource") || aURI->SchemeIs("chrome");
+}
+
 nsresult mozJSComponentLoader::Import(JSContext* aCx,
                                       const nsACString& aLocation,
                                       JS::MutableHandleObject aModuleGlobal,
@@ -1233,25 +1417,13 @@ nsresult mozJSComponentLoader::Import(JSContext* aCx,
     // Note: This implies EnsureURI().
     MOZ_TRY(info.EnsureResolvedURI());
 
-    // get the JAR if there is one
-    nsCOMPtr<nsIJARURI> jarURI;
-    jarURI = do_QueryInterface(info.ResolvedURI(), &rv);
-    nsCOMPtr<nsIFileURL> baseFileURL;
-    if (NS_SUCCEEDED(rv)) {
-      nsCOMPtr<nsIURI> baseURI;
-      while (jarURI) {
-        jarURI->GetJARFile(getter_AddRefs(baseURI));
-        jarURI = do_QueryInterface(baseURI, &rv);
-      }
-      baseFileURL = do_QueryInterface(baseURI, &rv);
-      NS_ENSURE_SUCCESS(rv, rv);
-    } else {
-      baseFileURL = do_QueryInterface(info.ResolvedURI(), &rv);
-      NS_ENSURE_SUCCESS(rv, rv);
+    // Reject imports from untrusted sources.
+    if (!IsTrustedScheme(info.URI())) {
+      return NS_ERROR_DOM_SECURITY_ERR;
     }
 
     nsCOMPtr<nsIFile> sourceFile;
-    rv = baseFileURL->GetFile(getter_AddRefs(sourceFile));
+    rv = GetSourceFile(info.ResolvedURI(), getter_AddRefs(sourceFile));
     NS_ENSURE_SUCCESS(rv, rv);
 
     rv = info.ResolvedURI()->GetSpec(newEntry->resolvedURL);
@@ -1289,6 +1461,11 @@ nsresult mozJSComponentLoader::Import(JSContext* aCx,
         return NS_ERROR_FAILURE;
       }
 
+      if (rv == NS_ERROR_FILE_NOT_FOUND) {
+        return TryFallbackToImportModule(aCx, aLocation, aModuleGlobal,
+                                         aModuleExports, aIgnoreExports);
+      }
+
       // Something failed, but we don't know what it is, guess.
       return NS_ERROR_FILE_NOT_FOUND;
     }
@@ -1303,7 +1480,19 @@ nsresult mozJSComponentLoader::Import(JSContext* aCx,
   }
 
   MOZ_ASSERT(mod->obj, "Import table contains entry with no object");
-  aModuleGlobal.set(mod->obj);
+  JS::RootedObject globalProxy(aCx);
+  {
+    JSAutoRealm ar(aCx, mod->obj);
+
+    globalProxy = CreateJSMEnvironmentProxy(aCx, mod->obj);
+    if (!globalProxy) {
+      return NS_ERROR_FAILURE;
+    }
+  }
+  if (!JS_WrapObject(aCx, &globalProxy)) {
+    return NS_ERROR_FAILURE;
+  }
+  aModuleGlobal.set(globalProxy);
 
   JS::RootedObject exports(aCx, mod->exports);
   if (!exports && !aIgnoreExports) {
@@ -1324,6 +1513,121 @@ nsresult mozJSComponentLoader::Import(JSContext* aCx,
   return NS_OK;
 }
 
+nsresult mozJSComponentLoader::TryFallbackToImportModule(
+    JSContext* aCx, const nsACString& aLocation,
+    JS::MutableHandleObject aModuleGlobal,
+    JS::MutableHandleObject aModuleExports, bool aIgnoreExports) {
+  if (!IsJSM(aLocation)) {
+    return NS_ERROR_FILE_NOT_FOUND;
+  }
+
+  nsAutoCString mjsLocation;
+  ToMJS(aLocation, mjsLocation);
+
+  JS::RootedObject moduleNamespace(aCx);
+  nsresult rv = ImportModule(aCx, mjsLocation, &moduleNamespace);
+  NS_ENSURE_SUCCESS(rv, rv);
+
+  JS::RootedObject globalProxy(aCx);
+  {
+    JSAutoRealm ar(aCx, moduleNamespace);
+
+    JS::RootedObject moduleObject(
+        aCx, JS::GetModuleForNamespace(aCx, moduleNamespace));
+    if (!moduleObject) {
+      return NS_ERROR_FAILURE;
+    }
+
+    globalProxy = CreateModuleEnvironmentProxy(aCx, moduleObject);
+    if (!globalProxy) {
+      return NS_ERROR_FAILURE;
+    }
+  }
+  if (!JS_WrapObject(aCx, &globalProxy)) {
+    return NS_ERROR_FAILURE;
+  }
+  aModuleGlobal.set(globalProxy);
+
+  if (!aIgnoreExports) {
+    JS::RootedObject exports(aCx, moduleNamespace);
+    if (!JS_WrapObject(aCx, &exports)) {
+      return NS_ERROR_FAILURE;
+    }
+    aModuleExports.set(exports);
+  }
+
+  return NS_OK;
+}
+
+nsresult mozJSComponentLoader::ImportModule(
+    JSContext* aCx, const nsACString& aLocation,
+    JS::MutableHandleObject aModuleNamespace) {
+  using namespace JS::loader;
+
+  MOZ_ASSERT(mModuleLoader);
+
+  // Called from ChromeUtils::ImportModule.
+  nsCString str(aLocation);
+
+  AUTO_PROFILER_MARKER_TEXT(
+      "ChromeUtils.importModule", JS,
+      MarkerOptions(MarkerStack::Capture(),
+                    MarkerInnerWindowIdFromJSContext(aCx)),
+      aLocation);
+
+  RootedObject globalObj(aCx, GetSharedGlobal(aCx));
+  NS_ENSURE_TRUE(globalObj, NS_ERROR_FAILURE);
+  MOZ_ASSERT(xpc::Scriptability::Get(globalObj).Allowed());
+
+  JSAutoRealm ar(aCx, globalObj);
+
+  nsCOMPtr<nsIURI> uri;
+  nsresult rv = NS_NewURI(getter_AddRefs(uri), aLocation);
+  NS_ENSURE_SUCCESS(rv, rv);
+
+  nsCOMPtr<nsIPrincipal> principal =
+      mModuleLoader->GetGlobalObject()->PrincipalOrNull();
+  MOZ_ASSERT(principal);
+
+  RefPtr<ScriptFetchOptions> options = new ScriptFetchOptions(
+      CORS_NONE, dom::ReferrerPolicy::No_referrer, principal);
+
+  RefPtr<ComponentLoadContext> context = new ComponentLoadContext();
+
+  RefPtr<VisitedURLSet> visitedSet =
+      ModuleLoadRequest::NewVisitedSetForTopLevelImport(uri);
+
+  RefPtr<ModuleLoadRequest> request = new ModuleLoadRequest(
+      uri, options, dom::SRIMetadata(),
+      /* aReferrer = */ nullptr, context,
+      /* aIsTopLevel = */ true,
+      /* aIsDynamicImport = */ false, mModuleLoader, visitedSet, nullptr);
+
+  rv = request->StartModuleLoad();
+  NS_ENSURE_SUCCESS(rv, rv);
+
+  rv = mModuleLoader->ProcessRequests();
+  NS_ENSURE_SUCCESS(rv, rv);
+
+  MOZ_ASSERT(request->IsReadyToRun());
+  if (!request->InstantiateModuleGraph()) {
+    return NS_ERROR_FAILURE;
+  }
+
+  rv = mModuleLoader->EvaluateModuleInContext(aCx, request,
+                                              JS::ThrowModuleErrorsSync);
+  NS_ENSURE_SUCCESS(rv, rv);
+  if (JS_IsExceptionPending(aCx)) {
+    return NS_ERROR_FAILURE;
+  }
+
+  RefPtr<ModuleScript> moduleScript = request->mModuleScript;
+  JS::Rooted<JSObject*> module(aCx, moduleScript->ModuleRecord());
+  aModuleNamespace.set(JS::GetModuleNamespace(aCx, module));
+
+  return NS_OK;
+}
+
 nsresult mozJSComponentLoader::Unload(const nsACString& aLocation) {
   if (!mInitialized) {
     return NS_OK;
