@@ -9,11 +9,13 @@
 #include "xpcprivate.h"
 #include "xpc_make_class.h"
 #include "mozilla/dom/BindingUtils.h"
+#include "mozilla/Maybe.h"
 #include "mozilla/Preferences.h"
 #include "js/CharacterEncoding.h"
 #include "js/Class.h"
 #include "js/Object.h"  // JS::GetClass
 #include "js/Printf.h"
+#include "js/PropertyAndElement.h"  // JS_DefineProperty, JS_DefinePropertyById, JS_GetProperty, JS_GetPropertyById
 #include "js/Symbol.h"
 
 using namespace mozilla;
@@ -342,11 +344,11 @@ static bool DefinePropertyIfFound(
 
       bool defineProperty = false;
       do {
-        if (!JSID_IS_STRING(id)) {
+        if (!id.isString()) {
           break;
         }
 
-        name = JS_EncodeStringToLatin1(ccx, JSID_TO_STRING(id));
+        name = JS_EncodeStringToLatin1(ccx, id.toString());
         if (!name) {
           break;
         }
@@ -550,6 +552,8 @@ bool XPC_WN_Shared_Enumerate(JSContext* cx, HandleObject obj) {
       uint16_t index;
       if (protoSet && protoSet->FindMember(name, nullptr, &index) && index == i)
         continue;
+
+      JS_MarkCrossZoneId(cx, name);
       if (!xpc_ForcePropertyResolve(cx, obj, name)) {
         return false;
       }
@@ -562,37 +566,35 @@ bool XPC_WN_Shared_Enumerate(JSContext* cx, HandleObject obj) {
 
 enum WNHelperType { WN_NOHELPER, WN_HELPER };
 
-static void WrappedNativeFinalize(JSFreeOp* fop, JSObject* obj,
+static void WrappedNativeFinalize(JS::GCContext* gcx, JSObject* obj,
                                   WNHelperType helperType) {
   const JSClass* clazz = JS::GetClass(obj);
   if (clazz->flags & JSCLASS_DOM_GLOBAL) {
     mozilla::dom::DestroyProtoAndIfaceCache(obj);
   }
-  nsISupports* p = static_cast<nsISupports*>(xpc_GetJSPrivate(obj));
-  if (!p) {
+  XPCWrappedNative* wrapper = JS::GetObjectISupports<XPCWrappedNative>(obj);
+  if (!wrapper) {
     return;
   }
 
-  XPCWrappedNative* wrapper = static_cast<XPCWrappedNative*>(p);
   if (helperType == WN_HELPER) {
-    wrapper->GetScriptable()->Finalize(wrapper, fop, obj);
+    wrapper->GetScriptable()->Finalize(wrapper, gcx, obj);
   }
   wrapper->FlatJSObjectFinalized();
 }
 
 static size_t WrappedNativeObjectMoved(JSObject* obj, JSObject* old) {
-  nsISupports* p = static_cast<nsISupports*>(xpc_GetJSPrivate(obj));
-  if (!p) {
+  XPCWrappedNative* wrapper = JS::GetObjectISupports<XPCWrappedNative>(obj);
+  if (!wrapper) {
     return 0;
   }
 
-  XPCWrappedNative* wrapper = static_cast<XPCWrappedNative*>(p);
   wrapper->FlatJSObjectMoved(obj, old);
   return 0;
 }
 
-void XPC_WN_NoHelper_Finalize(JSFreeOp* fop, JSObject* obj) {
-  WrappedNativeFinalize(fop, obj, WN_NOHELPER);
+void XPC_WN_NoHelper_Finalize(JS::GCContext* gcx, JSObject* obj) {
+  WrappedNativeFinalize(gcx, obj, WN_NOHELPER);
 }
 
 /*
@@ -652,7 +654,6 @@ static const JSClassOps XPC_WN_NoHelper_JSClassOps = {
     nullptr,                            // mayResolve
     XPC_WN_NoHelper_Finalize,           // finalize
     nullptr,                            // call
-    nullptr,                            // hasInstance
     nullptr,                            // construct
     XPCWrappedNative::Trace,            // trace
 };
@@ -663,8 +664,8 @@ const js::ClassExtension XPC_WN_JSClassExtension = {
 
 const JSClass XPC_WN_NoHelper_JSClass = {
     "XPCWrappedNative_NoHelper",
-    XPC_WRAPPER_FLAGS | JSCLASS_IS_WRAPPED_NATIVE |
-        JSCLASS_PRIVATE_IS_NSISUPPORTS | JSCLASS_FOREGROUND_FINALIZE,
+    JSCLASS_IS_WRAPPED_NATIVE | JSCLASS_HAS_RESERVED_SLOTS(1) |
+        JSCLASS_SLOT0_IS_NSISUPPORTS | JSCLASS_FOREGROUND_FINALIZE,
     &XPC_WN_NoHelper_JSClassOps,
     JS_NULL_CLASS_SPEC,
     &XPC_WN_JSClassExtension,
@@ -722,7 +723,7 @@ bool XPC_WN_Helper_Call(JSContext* cx, unsigned argc, Value* vp) {
   // N.B. we want obj to be the callee, not JS_THIS(cx, vp)
   RootedObject obj(cx, &args.callee());
 
-  XPCCallContext ccx(cx, obj, nullptr, JSID_VOIDHANDLE, args.length(),
+  XPCCallContext ccx(cx, obj, nullptr, JS::VoidHandlePropertyKey, args.length(),
                      args.array(), args.rval().address());
   if (!ccx.IsValid()) {
     return false;
@@ -740,7 +741,7 @@ bool XPC_WN_Helper_Construct(JSContext* cx, unsigned argc, Value* vp) {
     return false;
   }
 
-  XPCCallContext ccx(cx, obj, nullptr, JSID_VOIDHANDLE, args.length(),
+  XPCCallContext ccx(cx, obj, nullptr, JS::VoidHandlePropertyKey, args.length(),
                      args.array(), args.rval().address());
   if (!ccx.IsValid()) {
     return false;
@@ -751,19 +752,49 @@ bool XPC_WN_Helper_Construct(JSContext* cx, unsigned argc, Value* vp) {
   POST_HELPER_STUB
 }
 
-bool XPC_WN_Helper_HasInstance(JSContext* cx, HandleObject obj,
-                               MutableHandleValue valp, bool* bp) {
+static bool XPC_WN_Helper_HasInstance(JSContext* cx, unsigned argc, Value* vp) {
+  CallArgs args = CallArgsFromVp(argc, vp);
+  if (!args.requireAtLeast(cx, "WrappedNative[Symbol.hasInstance]", 1)) {
+    return false;
+  }
+
+  if (!args.thisv().isObject()) {
+    JS_ReportErrorASCII(
+        cx, "WrappedNative[Symbol.hasInstance]: unexpected this value");
+    return false;
+  }
+
+  RootedObject obj(cx, &args.thisv().toObject());
+  RootedValue val(cx, args.get(0));
+
   bool retval2;
   PRE_HELPER_STUB
-  HasInstance(wrapper, cx, obj, valp, &retval2, &retval);
-  *bp = retval2;
+  HasInstance(wrapper, cx, obj, val, &retval2, &retval);
+  args.rval().setBoolean(retval2);
   POST_HELPER_STUB
 }
 
-void XPC_WN_Helper_Finalize(JSFreeOp* fop, JSObject* obj) {
-  WrappedNativeFinalize(fop, obj, WN_HELPER);
+void XPC_WN_Helper_Finalize(JS::GCContext* gcx, JSObject* obj) {
+  WrappedNativeFinalize(gcx, obj, WN_HELPER);
 }
 
+// RAII class used to store the wrapper in the context when resolving a lazy
+// property on its JS reflector. This is used by XPC_WN_MaybeResolving to allow
+// adding properties while resolving.
+class MOZ_RAII AutoSetResolvingWrapper {
+ public:
+  AutoSetResolvingWrapper(XPCCallContext& ccx, XPCWrappedNative* wrapper)
+      : mCcx(ccx), mOldResolvingWrapper(ccx.SetResolvingWrapper(wrapper)) {}
+
+  ~AutoSetResolvingWrapper() {
+    (void)mCcx.SetResolvingWrapper(mOldResolvingWrapper);
+  }
+
+ private:
+  XPCCallContext& mCcx;
+  XPCWrappedNative* mOldResolvingWrapper;
+};
+
 bool XPC_WN_Helper_Resolve(JSContext* cx, HandleObject obj, HandleId id,
                            bool* resolvedp) {
   nsresult rv = NS_OK;
@@ -776,19 +807,30 @@ bool XPC_WN_Helper_Resolve(JSContext* cx, HandleObject obj, HandleId id,
   RootedId old(cx, ccx.SetResolveName(id));
 
   nsCOMPtr<nsIXPCScriptable> scr = wrapper->GetScriptable();
+
+  // Resolve a Symbol.hasInstance property if we want custom `instanceof`
+  // behavior.
+  if (scr && scr->WantHasInstance() &&
+      id.isWellKnownSymbol(SymbolCode::hasInstance)) {
+    mozilla::Maybe<AutoSetResolvingWrapper> asrw;
+    if (scr->AllowPropModsDuringResolve()) {
+      asrw.emplace(ccx, wrapper);
+    }
+    if (!JS_DefineFunctionById(
+            cx, obj, id, XPC_WN_Helper_HasInstance, 1,
+            JSPROP_READONLY | JSPROP_PERMANENT | JSPROP_RESOLVING)) {
+      rv = NS_ERROR_FAILURE;
+    } else {
+      resolved = true;
+    }
+  }
+
   if (scr && scr->WantResolve()) {
-    XPCWrappedNative* oldResolvingWrapper;
-    bool allowPropMods = scr->AllowPropModsDuringResolve();
-
-    if (allowPropMods) {
-      oldResolvingWrapper = ccx.SetResolvingWrapper(wrapper);
+    mozilla::Maybe<AutoSetResolvingWrapper> asrw;
+    if (scr->AllowPropModsDuringResolve()) {
+      asrw.emplace(ccx, wrapper);
     }
-
     rv = scr->Resolve(wrapper, cx, obj, id, &resolved, &retval);
-
-    if (allowPropMods) {
-      (void)ccx.SetResolvingWrapper(oldResolvingWrapper);
-    }
   }
 
   old = ccx.SetResolveName(old);
@@ -815,11 +857,10 @@ bool XPC_WN_Helper_Resolve(JSContext* cx, HandleObject obj, HandleId id,
       XPCWrappedNative* wrapperForInterfaceNames =
           (scr && scr->DontReflectInterfaceNames()) ? nullptr : wrapper;
 
-      XPCWrappedNative* oldResolvingWrapper = ccx.SetResolvingWrapper(wrapper);
+      AutoSetResolvingWrapper asrw(ccx, wrapper);
       retval = DefinePropertyIfFound(
           ccx, obj, id, set, iface, member, wrapper->GetScope(), false,
           wrapperForInterfaceNames, nullptr, scr, JSPROP_ENUMERATE, resolvedp);
-      (void)ccx.SetResolvingWrapper(oldResolvingWrapper);
     }
   }
 
@@ -907,7 +948,7 @@ bool XPC_WN_CallMethod(JSContext* cx, unsigned argc, Value* vp) {
   }
 
   obj = FixUpThisIfBroken(obj, funobj);
-  XPCCallContext ccx(cx, obj, funobj, JSID_VOIDHANDLE, args.length(),
+  XPCCallContext ccx(cx, obj, funobj, JS::VoidHandlePropertyKey, args.length(),
                      args.array(), vp);
   XPCWrappedNative* wrapper = ccx.GetWrapper();
   THROW_AND_RETURN_IF_BAD_WRAPPER(cx, wrapper);
@@ -936,7 +977,7 @@ bool XPC_WN_GetterSetter(JSContext* cx, unsigned argc, Value* vp) {
   RootedObject obj(cx, &args.thisv().toObject());
 
   obj = FixUpThisIfBroken(obj, funobj);
-  XPCCallContext ccx(cx, obj, funobj, JSID_VOIDHANDLE, args.length(),
+  XPCCallContext ccx(cx, obj, funobj, JS::VoidHandlePropertyKey, args.length(),
                      args.array(), vp);
   XPCWrappedNative* wrapper = ccx.GetWrapper();
   THROW_AND_RETURN_IF_BAD_WRAPPER(cx, wrapper);
@@ -964,9 +1005,22 @@ bool XPC_WN_GetterSetter(JSContext* cx, unsigned argc, Value* vp) {
 
 /***************************************************************************/
 
+/* static */
+XPCWrappedNativeProto* XPCWrappedNativeProto::Get(JSObject* obj) {
+  MOZ_ASSERT(JS::GetClass(obj) == &XPC_WN_Proto_JSClass);
+  return JS::GetMaybePtrFromReservedSlot<XPCWrappedNativeProto>(obj, ProtoSlot);
+}
+
+/* static */
+XPCWrappedNativeTearOff* XPCWrappedNativeTearOff::Get(JSObject* obj) {
+  MOZ_ASSERT(JS::GetClass(obj) == &XPC_WN_Tearoff_JSClass);
+  return JS::GetMaybePtrFromReservedSlot<XPCWrappedNativeTearOff>(obj,
+                                                                  TearOffSlot);
+}
+
 static bool XPC_WN_Proto_Enumerate(JSContext* cx, HandleObject obj) {
   MOZ_ASSERT(JS::GetClass(obj) == &XPC_WN_Proto_JSClass, "bad proto");
-  XPCWrappedNativeProto* self = (XPCWrappedNativeProto*)xpc_GetJSPrivate(obj);
+  XPCWrappedNativeProto* self = XPCWrappedNativeProto::Get(obj);
   if (!self) {
     return false;
   }
@@ -988,8 +1042,9 @@ static bool XPC_WN_Proto_Enumerate(JSContext* cx, HandleObject obj) {
     uint16_t member_count = iface->GetMemberCount();
 
     for (uint16_t k = 0; k < member_count; k++) {
-      if (!xpc_ForcePropertyResolve(cx, obj,
-                                    iface->GetMemberAt(k)->GetName())) {
+      jsid name = iface->GetMemberAt(k)->GetName();
+      JS_MarkCrossZoneId(cx, name);
+      if (!xpc_ForcePropertyResolve(cx, obj, name)) {
         return false;
       }
     }
@@ -998,17 +1053,17 @@ static bool XPC_WN_Proto_Enumerate(JSContext* cx, HandleObject obj) {
   return true;
 }
 
-static void XPC_WN_Proto_Finalize(JSFreeOp* fop, JSObject* obj) {
+static void XPC_WN_Proto_Finalize(JS::GCContext* gcx, JSObject* obj) {
   // This can be null if xpc shutdown has already happened
-  XPCWrappedNativeProto* p = (XPCWrappedNativeProto*)xpc_GetJSPrivate(obj);
+  XPCWrappedNativeProto* p = XPCWrappedNativeProto::Get(obj);
   if (p) {
-    p->JSProtoObjectFinalized(fop, obj);
+    p->JSProtoObjectFinalized(gcx, obj);
   }
 }
 
 static size_t XPC_WN_Proto_ObjectMoved(JSObject* obj, JSObject* old) {
   // This can be null if xpc shutdown has already happened
-  XPCWrappedNativeProto* p = (XPCWrappedNativeProto*)xpc_GetJSPrivate(obj);
+  XPCWrappedNativeProto* p = XPCWrappedNativeProto::Get(obj);
   if (!p) {
     return 0;
   }
@@ -1025,7 +1080,7 @@ static bool XPC_WN_OnlyIWrite_Proto_AddPropertyStub(JSContext* cx,
                                                     HandleValue v) {
   MOZ_ASSERT(JS::GetClass(obj) == &XPC_WN_Proto_JSClass, "bad proto");
 
-  XPCWrappedNativeProto* self = (XPCWrappedNativeProto*)xpc_GetJSPrivate(obj);
+  XPCWrappedNativeProto* self = XPCWrappedNativeProto::Get(obj);
   if (!self) {
     return false;
   }
@@ -1047,7 +1102,7 @@ static bool XPC_WN_Proto_Resolve(JSContext* cx, HandleObject obj, HandleId id,
                                  bool* resolvedp) {
   MOZ_ASSERT(JS::GetClass(obj) == &XPC_WN_Proto_JSClass, "bad proto");
 
-  XPCWrappedNativeProto* self = (XPCWrappedNativeProto*)xpc_GetJSPrivate(obj);
+  XPCWrappedNativeProto* self = XPCWrappedNativeProto::Get(obj);
   if (!self) {
     return false;
   }
@@ -1074,7 +1129,6 @@ static const JSClassOps XPC_WN_Proto_JSClassOps = {
     nullptr,                                  // mayResolve
     XPC_WN_Proto_Finalize,                    // finalize
     nullptr,                                  // call
-    nullptr,                                  // hasInstance
     nullptr,                                  // construct
     nullptr,                                  // trace
 };
@@ -1084,9 +1138,13 @@ static const js::ClassExtension XPC_WN_Proto_ClassExtension = {
 };
 
 const JSClass XPC_WN_Proto_JSClass = {
-    "XPC_WN_Proto_JSClass",       XPC_WRAPPER_FLAGS,
-    &XPC_WN_Proto_JSClassOps,     JS_NULL_CLASS_SPEC,
-    &XPC_WN_Proto_ClassExtension, JS_NULL_OBJECT_OPS};
+    "XPC_WN_Proto_JSClass",
+    JSCLASS_HAS_RESERVED_SLOTS(XPCWrappedNativeProto::SlotCount) |
+        JSCLASS_FOREGROUND_FINALIZE,
+    &XPC_WN_Proto_JSClassOps,
+    JS_NULL_CLASS_SPEC,
+    &XPC_WN_Proto_ClassExtension,
+    JS_NULL_OBJECT_OPS};
 
 /***************************************************************************/
 
@@ -1104,7 +1162,9 @@ static bool XPC_WN_TearOff_Enumerate(JSContext* cx, HandleObject obj) {
 
   uint16_t member_count = iface->GetMemberCount();
   for (uint16_t k = 0; k < member_count; k++) {
-    if (!xpc_ForcePropertyResolve(cx, obj, iface->GetMemberAt(k)->GetName())) {
+    jsid name = iface->GetMemberAt(k)->GetName();
+    JS_MarkCrossZoneId(cx, name);
+    if (!xpc_ForcePropertyResolve(cx, obj, name)) {
       return false;
     }
   }
@@ -1131,8 +1191,8 @@ static bool XPC_WN_TearOff_Resolve(JSContext* cx, HandleObject obj, HandleId id,
       resolvedp);
 }
 
-static void XPC_WN_TearOff_Finalize(JSFreeOp* fop, JSObject* obj) {
-  XPCWrappedNativeTearOff* p = (XPCWrappedNativeTearOff*)xpc_GetJSPrivate(obj);
+static void XPC_WN_TearOff_Finalize(JS::GCContext* gcx, JSObject* obj) {
+  XPCWrappedNativeTearOff* p = XPCWrappedNativeTearOff::Get(obj);
   if (!p) {
     return;
   }
@@ -1140,7 +1200,7 @@ static void XPC_WN_TearOff_Finalize(JSFreeOp* fop, JSObject* obj) {
 }
 
 static size_t XPC_WN_TearOff_ObjectMoved(JSObject* obj, JSObject* old) {
-  XPCWrappedNativeTearOff* p = (XPCWrappedNativeTearOff*)xpc_GetJSPrivate(obj);
+  XPCWrappedNativeTearOff* p = XPCWrappedNativeTearOff::Get(obj);
   if (!p) {
     return 0;
   }
@@ -1148,13 +1208,6 @@ static size_t XPC_WN_TearOff_ObjectMoved(JSObject* obj, JSObject* old) {
   return 0;
 }
 
-// Make sure XPC_WRAPPER_FLAGS has no reserved slots, so our
-// XPC_WN_TEAROFF_RESERVED_SLOTS value is OK.
-
-static_assert(((XPC_WRAPPER_FLAGS >> JSCLASS_RESERVED_SLOTS_SHIFT) &
-               JSCLASS_RESERVED_SLOTS_MASK) == 0,
-              "XPC_WRAPPER_FLAGS should not include any reserved slots");
-
 static const JSClassOps XPC_WN_Tearoff_JSClassOps = {
     XPC_WN_OnlyIWrite_AddPropertyStub,  // addProperty
     XPC_WN_CannotDeletePropertyStub,    // delProperty
@@ -1164,7 +1217,6 @@ static const JSClassOps XPC_WN_Tearoff_JSClassOps = {
     nullptr,                            // mayResolve
     XPC_WN_TearOff_Finalize,            // finalize
     nullptr,                            // call
-    nullptr,                            // hasInstance
     nullptr,                            // construct
     nullptr,                            // trace
 };
@@ -1175,7 +1227,7 @@ static const js::ClassExtension XPC_WN_Tearoff_JSClassExtension = {
 
 const JSClass XPC_WN_Tearoff_JSClass = {
     "WrappedNative_TearOff",
-    XPC_WRAPPER_FLAGS |
-        JSCLASS_HAS_RESERVED_SLOTS(XPC_WN_TEAROFF_RESERVED_SLOTS),
+    JSCLASS_HAS_RESERVED_SLOTS(XPCWrappedNativeTearOff::SlotCount) |
+        JSCLASS_FOREGROUND_FINALIZE,
     &XPC_WN_Tearoff_JSClassOps, JS_NULL_CLASS_SPEC,
     &XPC_WN_Tearoff_JSClassExtension};
