@@ -136,9 +136,10 @@ TEST(Jemalloc, PtrInfo)
   jemalloc_ptr_info_t info;
   Vector<char*> small, large, huge;
 
-  // For small (<= 2KiB) allocations, test every position within many possible
-  // sizes.
-  size_t small_max = stats.page_size / 2;
+  // For small (less than half the page size) allocations, test every position
+  // within many possible sizes.
+  size_t small_max =
+      stats.subpage_max ? stats.subpage_max : stats.quantum_wide_max;
   for (size_t n = 0; n <= small_max; n += 8) {
     auto p = (char*)moz_arena_malloc(arenaId, n);
     size_t usable = moz_malloc_size_of(p);
@@ -149,7 +150,7 @@ TEST(Jemalloc, PtrInfo)
     }
   }
 
-  // Similar for large (2KiB + 1 KiB .. 1MiB - 8KiB) allocations.
+  // Similar for large (small_max + 1 KiB .. 1MiB - 8KiB) allocations.
   for (size_t n = small_max + 1_KiB; n <= stats.large_max; n += 1_KiB) {
     auto p = (char*)moz_arena_malloc(arenaId, n);
     size_t usable = moz_malloc_size_of(p);
@@ -199,7 +200,7 @@ TEST(Jemalloc, PtrInfo)
   // the former.
   ASSERT_TRUE(isFreedAlloc != 0);
   ASSERT_TRUE(isFreedPage != 0);
-  ASSERT_TRUE(isFreedAlloc / isFreedPage > 10);
+  ASSERT_TRUE(isFreedAlloc / isFreedPage > 8);
 
   // Free the large allocations and recheck them.
   len = large.length();
@@ -277,7 +278,7 @@ TEST(Jemalloc, PtrInfo)
   moz_dispose_arena(arenaId);
 }
 
-size_t sSizes[] = {1,      42,      79,      918,     1.5_KiB,
+size_t sSizes[] = {1,      42,      79,      918,     1.4_KiB,
                    73_KiB, 129_KiB, 1.1_MiB, 2.6_MiB, 5.1_MiB};
 
 TEST(Jemalloc, Arenas)
@@ -382,7 +383,9 @@ class SizeClassesBetween {
 };
 
 #define ALIGNMENT_CEILING(s, alignment) \
-  (((s) + (alignment - 1)) & (~(alignment - 1)))
+  (((s) + ((alignment)-1)) & (~((alignment)-1)))
+
+#define ALIGNMENT_FLOOR(s, alignment) ((s) & (~((alignment)-1)))
 
 static bool IsSameRoundedHugeClass(size_t aSize1, size_t aSize2,
                                    jemalloc_stats_t& aStats) {
@@ -655,8 +658,9 @@ TEST(Jemalloc, JunkPoison)
   _gdb_sleep_duration = old_gdb_sleep_duration;
 #  endif
 }
+#endif  // !defined(XP_WIN) || !defined(MOZ_CODE_COVERAGE)
 
-TEST(Jemalloc, GuardRegion)
+TEST(Jemalloc, TrailingGuard)
 {
   // Disable PHC allocations for this test, because even a single PHC
   // allocation occurring can throw it off.
@@ -665,11 +669,11 @@ TEST(Jemalloc, GuardRegion)
   jemalloc_stats_t stats;
   jemalloc_stats(&stats);
 
-#  ifdef HAS_GDB_SLEEP_DURATION
+#ifdef HAS_GDB_SLEEP_DURATION
   // Avoid death tests adding some unnecessary (long) delays.
   unsigned int old_gdb_sleep_duration = _gdb_sleep_duration;
   _gdb_sleep_duration = 0;
-#  endif
+#endif
 
   arena_id_t arena = moz_create_arena();
   ASSERT_TRUE(arena != 0);
@@ -691,7 +695,6 @@ TEST(Jemalloc, GuardRegion)
   jemalloc_ptr_info_t info;
   jemalloc_ptr_info(guard_page, &info);
   ASSERT_TRUE(jemalloc_ptr_is_freed_page(&info));
-  ASSERT_TRUE(info.tag == TagFreedPage);
 
   ASSERT_DEATH_WRAP(*(char*)guard_page = 0, "");
 
@@ -702,9 +705,59 @@ TEST(Jemalloc, GuardRegion)
 
   moz_dispose_arena(arena);
 
-#  ifdef HAS_GDB_SLEEP_DURATION
+#ifdef HAS_GDB_SLEEP_DURATION
   _gdb_sleep_duration = old_gdb_sleep_duration;
-#  endif
+#endif
+}
+
+TEST(Jemalloc, LeadingGuard)
+{
+  // Disable PHC allocations for this test, because even a single PHC
+  // allocation occurring can throw it off.
+  AutoDisablePHCOnCurrentThread disable;
+
+  jemalloc_stats_t stats;
+  jemalloc_stats(&stats);
+
+#ifdef HAS_GDB_SLEEP_DURATION
+  // Avoid death tests adding some unnecessary (long) delays.
+  unsigned int old_gdb_sleep_duration = _gdb_sleep_duration;
+  _gdb_sleep_duration = 0;
+#endif
+
+  arena_id_t arena = moz_create_arena();
+  ASSERT_TRUE(arena != 0);
+
+  // Do a simple normal allocation, but force all the allocation space
+  // in the chunk to be used up. This allows us to check that we get
+  // the safe area right in the logic that follows (all memory will be
+  // committed and initialized), and it forces this pointer to the start
+  // of the zone to sit at the very start of the usable chunk area.
+  void* ptr = moz_arena_malloc(arena, stats.large_max);
+  ASSERT_TRUE(ptr != nullptr);
+  // If ptr is chunk-aligned, the above allocation went wrong.
+  void* chunk_start = (void*)ALIGNMENT_FLOOR((uintptr_t)ptr, stats.chunksize);
+  ASSERT_NE((uintptr_t)ptr, (uintptr_t)chunk_start);
+  // If ptr is 1 page after the chunk start (so right after the header),
+  // we must have missed adding the guard page.
+  ASSERT_NE((uintptr_t)ptr, (uintptr_t)chunk_start + stats.page_size);
+  // The actual start depends on the amount of metadata versus the page
+  // size, so we can't check equality without pulling in too many
+  // implementation details.
+
+  // Guard page should be right before data area
+  void* guard_page = (void*)(((uintptr_t)ptr) - sizeof(void*));
+  jemalloc_ptr_info_t info;
+  jemalloc_ptr_info(guard_page, &info);
+  ASSERT_TRUE(info.tag == TagUnknown);
+  ASSERT_DEATH_WRAP(*(char*)guard_page = 0, "");
+
+  moz_arena_free(arena, ptr);
+  moz_dispose_arena(arena);
+
+#ifdef HAS_GDB_SLEEP_DURATION
+  _gdb_sleep_duration = old_gdb_sleep_duration;
+#endif
 }
 
 TEST(Jemalloc, DisposeArena)
@@ -712,11 +765,11 @@ TEST(Jemalloc, DisposeArena)
   jemalloc_stats_t stats;
   jemalloc_stats(&stats);
 
-#  ifdef HAS_GDB_SLEEP_DURATION
+#ifdef HAS_GDB_SLEEP_DURATION
   // Avoid death tests adding some unnecessary (long) delays.
   unsigned int old_gdb_sleep_duration = _gdb_sleep_duration;
   _gdb_sleep_duration = 0;
-#  endif
+#endif
 
   arena_id_t arena = moz_create_arena();
   void* ptr = moz_arena_malloc(arena, 42);
@@ -734,12 +787,12 @@ TEST(Jemalloc, DisposeArena)
 
   arena = moz_create_arena();
   ptr = moz_arena_malloc(arena, stats.chunksize * 2);
-#  ifdef MOZ_DEBUG
+#ifdef MOZ_DEBUG
   // On debug builds, we do the expensive check that arenas are empty.
   ASSERT_DEATH_WRAP(moz_dispose_arena(arena), "");
   moz_arena_free(arena, ptr);
   moz_dispose_arena(arena);
-#  else
+#else
   // Currently, the allocator can't trivially check whether the arena is empty
   // of huge allocations, so disposing of it works.
   moz_dispose_arena(arena);
@@ -747,14 +800,12 @@ TEST(Jemalloc, DisposeArena)
   ASSERT_DEATH_WRAP(free(ptr), "");
   // Likewise for realloc
   ASSERT_DEATH_WRAP(ptr = realloc(ptr, stats.chunksize * 3), "");
-#  endif
+#endif
 
   // Using the arena after it's been disposed of is MOZ_CRASH-worthy.
   ASSERT_DEATH_WRAP(moz_arena_malloc(arena, 42), "");
 
-#  ifdef HAS_GDB_SLEEP_DURATION
+#ifdef HAS_GDB_SLEEP_DURATION
   _gdb_sleep_duration = old_gdb_sleep_duration;
-#  endif
-}
-
 #endif
+}
